/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "liberty.h"
#include "patterns.h"


static void dragon_ring(int m, int n, int color, int *i, int *j);
static int compute_dragon_status(int i, int j);
static void dragon_eye(int m, int n, struct eye_data[MAX_BOARD][MAX_BOARD]);
static void escape_callback(int m, int n, int color,
			    struct pattern *pattern, int ll, void *data);


/* This basic function finds all dragons and collects some basic information
 * about them in the dragon array.
 *
 * color is the player in turn to move. This does in no way affect the
 * information collected about the dragons, but it does affect what
 * information is passed on to the move generation code. If
 * color == EMPTY no information at all is passed on to the move generation.
 */

void 
make_dragons(int color)
{
  int m, n, i, j;
  int border_color;

  /* We start with the dragon data copied from the worm data, then
   * modify it as the worms are amalgamated into larger dragons.
   */

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      dragon[m][n].size           = worm[m][n].size;
      dragon[m][n].effective_size = worm[m][n].effective_size;
      dragon[m][n].color          = worm[m][n].color;
      dragon[m][n].origini        = worm[m][n].origini;
      dragon[m][n].originj        = worm[m][n].originj;
      dragon[m][n].lunchi         = -1;
      dragon[m][n].lunchj         = -1;
      dragon[m][n].mouse_attacki  = -1;
      dragon[m][n].mouse_attackj  = -1;
      dragon[m][n].mouse_defendi  = -1;
      dragon[m][n].mouse_defendj  = -1;
      dragon[m][n].escape_route   = worm[m][n].liberties4;
      dragon[m][n].heyes          = 0;
      dragon[m][n].semeai         = 0;
				  
      half_eye[m][n].type         = 0;
      
      if (worm[m][n].origini == m && worm[m][n].originj == n)
	DEBUG(DEBUG_DRAGONS, 
	      "Initialising dragon from worm at %m, size %d, escape_route=%d\n", 
	      m, n, worm[m][n].size, worm[m][n].liberties4);
    }

  /* Amalgamate cavities. 
   *
   * Begin by finding the INESSENTIAL strings. These are defined as
   * surrounded strings which have no life potential unless part of
   * their surrounding chain can be captured. We give a conservative
   * definition of inessential: 
   *  - the genus must be zero 
   *  - there can no second order liberties
   *  - there can be no edge liberties
   *  - if it is removed from the board, the remaining cavity has
   *    border color the opposite color of the string 
   *  - it contains at most two edge vertices.
   *
   * An inessential string can be ignored for life and death purposes. It then
   * makes sense to amalgamate the surrounding cavities into a single cave
   * (empty dragon) with bordercolor the opposite of the inessential worm.
   *
   * For example, in the following situation:
   *
   *   OOOOO
   *   O.X.O
   *   OOOOO
   *
   * we find two graybordered cavities of size one. The X string is
   * inessential, so these two cavities are amalgamated into a single cave. 
   */

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if ((p[m][n])
	  && (worm[m][n].origini == m)
	  && (worm[m][n].originj == n)
	  && (worm[m][n].genus == 0)
	  && (worm[m][n].liberties2 == 0)
	  && (worm[m][n].lunchi == -1)) 
      {
	int edge;
	int borigini=-1, boriginj=-1;

	border_color=find_border(m, n, &edge, NULL);
	if ((border_color != GRAY_BORDER) && (edge<3)) {
	  dragon_ring(m, n, EMPTY, &borigini, &boriginj);
	  worm[m][n].inessential=1;
	  propagate_worm(m, n);
	  dragon[borigini][boriginj].color=border_color;
	}
      }
    }

  /* The type of reading done in make_domains doesn't need much depth */
  set_temporary_depth_values(depth, backfill_depth, 0, 0);
  make_domains(black_eye, white_eye);
  restore_depth_values();

  /* Find explicit connections patterns in database and amalgamate
   * involved dragons.
   */
  find_connections();
  
  /* Amalgamate dragons sharing an eyespace (not ko). At the same time
   * we decide to which dragon an eyespace belongs. Ko eyespaces
   * (typically false eyes but sometimes halfeyes) get assigned to an
   * arbitrary neighbor that is not the ko stone.
   */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {

      if ((black_eye[m][n].color == BLACK_BORDER) 
	  && (black_eye[m][n].origini == m)
	  && (black_eye[m][n].originj == n)) {
	if (!worm[m][n].ko
	   || (black_eye[m][n].esize > 1)) /* Only exclude living kos. */
	  dragon_eye(m, n, black_eye);
	else {
	  if (m > 0 && !worm[m-1][n].ko) {
	    black_eye[m][n].dragoni = dragon[m-1][n].origini;
	    black_eye[m][n].dragonj = dragon[m-1][n].originj;
	  } 
	  else if (m < board_size-1 && !worm[m+1][n].ko) {
	    black_eye[m][n].dragoni = dragon[m+1][n].origini;
	    black_eye[m][n].dragonj = dragon[m+1][n].originj;
	  } 
	  else if (n > 0 && !worm[m][n-1].ko) {
	    black_eye[m][n].dragoni = dragon[m][n-1].origini;
	    black_eye[m][n].dragonj = dragon[m][n-1].originj;
	  } 
	  else if (n < board_size-1 && !worm[m][n+1].ko) {
	    black_eye[m][n].dragoni = dragon[m][n+1].origini;
	    black_eye[m][n].dragonj = dragon[m][n+1].originj;
	  } 
	}
      }
	  
      if ((white_eye[m][n].color == WHITE_BORDER)
	  && (white_eye[m][n].origini == m)
	  && (white_eye[m][n].originj == n)) {
	if (!worm[m][n].ko
	    || (white_eye[m][n].esize > 1)) /* Only exclude living kos. */
	  dragon_eye(m, n, white_eye);
	else {
	  if (m > 0 && !worm[m-1][n].ko) {
	    white_eye[m][n].dragoni = dragon[m-1][n].origini;
	    white_eye[m][n].dragonj = dragon[m-1][n].originj;
	  } 
	  else if (m < board_size-1 && !worm[m+1][n].ko) {
	    white_eye[m][n].dragoni = dragon[m+1][n].origini;
	    white_eye[m][n].dragonj = dragon[m+1][n].originj;
	  } 
	  else if (n > 0 && !worm[m][n-1].ko) {
	    white_eye[m][n].dragoni = dragon[m][n-1].origini;
	    white_eye[m][n].dragonj = dragon[m][n-1].originj;
	  } 
	  else if (n < board_size-1 && !worm[m][n+1].ko) {
	    white_eye[m][n].dragoni = dragon[m][n+1].origini;
	    white_eye[m][n].dragonj = dragon[m][n+1].originj;
	  }
	}
      }
    }

  /* Find adjacent worms which can be easily captured: */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if ((worm[m][n].origini != m) 
	  || (worm[m][n].originj != n)
	  || p[m][n] == EMPTY
	  || (worm[m][n].lunchi == -1)) 
	continue;

      i=worm[m][n].lunchi;
      j=worm[m][n].lunchj;

      /* In contrast to worm lunches, a dragon lunch must also be
       * able to defend itself. 
       */
      if (worm[i][j].defendi == -1)
	continue;

      /* Tell the move generation code about the lunch. */
      if (color != EMPTY)
	add_lunch(m, n, i, j);
	
      /* If several lunches are found, we pick the juiciest.
       * First maximize cutstone, then minimize liberties.
       */
      {
	int origini = dragon[m][n].origini;
	int originj = dragon[m][n].originj;

	if ((dragon[origini][originj].lunchi == -1)
	    || (worm[i][j].cutstone
		> worm[dragon[origini][originj].lunchi]
		[dragon[origini][originj].lunchj].cutstone)
	    || ((worm[i][j].cutstone 
		 == worm[dragon[origini][originj].lunchi]
		 [dragon[origini][originj].lunchj].cutstone) 
		&& (worm[i][j].liberties
		    < worm[dragon[m][n].lunchi][dragon[m][n].lunchj].liberties))) 
	  {
	    dragon[origini][originj].lunchi = worm[i][j].origini;
	    dragon[origini][originj].lunchj = worm[i][j].originj;
	    TRACE("at %m setting %m.lunch to %m (cutstone=%d)\n",
		  m, n, origini, originj,
		  worm[i][j].origini, worm[i][j].originj, worm[i][j].cutstone);
	  }
      }
    }

  /* Propagate lunch to rest of the dragon. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      dragon[i][j].lunchi
	= dragon[dragon[i][j].origini][dragon[i][j].originj].lunchi;
      dragon[i][j].lunchj
	= dragon[dragon[i][j].origini][dragon[i][j].originj].lunchj;
    }
  
  /* In case origins of dragons got moved, put the dragons of eyes aright. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if (black_eye[i][j].dragoni != -1) {
	  int di=dragon[black_eye[i][j].dragoni]
                       [black_eye[i][j].dragonj].origini;
	  int dj=dragon[black_eye[i][j].dragoni]
	               [black_eye[i][j].dragonj].originj;
	  black_eye[i][j].dragoni=di;
	  black_eye[i][j].dragonj=dj;
      }

      if (white_eye[i][j].dragoni != -1) {
	  int di=dragon[white_eye[i][j].dragoni]
	               [white_eye[i][j].dragonj].origini;
	  int dj=dragon[white_eye[i][j].dragoni] 
	               [white_eye[i][j].dragonj].originj;
	  white_eye[i][j].dragoni=di;
	  white_eye[i][j].dragonj=dj;
      }
    }

  /* Find topological half eyes and false eyes by analyzing the
   * diagonal intersections, as described in the Texinfo
   * documentation (Eyes/Eye Topology).
   */

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      int sum;
      int ai = -1;
      int aj = -1;
      int di = -1;
      int dj = -1;

      if (black_eye[m][n].color == BLACK_BORDER
	  && !black_eye[m][n].marginal
	  && (black_eye[m][n].neighbors <= 1)
	  && (black_eye[m][n].dragoni != -1)) {
	sum = topological_eye(m, n, BLACK, &ai, &aj, &di, &dj, black_eye, white_eye);
	if (sum >= 4) {
	  half_eye[m][n].type = FALSE_EYE;
	  if ((black_eye[m][n].esize == 1) || legal(m, n, WHITE))
	    add_half_eye(m, n, black_eye, half_eye);
	}
	else if (sum == 3) {
	  half_eye[m][n].type = HALF_EYE;
#if 0
	  /* See comment in add_halfeye() */
	  half_eye[m][n].ai = ai;
	  half_eye[m][n].aj = aj;
	  half_eye[m][n].di = di;
	  half_eye[m][n].dj = dj;
	  add_half_eye(m, n, black_eye, half_eye);
#endif
	}
      }
      
      if (white_eye[m][n].color == WHITE_BORDER
	  && !white_eye[m][n].marginal
	  && (white_eye[m][n].neighbors <= 1)
	  && (white_eye[m][n].dragoni != -1)) {
	sum = topological_eye(m, n, WHITE, &ai, &aj, &di, &dj, black_eye, white_eye);
	if (sum >= 4) {
	  half_eye[m][n].type = FALSE_EYE;
	  if ((white_eye[m][n].esize == 1) || legal(m, n, BLACK))
	    add_half_eye(m, n, white_eye, half_eye);
	}
	else if (sum == 3) {
	  half_eye[m][n].type = HALF_EYE;
#if 0
	  /* See comment in add_halfeye() */
	  half_eye[m][n].ai = ai;
	  half_eye[m][n].aj = aj;
	  half_eye[m][n].di = di;
	  half_eye[m][n].dj = dj;
	  add_half_eye(m, n, white_eye, half_eye);
#endif
	}
      }
    }

  /* Pattern based modification of the eye shapes computed by
   * make_domains and halfeye analysis.
   */
  modify_eye_spaces2();
  
  /* Compute the number of eyes, half eyes, etc. in an eye space. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((black_eye[i][j].color == BLACK_BORDER) 
	  && (black_eye[i][j].origini == i)
	  && (black_eye[i][j].originj == j)) 
      {
	int max, min, attacki, attackj, defendi, defendj;

	compute_eyes(i, j, &max, &min, &attacki, &attackj,
		     &defendi, &defendj, black_eye, 1, color);
	DEBUG(DEBUG_EYES, "Black eyespace at %m: min=%d, max=%d\n",
	      i, j, min, max);
	black_eye[i][j].maxeye = max;
	black_eye[i][j].mineye = min;
	black_eye[i][j].attacki = attacki;
	black_eye[i][j].attackj = attackj;	  
	black_eye[i][j].defendi = defendi;
	black_eye[i][j].defendj = defendj;	  
	propagate_eye(i, j, black_eye);
      }

      if ((white_eye[i][j].color == WHITE_BORDER) 
	  && (white_eye[i][j].origini == i)
	  && (white_eye[i][j].originj == j)) 
      {
	int max, min, attacki, attackj, defendi, defendj;

	compute_eyes(i, j, &max, &min, &attacki, &attackj,
		     &defendi, &defendj, white_eye, 1, color);
	DEBUG(DEBUG_EYES, "White eyespace at %m: min=%d, max=%d\n",
	      i, j, min, max);
	white_eye[i][j].maxeye = max;
	white_eye[i][j].mineye = min;
	white_eye[i][j].attacki = attacki;
	white_eye[i][j].attackj = attackj;	  
	white_eye[i][j].defendi = defendi;
	white_eye[i][j].defendj = defendj;	  
	propagate_eye(i, j, white_eye);
      }
    }

  /* Now we compute the genus. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((black_eye[i][j].color == BLACK_BORDER) 
	  && (black_eye[i][j].dragoni != -1) 
	  && (black_eye[i][j].origini == i)
	  && (black_eye[i][j].originj == j)) 
      {
	m = black_eye[i][j].dragoni;
	n = black_eye[i][j].dragonj;
	assert (p[m][n] == BLACK);
	TRACE("eye at %m found for dragon at %m--augmenting genus\n",
	      i, j, m, n);
	dragon[m][n].genus += (black_eye[i][j].mineye);
	dragon[m][n].heyes += (black_eye[i][j].maxeye-black_eye[i][j].mineye);
	if (black_eye[i][j].maxeye - black_eye[i][j].mineye > 0) {
	  dragon[m][n].heyei = black_eye[i][j].attacki;
	  dragon[m][n].heyej = black_eye[i][j].attackj;
	}
      }
      if ((white_eye[i][j].color == WHITE_BORDER) 
	  && (white_eye[i][j].dragoni != -1)
	  && (white_eye[i][j].origini == i)
	  && (white_eye[i][j].originj == j)) 
      {
	m = white_eye[i][j].dragoni;
	n = white_eye[i][j].dragonj;
	assert (p[m][n] == WHITE);
	TRACE("eye at %m found for dragon at %m--augmenting genus\n",
	      i, j, m, n);
	dragon[m][n].genus += (white_eye[i][j].mineye);
	dragon[m][n].heyes += (white_eye[i][j].maxeye-white_eye[i][j].mineye);
	if (white_eye[i][j].maxeye - white_eye[i][j].mineye > 0) {
	  dragon[m][n].heyei = white_eye[i][j].attacki;
	  dragon[m][n].heyej = white_eye[i][j].attackj;
	}
      }
    }

  /* Propagate genus to rest of the dragon. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      dragon[i][j].genus
	= dragon[dragon[i][j].origini][dragon[i][j].originj].genus;

  

  /* Determine status: ALIVE, DEAD, CRITICAL or UNKNOWN */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if ((dragon[m][n].origini==m)
	  && (dragon[m][n].originj==n) && p[m][n])
	{
	  /* entirely ad-hoc : aim for -ve value to mean we need
	   * to worry about this dragon
	   * 2.5.38: Added effective_size to heuristics. /gf
	   */
	  dragon[m][n].vitality = 6 * dragon[m][n].genus 
	    + dragon[m][n].heyes + dragon[m][n].escape_route
	    + (dragon[m][n].effective_size - dragon[m][n].size)/6 - 10;
	  dragon[m][n].status=compute_dragon_status(m, n);
	  dragon[m][n].safety=dragon[m][n].status;
	  sgffile_dragon_status(m,n,dragon[m][n].status);
	}
    }
  
  if (mouse) {
    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++) {
	int attacki = -1;
	int attackj = -1;
	int defendi = -1;
	int defendj = -1;
	char goal[MAX_BOARD][MAX_BOARD];
	  
	if (p[m][n] == EMPTY
	    || (dragon[m][n].origini != m)
	    || (dragon[m][n].originj != n))
	  continue;

	if (dragon[m][n].escape_route > 5) {
	  dragon[m][n].mouse_status = UNCHECKED;
	  dragon[m][n].mouse_defendi = -1;
	  dragon[m][n].mouse_defendj = -1;
	  dragon[m][n].mouse_attacki = -1;
	  dragon[m][n].mouse_attackj = -1;
	}
	else {
	  if (mouse_attack(m, n, &attacki, &attackj, goal)) {
	    dragon[m][n].mouse_attacki = attacki;
	    dragon[m][n].mouse_attackj = attackj;
	    if ((attacki != -1 )
		&& mouse_defend(m, n, &defendi, &defendj, goal)) {
	      if (defendi != -1) {
		dragon[m][n].mouse_defendi = defendi;
		dragon[m][n].mouse_defendj = defendj;
		dragon[m][n].mouse_status = CRITICAL;
	      }
	    }
	    else {
	      dragon[m][n].mouse_status = DEAD; 
	      dragon[m][n].mouse_defendi = -1;
	      dragon[m][n].mouse_defendj = -1;
	    }
	  }
	  else {
	    dragon[m][n].mouse_status = ALIVE;
	    dragon[m][n].mouse_attacki = -1;
	    dragon[m][n].mouse_attackj = -1;
	  }
	}
      }
  }

  /* The dragon data is now correct at the origin of each dragon but
   * we need to copy it to every vertex.  
   */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      struct dragon_data *d = &(dragon[m][n]);
      dragon[m][n] = dragon[d->origini][d->originj];
    }

}



/* print status info on all dragons. (Can be invoked from gdb) 
 */
void 
show_dragons(void)
{
  static const char *cnames[] = 
    {"(empty)", "white dragon", "black dragon",
     "gray-bordered cave", "black-bordered cave", "white-bordered cave"};
  static const char *snames[] = 
    {"dead", "alive", "critical", "unknown", "unchecked"};

  int m, n;

  for (m=0;m<board_size;m++)
    for (n=0;n<board_size;n++) {
      struct worm_data *w = &(worm[m][n]);

      if (((w->origini)==m)&&((w->originj)==n)) {
	if (p[m][n]) {
	  gprintf("%m : (dragon %m) %s string of size %d (%f), genus %d: (%d,%d,%d,%d)",
		  m, n, dragon[m][n].origini, dragon[m][n].originj,
		  color_to_string(p[m][n]),
		  w->size,
		  w->effective_size,
		  w->genus,
		  w->liberties,
		  w->liberties2,
		  w->liberties3,
		  w->liberties4);
	  if (w->cutstone==1)
	    gprintf ("%o - is a potential cutting stone\n");
	  else if (w->cutstone==2)
	    gprintf("%o - is a cutting stone\n");
	  else gprintf("%o\n");

	  if (w->cutstone2 > 0)
	    gprintf("- cutstone2 = %d\n", w->cutstone2);
	  
	  if (w->attacki != -1)
	    gprintf("- attackable at %m, attack code = %d\n",
		    w->attacki, w->attackj, w->attack_code);
	  if (w->defendi != -1)
	    gprintf("- defendable at %m, defend code = %d\n",
		    w->defendi, w->defendj, w->defend_code);

	  if (w->inessential)
	    gprintf("- is inessential\n");
	  
	  if (w->ko == 1)
	    gprintf("- is a ko stone\n");
	}
	else 
	  gprintf("%m : cavity of size %d\n",m,n,w->size);
      }
    }

  gprintf("%o\n");
  for (m=0;m<board_size;m++)
    for (n=0;n<board_size;n++) {
      struct dragon_data *d = &(dragon[m][n]);

      if (((d->origini)==m) && ((d->originj)==n)) {
	if (p[m][n]) {
	  gprintf("%m : %s dragon size %d (%f), genus %d, half eyes %d, escape factor %d, status %s, vitality %d, escape2=%d%s\n",
		  m,n,
		  p[m][n]==BLACK ? "B" : "W",
		  d->size,
		  d->effective_size,
		  d->genus,
		  d->heyes,
		  d->escape_route,
		  snames[d->status],
		  d->vitality,
		  d->escape2,
		  d->safety == CRITICAL ? ", weak" : "");
	  if (d->lunchi != -1)
	    gprintf("... adjacent worm %m is lunch\n", d->lunchi, d->lunchj);
	}
	else {
	  gprintf("%m : cave of size %d and border color %s\n",
		  m,n,
		  d->size,
		  cnames[d->color]);
	  if (d->color == BLACK_BORDER || d->color == WHITE_BORDER) {
	    if (!worm[m][n].ko)
	      gprintf("... surrounded by dragon at %m\n",
		      d->borderi, d->borderj);
	    else
	      gprintf("... is a ko\n");
	  }
	}
      }
    }
}


  
/* The function dragon_ring(m, n, color, *di, *dj) amalgamates every 
 * dragon adjacent to the worm at (m, n) with the given color. The
 * amalgamated dragons then form a bigger dragon ringing the worm at
 * (m, n). 
 *
 * Amalgamation means that the individual dragons are to be combined
 * into the same dragon.  If color is not EMPTY, then p[m][n]==0 and
 * we have found an eye (possibly false). In this (*di, *dj) is set to
 * the origin of the surrounding dragon and dragon[*di][*dj].genus is
 * incremented.  
 */

static void
dragon_ring(int m, int n, int color, int *di, int *dj)
{
  int i, j;
  int dragoni=-1, dragonj=-1;

  DEBUG(DEBUG_DRAGONS, "amalgamate dragons around %m\n", m, n);
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((worm[i][j].origini==m) && (worm[i][j].originj==n)) {

	if ((i>0)&&(p[i-1][j]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i-1][j].origini;
	    dragonj=dragon[i-1][j].originj;
	  } else if ((dragoni != dragon[i-1][j].origini) 
		     || (dragonj != dragon[i-1][j].originj))
	    join_dragons(i-1, j, dragoni, dragonj);
	}

	if ((i<board_size-1)&&(p[i+1][j]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i+1][j].origini;
	    dragonj=dragon[i+1][j].originj;
	  } else if ((dragoni != dragon[i+1][j].origini) 
		     || (dragonj != dragon[i+1][j].originj))
	    join_dragons(i+1, j, dragoni, dragonj);
	}

	if ((j>0)&&(p[i][j-1]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i][j-1].origini;
	    dragonj=dragon[i][j-1].originj;
	  } else if ((dragoni != dragon[i][j-1].origini) 
		     || (dragonj != dragon[i][j-1].originj))
	      join_dragons(i, j-1, dragoni, dragonj);
	}

	if ((j<board_size-1)&&(p[i][j+1]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i][j+1].origini;
	    dragonj=dragon[i][j+1].originj;
	  } else if ((dragoni != dragon[i][j+1].origini) 
		     || (dragonj != dragon[i][j+1].originj))
	      join_dragons(i, j+1, dragoni, dragonj);
	}
      }
    }

  if (color) {
    ASSERT(p[m][n] == EMPTY, m, n); /* (m, n) is a cave */

    /* record the fact that we found an eye (could be false) */
    (dragon[dragoni][dragonj].genus)++;
  }
  *di=dragoni;   /* (m, n) is an inessential string. */
  *dj=dragonj;
}


/* 
 * dragon_eye(m, n, *di, *dj) is invoked with (m, n) the origin of an
 * eyespace. It unites all the worms adjacent to non-marginal points of
 * the eyespace into a single dragon with origin (*di, *dj). In addition
 * to marginal eye space points, amalgamation is inhibited for points
 * with the INHIBIT_CONNECTION type set.
 *
 * This is based on the older function dragon_ring. Unlike dragon_ring
 * this function does not attempt to keep track of the genus.
 */

static void
dragon_eye(int m, int n, struct eye_data eye[MAX_BOARD][MAX_BOARD])
{
  int i, j;
  int dragoni=-1, dragonj=-1;
  int color;

  /* don't amalgamate across ikken tobi */
  if ((eye[m][n].esize==3) && (eye[m][n].msize>1))
    return;

  DEBUG(DEBUG_DRAGONS, "amalgamate dragons around %m\n", m, n);
  if (eye[m][n].color==BLACK_BORDER)
    color=BLACK;
  else {
    assert(eye[m][n].color==WHITE_BORDER);
    color=WHITE;
  }

  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((eye[i][j].origini==m)
	  && (eye[i][j].originj==n)
	  && !eye[i][j].marginal
	  && !(eye[i][j].type & INHIBIT_CONNECTION)) 
      {
	if ((i>0)&&(p[i-1][j]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i-1][j].origini;
	    dragonj=dragon[i-1][j].originj;
	  } else if ((dragoni != dragon[i-1][j].origini) 
		     || (dragonj != dragon[i-1][j].originj))
	    join_dragons(i-1, j, dragoni, dragonj);
	}

	if ((i<board_size-1)&&(p[i+1][j]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i+1][j].origini;
	    dragonj=dragon[i+1][j].originj;
	  } else if ((dragoni != dragon[i+1][j].origini) 
		     || (dragonj != dragon[i+1][j].originj))
	    join_dragons(i+1, j, dragoni, dragonj);
	}

	if ((j>0)&&(p[i][j-1]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i][j-1].origini;
	    dragonj=dragon[i][j-1].originj;
	  } else if ((dragoni != dragon[i][j-1].origini) 
		     || (dragonj != dragon[i][j-1].originj))
	      join_dragons(i, j-1, dragoni, dragonj);
	}

	if ((j<board_size-1)&&(p[i][j+1]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i][j+1].origini;
	    dragonj=dragon[i][j+1].originj;
	  } else if ((dragoni != dragon[i][j+1].origini) 
		     || (dragonj != dragon[i][j+1].originj))
	      join_dragons(i, j+1, dragoni, dragonj);
	}
      }
    }

  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((eye[i][j].color == BLACK_BORDER 
	   || eye[i][j].color == WHITE_BORDER) 
	  && (eye[i][j].origini==m)
	  && (eye[i][j].originj==n)) 
      {
	eye[i][j].dragoni=dragoni;
	eye[i][j].dragonj=dragonj;
      }
    }
}


/* 
 * join_dragons amalgamates the dragon at (ai, aj) to the
 * dragon at (bi, bj).
 */

void 
join_dragons(int ai, int aj, int bi, int bj)
{
  int t, u;

  int i = dragon[ai][aj].origini;
  int j = dragon[ai][aj].originj;
  int m = dragon[bi][bj].origini;
  int n = dragon[bi][bj].originj;
  
  assert(worm[i][j].color==worm[m][n].color);

  DEBUG(DEBUG_DRAGONS,
	"joining dragon at %m with escape route %d to dragon at %m with %d\n", 
	i, j, dragon[i][j].escape_route, m, n, dragon[m][n].escape_route);

  dragon[m][n].size  = dragon[m][n].size + dragon[i][j].size;
  dragon[m][n].effective_size  = (dragon[m][n].effective_size
				  + dragon[i][j].effective_size);
  dragon[m][n].genus = dragon[m][n].genus + dragon[i][j].genus;
  if (dragon[i][j].escape_route > dragon[m][n].escape_route)
    dragon[m][n].escape_route = dragon[i][j].escape_route;

  for (t=0; t<board_size; t++)
    for (u=0; u<board_size; u++){
      if ((dragon[t][u].origini == i) && (dragon[t][u].originj == j)) {
	dragon[t][u].origini = m;
	dragon[t][u].originj = n;
      }
    }
}



/*
 * compute_dragon_status(i, j) tries to determine whether the dragon
 * at (i, j) is ALIVE, DEAD, or UNKNOWN. The algorithm is not perfect
 * and can give incorrect answers.
 *
 * The dragon is judged alive if its genus is >1. It is judged dead if
 * the genus is <2, it has no escape route, and no adjoining string can
 * be easily captured. Otherwise it is judged UNKNOWN.  */

static int 
compute_dragon_status(int i, int j)
{

  int true_genus = 2*dragon[i][j].genus + dragon[i][j].heyes;

  /* If it has two sure eyes, everything is just dandy. */
  if (true_genus > 3)
    return ALIVE;

  /* If the dragon consists of one worm, there is an attack, but 
   * no defense and there is less than one eye and one half eye,
   * the situation is hopeless. */
  if ((dragon[i][j].size==worm[i][j].size)
      && (worm[i][j].attacki != -1) 
      && (worm[i][j].defendi == -1)
      && (true_genus < 3))
    return DEAD;
  
  if ((dragon[i][j].lunchi != -1)
      && (true_genus < 3)
      && (worm[dragon[i][j].lunchi][dragon[i][j].lunchj].defendi != -1)
      && (dragon[i][j].escape_route <5))
    if ((true_genus == 2)
	|| (worm[dragon[i][j].lunchi][dragon[i][j].lunchj].size >2))
      return CRITICAL;

  if ((dragon[i][j].lunchi != -1)
      && (true_genus >= 3))
    return ALIVE;

  if ((dragon[i][j].lunchi == -1)
      || (worm[dragon[i][j].lunchi][dragon[i][j].lunchj].cutstone <2)) 
  {
    if ((true_genus < 3) && (dragon[i][j].escape_route==0))
      return DEAD;

    if ((true_genus == 3) && (dragon[i][j].escape_route < 6))
      return CRITICAL;
  }

  return UNKNOWN;
}


/*
 * compute_dragon_status2(i, j) tries to determine whether the dragon
 * at (i, j) is ALIVE, DEAD, or UNKNOWN. The algorithm is not perfect
 * and can give incorrect answers.
 *
 * This differs from compute_dragon_status() in the use of the escape2
 * field instead of the escape_route field.
 *
 * The dragon is judged alive if its genus is >1. It is judged dead if
 * the genus is <2, it has no escape route, and no adjoining string can
 * be easily captured. Otherwise it is judged UNKNOWN.  */
#if 0
static int 
compute_dragon_status2(int i, int j)
{

  int true_genus = 2*dragon[i][j].genus + dragon[i][j].heyes;

  if (true_genus > 3
      || (true_genus == 3 && dragon[i][j].lunchi != -1))
    return ALIVE;

  if ((dragon[i][j].size==worm[i][j].size)
      && (worm[i][j].attack_code != 0) 
      && (worm[i][j].defend_code == 0)
      && (true_genus < 3))
    return DEAD;
  
  if ((dragon[i][j].lunchi != -1)
      && (true_genus < 3)
      && (worm[dragon[i][j].lunchi][dragon[i][j].lunchj].defend_code != 0)
      && (dragon[i][j].escape2 < 2))
    if ((true_genus == 2)
	|| (worm[dragon[i][j].lunchi][dragon[i][j].lunchj].size >2))
      return CRITICAL;

  if ((dragon[i][j].lunchi == -1)
      || (worm[dragon[i][j].lunchi][dragon[i][j].lunchj].cutstone < 2)) {
    if ((true_genus < 3) && (dragon[i][j].escape2 == 0))
      return DEAD;

    if ((true_genus == 3) && (dragon[i][j].escape2 < 2))
      return CRITICAL;
  }

  return UNKNOWN;
}
#endif


int route_used[MAX_BOARD][MAX_BOARD];
int escape_routes;

/* Compute a new escape route measure. See Texinfo documentation
 * for the escape2 field under Worms and Dragons: Dragons.
 */
void
compute_escape_potential(void)
{
  int m, n;
  int i, j;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      if (p[m][n]
	  && dragon[m][n].origini == m
	  && dragon[m][n].originj == n) {
	
	/* For each distinct dragon. */
	memset(route_used, 0, sizeof(route_used));
	escape_routes = 0;

	/* Match each stone of the dragon with the patterns in escape.db. */
	for (i=0; i<board_size; i++)
	  for (j=0; j<board_size; j++)
	    if (dragon[i][j].origini == m
		&& dragon[i][j].originj == n)
	      matchpat(i, j, escape_callback, p[i][j], &escpat_db, NULL, NULL);

	/* Store the number of found escape routes in the dragon data. */
	dragon[m][n].escape2 = escape_routes;
      }

  /* FIXME: When escape2 field seems reasonably accurate, recompute dragon
   *        status here.
   *
   * We don't trust this code yet.
   */
#if 0
  /* Determine status: ALIVE, DEAD, CRITICAL or UNKNOWN */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if ((dragon[m][n].origini==m)
	  && (dragon[m][n].originj==n) && p[m][n]) {
	/* entirely ad-hoc : aim for -ve value to mean we need
	 * to worry about this dragon
	 * 2.5.38: Added effective_size to heuristics. /gf
	 */
	dragon[m][n].vitality = 6 * dragon[m][n].genus 
	  + dragon[m][n].heyes + 5 * dragon[m][n].escape2
	  + (dragon[m][n].effective_size - dragon[m][n].size)/5
	  + 3 * (dragon[m][n].lunchi != -1)
	  - 10 - 5 * (dragon[m][n].safety == CRITICAL);
	dragon[m][n].status=compute_dragon_status2(m, n);
	sgf_dragon_status(m,n,dragon[m][n].status);
      }
    }
#endif
  
  /* The dragon data is now correct at the origin of each dragon but
   * we need to copy it to every vertex.  
   */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      struct dragon_data *d = &(dragon[m][n]);
      
      dragon[m][n] = dragon[d->origini][d->originj];
    }
}

static void
escape_callback(int m, int n, int color, struct pattern *pattern, int ll,
		void *data)
{
  int ti, tj;
  int k;
  UNUSED(data);

  TRANSFORM(pattern->movei, pattern->movej, &ti, &tj, ll);
  ti += m;
  tj += n;

  /* If the pattern has a constraint, call the autohelper to see
   * if the pattern must be rejected.
   */
  if (pattern->autohelper_flag & HAVE_CONSTRAINT) {
    if (!pattern->autohelper(pattern, ll, ti, tj, color, 0))
      return;
  }

  /* If the pattern has a helper, call it to see if the pattern must
   * be rejected.
   */
  if (pattern->helper) {
    if (!pattern->helper(pattern, ll, ti, tj, color)) {
      DEBUG(DEBUG_ESCAPE, "Escape pattern %s+%d rejected by helper at %m\n",
	    pattern->name, ll, ti, tj);
      return;
    }
  }
  
  /* Loop through pattern elements to see if the escape route has
   * already been used.
   */
  for (k = 0; k < pattern->patlen; ++k) { /* match each point */
    if (pattern->patn[k].att == ATT_comma) {
      /* transform pattern real coordinate */
      int x, y;
      TRANSFORM(pattern->patn[k].x,pattern->patn[k].y,&x,&y,ll);
      x += m;
      y += n;

      if (route_used[x][y]) {
	return;
      }
    }
    else
      break; /* All commas are guaranteed to come first. */
  }

  /* Route accepted. Increase the number of routes and mark this one
   * as used.
   */
  escape_routes += pattern->value;

  DEBUG(DEBUG_ESCAPE, "Escape pattern %s+%d, value %d, found at %m\n",
	pattern->name, ll, (int) pattern->value, ti, tj);

  for (k = 0; k < pattern->patlen; ++k) { /* match each point */
    if (pattern->patn[k].att == ATT_comma) {
      /* transform pattern real coordinate */
      int x, y;
      TRANSFORM(pattern->patn[k].x,pattern->patn[k].y,&x,&y,ll);
      x += m;
      y += n;

      route_used[x][y] = 1;
    }
    else
      break; /* All commas are guaranteed to come first. */
  }
}


/* ================================================================ */
/*                       A few status functions                     */
/* ================================================================ */

/*
 * These functions are only here because then we don't need to expose
 * the dragon structure to the external program.
 */

int
dragon_status(int i, int j)
{
  return dragon[i][j].status;
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
