/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "liberty.h"
#include "influence.h"
#include "patterns.h"

static void segment_influence(struct influence_data *q);
static void print_influence(struct influence_data *q);

/* Influence computed for the initial position, i.e. before making
 * some move.
 */
static struct influence_data initial_influence;

/* Influence computed after some move has been made. */
static struct influence_data move_influence;

/* Coordinates for the move influence was last computed for. */
static int influence_movei = -1;
static int influence_movej = -1;
static int influence_color = EMPTY;

/* Cache of delta_territory_values. */
static int delta_territory_cache[MAX_BOARD][MAX_BOARD];

/* If set, print influence map when computing this move. Purely for
 * debugging.
 */
static int debug_influence_i = -1;
static int debug_influence_j = -1;

/* This is the core of the influence function. Given the coordinates
 * and color of an influence source, it radiates the influence
 * outwards until it hits a barrier or the strength of the influence
 * falls under a certain threshold.
 *
 * The radiation is performed by a breadth first propagation,
 * implemented by means of an internal queue.
 *
 * Since this function has turned out be one of the bottlenecks, loop
 * unrolling makes a noticeable performance difference. It does,
 * however, make the code much harder to read and maintain. Therefore
 * we include both the original and the unrolled versions.
 */

#define EXPLICIT_LOOP_UNROLLING 1

#if !EXPLICIT_LOOP_UNROLLING

static void
accumulate_influence(struct influence_data *q, int m, int n, int color)
{
  int i, j;
  int k;
  int d;
  float damping;
  float current_strength;
  float cos2phi;
  float permeability;
  float contribution;

  /* Clear the queue. */
  int queue_start = 0;
  int queue_end = 0;

  assert((q->color[m][n] & color) != 0);
  
  if (0)
    gprintf("Accumulating influence for %s at %m\n",
	    color_to_string(color), m, n);
  
  /* Put the influence origin on the stack. */
  q->w[m][n] = q->strength[m][n];
  q->queuei[queue_end] = m;
  q->queuej[queue_end] = n;
  queue_end++;

  /* Spread influence until the stack is empty. */
  while (queue_start < queue_end) {

    /* Pick first element in queue. */
    i = q->queuei[queue_start];
    j = q->queuej[queue_start];
    if (0)
      gprintf("Picked %m from queue. w=%f start=%d end=%d\n",
	      i, j, q->w[i][j], queue_start, queue_end);
    current_strength = q->w[i][j];
    queue_start++;

    /* Try to spread influence in each of the eight directions. */    
    for (d=0; d<8; d++) {
      int di = deltai[d];
      int dj = deltaj[d];

      /* Verify that (i+di, j+dj) is
       * 1. Inside the board.
       * 2. Not occupied.
       * 3. Directed outwards. For the origin all directions are outwards.
       */
      if (ON_BOARD(i+di, j+dj)
	  && q->p[i+di][j+dj] == EMPTY
	  && (di*(i-m) + dj*(j-n) > 0
	      || queue_start == 1)) {

	/* Now compute the damping of the influence.
	 * First we have the permeability at the point we are
	 * spreading from. For diagonal movement we also take the
	 * permeability of the vertices we are "passing by" into
	 * account.
	 */
	if (color == WHITE) {
	  permeability = q->white_permeability[i][j];
	  if (d > 3) /* diagonal movement */
	    permeability *= max(q->white_permeability[i+di][j],
				q->white_permeability[i][j+dj]);
	}
	else {
	  permeability = q->black_permeability[i][j];
	  if (d > 3) /* diagonal movement */
	    permeability *= max(q->black_permeability[i+di][j],
				q->black_permeability[i][j+dj]);
	}

	if (permeability == 0.0)
	  continue;
	
	/* Then we have the distance dependent damping. */
	damping = 1.0;
	switch (d) {
	case 0: /* south */
	  damping = q->south_attenuation[m][n];
	  break;
	case 1: /* west */
	  damping = q->west_attenuation[m][n];
	  break;
	case 2: /* north */
	  damping = q->north_attenuation[m][n];
	  break;
	case 3: /* east */
	  damping = q->east_attenuation[m][n];
	  break;
	case 4: /* southwest */
	  damping = (max(q->south_attenuation[m][n], q->west_attenuation[m][n])
		     * DIAGONAL_DAMPING);
	  break;
	case 5: /* northwest */
	  damping = (max(q->north_attenuation[m][n], q->west_attenuation[m][n])
		     * DIAGONAL_DAMPING);
	  break;
	case 6: /* northeast */
	  damping = (max(q->north_attenuation[m][n], q->east_attenuation[m][n])
		     * DIAGONAL_DAMPING);
	  break;
	case 7: /* southeast */
	  damping = (max(q->south_attenuation[m][n], q->east_attenuation[m][n])
		     * DIAGONAL_DAMPING);
	  break;
	}
	
	/* Finally direction dependent damping. */
	if (i==m && j==n)
	  cos2phi = 1.0;
	else {
	  float a = di*(i-m) + dj*(j-n);
	  float b = (i-m)*(i-m) + (j-n)*(j-n);
	  float c = di*di + dj*dj;
	  assert(a > 0.0);
	  cos2phi = (a*a) / (b*c);
	}

	contribution = cos2phi * permeability * current_strength / damping;

	/* Stop spreading influence if the contribution becomes too low. */
	if (contribution <= INFLUENCE_CUTOFF)
	  continue;
	
	/* If no influence here before, add the point to the queue for
	 * further spreading.
	 */
	if (0)
	  gprintf("  Spreading %s influence from %m to %m, d=%d\n",
		  color_to_string(color), i, j, i+di, j+dj, d);
	if (q->w[i+di][j+dj] == 0.0) {
	  q->queuei[queue_end] = i + di;
	  q->queuej[queue_end] = j + dj;
	  queue_end++;
	}
	
	q->w[i+di][j+dj] += contribution;
      }
    }
  }
  
  /* Add the values in the working area to the accumulated influence
   * and simultaneously reset the working area. We know that all
   * influenced points were stored in the queue, so we just traverse
   * it.
   */
  for (k=0; k<queue_end; k++) {
    i = q->queuei[k];
    j = q->queuej[k];

    if (color == WHITE)
      q->white_influence[i][j] += q->w[i][j];
    else
      q->black_influence[i][j] += q->w[i][j];
    
    q->w[i][j] = 0.0;
  }
}

#else /* EXPLICIT_LOOP_UNROLLING */

static void
accumulate_influence(struct influence_data *q, int m, int n, int color)
{
  int i, j;
  int k;
  float damping;
  float current_strength;
  float cos2phi;
  float permeability;
  float contribution;

  /* Clear the queue. */
  int queue_start = 0;
  int queue_end = 0;

  assert((q->color[m][n] & color) != 0);
  
  if (0)
    gprintf("Accumulating influence for %s at %m\n",
	    color_to_string(color), m, n);
  
  /* Put the influence origin on the stack. */
  q->w[m][n] = q->strength[m][n];
  q->queuei[queue_end] = m;
  q->queuej[queue_end] = n;
  queue_end++;

  /* Spread influence until the stack is empty. */
  while (queue_start < queue_end) {
    float b_inv;
    
    /* Pick first element in queue. */
    i = q->queuei[queue_start];
    j = q->queuej[queue_start];
    if (0)
      gprintf("Picked %m from queue. w=%f start=%d end=%d\n",
	      i, j, q->w[i][j], queue_start, queue_end);
    current_strength = q->w[i][j];
    queue_start++;

    if (i==m && j==n)
      b_inv = 1.0;
    else 
      b_inv = 1.0 / ((i-m)*(i-m) + (j-n)*(j-n));
    
    if (color == WHITE)
      permeability = q->white_permeability[i][j];
    else
      permeability = q->black_permeability[i][j];
    
    if (permeability > 0.0)
    {
      /* Try to spread influence south. */    
      if (i < board_size-1
	  && q->p[i+1][j] == EMPTY
	  && (i-m > 0 || queue_start == 1)) {

	damping = q->south_attenuation[m][n];
	
	if (i==m && j==n)
	  cos2phi = 1.0;
	else {
	  float a = i-m;
	  cos2phi = a * a * b_inv;
	}

	contribution = cos2phi * permeability * current_strength / damping;

	if (contribution > INFLUENCE_CUTOFF) {
	  if (0)
	    gprintf("  Spreading %s influence from %m to %m, d=%d\n",
		    color_to_string(color), i, j, i+1, j, 0);
	  if (q->w[i+1][j] == 0.0) {
	    q->queuei[queue_end] = i+1;
	    q->queuej[queue_end] = j;
	    queue_end++;
	  }
	
	  q->w[i+1][j] += contribution;
	}
      }
    
      /* Try to spread influence west. */    
      if (j > 0
	  && q->p[i][j-1] == EMPTY
	  && (j-n < 0 || queue_start == 1)) {
	
	damping = q->west_attenuation[m][n];
	
	if (i==m && j==n)
	  cos2phi = 1.0;
	else {
	  float a = n-j;
	  cos2phi = a * a * b_inv;
	}

	contribution = cos2phi * permeability * current_strength / damping;

	if (contribution > INFLUENCE_CUTOFF) {
	  if (0)
	    gprintf("  Spreading %s influence from %m to %m, d=%d\n",
		    color_to_string(color), i, j, i, j-1, 1);
	  if (q->w[i][j-1] == 0.0) {
	    q->queuei[queue_end] = i;
	    q->queuej[queue_end] = j-1;
	    queue_end++;
	  }
	  
	  q->w[i][j-1] += contribution;
	}
      }
      
      /* Try to spread influence north. */
      if (i > 0
	  && q->p[i-1][j] == EMPTY
	  && (i-m < 0 || queue_start == 1)) {
	
	damping = q->north_attenuation[m][n];
	
	if (i==m && j==n)
	  cos2phi = 1.0;
	else {
	  float a = m-i;
	  cos2phi = a * a * b_inv;
	}

	contribution = cos2phi * permeability * current_strength / damping;

	if (contribution > INFLUENCE_CUTOFF) {
	  if (0)
	    gprintf("  Spreading %s influence from %m to %m, d=%d\n",
		    color_to_string(color), i, j, i-1, j, 2);
	  if (q->w[i-1][j] == 0.0) {
	    q->queuei[queue_end] = i-1;
	    q->queuej[queue_end] = j;
	    queue_end++;
	  }
	}
	q->w[i-1][j] += contribution;
      }
      
      /* Try to spread influence east. */    
      if (j < board_size-1
	  && q->p[i][j+1] == EMPTY
	  && (j-n > 0 || queue_start == 1)) {

	damping = q->east_attenuation[m][n];
	
	if (i==m && j==n)
	  cos2phi = 1.0;
	else {
	  float a = j-n;
	  cos2phi = a * a * b_inv;
	}

	contribution = cos2phi * permeability * current_strength / damping;

	if (contribution > INFLUENCE_CUTOFF) {
	  if (0)
	    gprintf("  Spreading %s influence from %m to %m, d=%d\n",
		    color_to_string(color), i, j, i, j+1, 3);
	  if (q->w[i][j+1] == 0.0) {
	    q->queuei[queue_end] = i;
	    q->queuej[queue_end] = j+1;
	    queue_end++;
	  }
	  
	  q->w[i][j+1] += contribution;
	}
      }
    }
    
    /* Try to spread influence southwest. */    
    if (i < board_size-1 && j > 0
	&& q->p[i+1][j-1] == EMPTY
	&& (i-m > j-n
	    || queue_start == 1)) {
      
      if (color == WHITE) {
	permeability = q->white_permeability[i][j];
	permeability *= max(q->white_permeability[i+1][j],
			    q->white_permeability[i][j-1]);
      }
      else {
	permeability = q->black_permeability[i][j];
	permeability *= max(q->black_permeability[i+1][j],
			    q->black_permeability[i][j-1]);
      }

      if (permeability > 0.0) {
	damping = (max(q->south_attenuation[m][n], q->west_attenuation[m][n])
		   * DIAGONAL_DAMPING);
	
	if (i==m && j==n)
	  cos2phi = 1.0;
	else {
	  float a = (i-m) - (j-n);
	  cos2phi = 0.5 * a * a * b_inv;
	}

	contribution = cos2phi * permeability * current_strength / damping;

	if (contribution > INFLUENCE_CUTOFF) {
	  if (0)
	    gprintf("  Spreading %s influence from %m to %m, d=%d\n",
		    color_to_string(color), i, j, i+1, j-1, 4);
	  if (q->w[i+1][j-1] == 0.0) {
	    q->queuei[queue_end] = i+1;
	    q->queuej[queue_end] = j-1;
	    queue_end++;
	  }
	  
	  q->w[i+1][j-1] += contribution;
	}
      }
    }
    
    /* Try to spread influence northwest. */    
    if (i > 0 && j > 0
	&& q->p[i-1][j-1] == EMPTY
	&& ((i-m) + (j-n) < 0
	    || queue_start == 1)) {

      if (color == WHITE) {
	permeability = q->white_permeability[i][j];
	permeability *= max(q->white_permeability[i-1][j],
			    q->white_permeability[i][j-1]);
      }
      else {
	permeability = q->black_permeability[i][j];
	permeability *= max(q->black_permeability[i-1][j],
			    q->black_permeability[i][j-1]);
      }
      
      if (permeability > 0.0) {
	damping = (max(q->north_attenuation[m][n], q->west_attenuation[m][n])
		   * DIAGONAL_DAMPING);
	
	if (i==m && j==n)
	  cos2phi = 1.0;
	else {
	  float a = -(i-m) - (j-n);
	  cos2phi = 0.5 * a * a * b_inv;
	}

	contribution = cos2phi * permeability * current_strength / damping;

	if (contribution > INFLUENCE_CUTOFF) {
	  if (0)
	    gprintf("  Spreading %s influence from %m to %m, d=%d\n",
		    color_to_string(color), i, j, i-1, j-1, 5);
	  if (q->w[i-1][j-1] == 0.0) {
	    q->queuei[queue_end] = i-1;
	    q->queuej[queue_end] = j-1;
	    queue_end++;
	  }
	  
	  q->w[i-1][j-1] += contribution;
	}
      }
    }
    
    /* Try to spread influence northeast. */
    if (i > 0 && j < board_size-1
	&& q->p[i-1][j+1] == EMPTY
	&& (j-n > i-m
	    || queue_start == 1)) {

      if (color == WHITE) {
	permeability = q->white_permeability[i][j];
	permeability *= max(q->white_permeability[i-1][j],
			    q->white_permeability[i][j+1]);
      }
      else {
	permeability = q->black_permeability[i][j];
	permeability *= max(q->black_permeability[i-1][j],
			    q->black_permeability[i][j+1]);
      }

      if (permeability > 0.0) {
	damping = (max(q->north_attenuation[m][n], q->east_attenuation[m][n])
		   * DIAGONAL_DAMPING);
	
	if (i==m && j==n)
	  cos2phi = 1.0;
	else {
	  float a = -(i-m) + (j-n);
	  cos2phi = 0.5 * a * a * b_inv;
	}

	contribution = cos2phi * permeability * current_strength / damping;

	if (contribution > INFLUENCE_CUTOFF) {
	  if (0)
	    gprintf("  Spreading %s influence from %m to %m, d=%d\n",
		    color_to_string(color), i, j, i-1, j+1, 6);
	  if (q->w[i-1][j+1] == 0.0) {
	    q->queuei[queue_end] = i-1;
	    q->queuej[queue_end] = j+1;
	    queue_end++;
	  }
	  
	  q->w[i-1][j+1] += contribution;
	}
      }
    }
    
    /* Try to spread influence southeast. */    
    if (i < board_size-1 && j < board_size-1
	&& q->p[i+1][j+1] == EMPTY
	&& ((i-m) + (j-n) > 0
	    || queue_start == 1)) {
      
      if (color == WHITE) {
	permeability = q->white_permeability[i][j];
	permeability *= max(q->white_permeability[i+1][j],
			    q->white_permeability[i][j+1]);
      }
      else {
	permeability = q->black_permeability[i][j];
	permeability *= max(q->black_permeability[i+1][j],
			    q->black_permeability[i][j+1]);
      }

      if (permeability > 0.0) {
	damping = (max(q->south_attenuation[m][n], q->east_attenuation[m][n])
		   * DIAGONAL_DAMPING);
	
	if (i==m && j==n)
	  cos2phi = 1.0;
	else {
	  float a = (i-m) + (j-n);
	  cos2phi = 0.5 * a * a * b_inv;
	}

	contribution = cos2phi * permeability * current_strength / damping;

	if (contribution > INFLUENCE_CUTOFF) {
	  if (0)
	    gprintf("  Spreading %s influence from %m to %m, d=%d\n",
		    color_to_string(color), i, j, i+1, j+1, 7);
	  if (q->w[i+1][j+1] == 0.0) {
	    q->queuei[queue_end] = i+1;
	    q->queuej[queue_end] = j+1;
	    queue_end++;
	  }
	
	  q->w[i+1][j+1] += contribution;
	}
      }
    }
  }
  
  /* Add the values in the working area to the accumulated influence
   * and simultaneously reset the working area. We know that all
   * influenced points were stored in the queue, so we just traverse
   * it.
   */
  for (k=0; k<queue_end; k++) {
    i = q->queuei[k];
    j = q->queuej[k];

    if (color == WHITE)
      q->white_influence[i][j] += q->w[i][j];
    else
      q->black_influence[i][j] += q->w[i][j];
    
    q->w[i][j] = 0.0;
  }
}

#endif /* EXPLICIT_LOOP_UNROLLING */
    

/* Initialize the influence_data structure. If the dragons have been
 * computed, we weight the strength of the influence with the dragon
 * status.
 */
static void
init_influence(struct influence_data *q, int color, int dragons_known)
{
  int i, j;
  q->color_to_move = color;
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      q->white_influence[i][j] = 0.0;
      q->black_influence[i][j] = 0.0;
      q->w[i][j] = 0.0;
      q->south_attenuation[i][j] = DEFAULT_ATTENUATION;
      q->west_attenuation[i][j] = DEFAULT_ATTENUATION;
      q->north_attenuation[i][j] = DEFAULT_ATTENUATION;
      q->east_attenuation[i][j] = DEFAULT_ATTENUATION;
      q->white_permeability[i][j] = 1.0;
      q->black_permeability[i][j] = 1.0;
      q->p[i][j] = p[i][j];
      if (p[i][j] != EMPTY) {
	if (worm[i][j].attack_code == 1
	    && (q->p[i][j] != color || worm[i][j].defend_code == 0)) {
	  if (q->p[i][j] == WHITE)
	    q->white_permeability[i][j] = 0.0;
	  else
	    q->black_permeability[i][j] = 0.0;
	  q->p[i][j] = EMPTY;
	}
	else {
	  if (q->p[i][j] == WHITE)
	    q->black_permeability[i][j] = 0.0;
	  else
	    q->white_permeability[i][j] = 0.0;
	}
      }
      q->color[i][j] = q->p[i][j];
      if (q->p[i][j] == EMPTY)
	q->strength[i][j] = 0.0;
      else if (!dragons_known)
	q->strength[i][j] = DEFAULT_STRENGTH;
      else if (dragon[i][j].status == ALIVE)
	q->strength[i][j] = DEFAULT_STRENGTH;
      else if (dragon[i][j].status == CRITICAL)
	q->strength[i][j] = 0.9 * DEFAULT_STRENGTH;
      else if (dragon[i][j].safety == CRITICAL)
	q->strength[i][j] = 0.3 * DEFAULT_STRENGTH;
      else if (dragon[i][j].status == UNKNOWN) {
	q->strength[i][j] = (0.5 + min(1.0,
				       max(0.5,
					   0.5+dragon[i][j].vitality/20.0))
			     * DEFAULT_STRENGTH);
      }
      else if (dragon[i][j].status == DEAD)
	q->strength[i][j] = 0.0;
    }
}

/* Enhance the influence in a certain direction by decreasing the
 * attenuation in that direction. The direction is given implicitly by
 * the transformation number ll of a matched pattern.
 */
static void
enhance_influence(struct influence_data *q, int m, int n, int ll, float value)
{
  float *attenuation = NULL;
  /* The convention is that the pattern enhances the influence
   * eastwards, when it is oriented as in the database. Now we
   * need to transform this direction.
   */
  if (0)
    gprintf("Enhancing influence at %m, rotation %d value %f\n",
	    m, n, ll, value);
  switch (ll) {
  case 0:
  case 5:
    attenuation = &(q->east_attenuation[m][n]);
    break;
  case 2:
  case 7:
    attenuation = &(q->west_attenuation[m][n]);
    break;
  case 4:
  case 3:
    attenuation = &(q->north_attenuation[m][n]);
    break;
  case 6:
  case 1:
    attenuation = &(q->south_attenuation[m][n]);
    break;
  }
  *attenuation = min(*attenuation, DEFAULT_ATTENUATION - value/4.0);
}

/* Callback for the matched patterns in influence.db and barriers.db.
 * The pattern classes used here are:
 * A - Barrier pattern, where O plays first and X tries to block influence.
 * D - Barrier pattern, where O plays first and O tries to block influence.
 * B - Intrusion patterns, adding a low intensity influence source.
 * E - Enhance patterns, decreasing the attenuation in a given direction.
 * I - Invasion patterns, adding a low intensity influence source.
 *
 * Classes A, D, and B are matched with color as O, and it is assumed
 * that O is in turn to move. Classes E and I are matched with either
 * color as O.
 */
static void
influence_callback(int m, int n, int color, struct pattern *pattern, int ll,
		   void *data)
{
  int ti, tj;
  int k;
  struct influence_data *q = data;
  
  /* Loop through pattern elements to see if an A or D pattern
   * can possibly have any effect. If not we can skip evaluating
   * constraint and/or helper. */
  if (pattern->class & (CLASS_A | CLASS_D)) {
    int something_to_do = 0;
    for (k = 0; k < pattern->patlen; ++k) { /* match each point */
      int x, y;
      int blocking_color;
      if (pattern->patn[k].att != ATT_comma)
	break;  /* All commas are guaranteed to come first. */

      /* transform pattern real coordinate */
      TRANSFORM(pattern->patn[k].x, pattern->patn[k].y, &x, &y, ll);
      x += m;
      y += n;

      /* Territorial connection, making a barrier for opponent influence. */
      if (pattern->class & CLASS_D)
	blocking_color = color;
      else
	blocking_color = OTHER_COLOR(color);
      if ((blocking_color == WHITE
	   && q->black_permeability[x][y] != 0.0)
	  || (blocking_color == BLACK
	      && q->white_permeability[x][y] != 0.0)) {
	something_to_do = 1;
	break;
      }
    }
    if (!something_to_do)
      return;
  }

  /* FIXME: Integrate the following loop into the previous one. */
  
  /* Require that all O stones in the pattern are tactically safe for
   * territorial connection patterns of type D and X stones for type A.
   * Both colors must be tactically safe for patterns of class B.
   */
  if (pattern->class & (CLASS_D | CLASS_A | CLASS_B)) {
    for (k = 0; k < pattern->patlen; ++k) { /* match each point */
      if ((pattern->patn[k].att == ATT_O
	   && (pattern->class & (CLASS_D | CLASS_B)))
	  || (pattern->patn[k].att == ATT_X
	      && (pattern->class & (CLASS_A | CLASS_B)))) {
	/* transform pattern real coordinate */
	int x, y;
	TRANSFORM(pattern->patn[k].x,pattern->patn[k].y,&x,&y,ll);
	x += m;
	y += n;
	if (worm[x][y].attack_code != 0)
	  return; /* Match failed */
      }
    }
  }

  TRANSFORM(pattern->movei, pattern->movej, &ti, &tj, ll);
  ti += m;
  tj += n;

  /* If the pattern has a constraint, call the autohelper to see
   * if the pattern must be rejected.
   */
  if (pattern->autohelper_flag & HAVE_CONSTRAINT) {
    if (!pattern->autohelper(pattern, ll, ti, tj, color, 0))
      return;
  }

  /* If the pattern has a helper, call it to see if the pattern must
   * be rejected.
   */
  if (pattern->helper) {
    if (!pattern->helper(pattern, ll, ti, tj, color)) {
      DEBUG(DEBUG_INFLUENCE,
	    "Influence pattern %s+%d rejected by helper at %m\n",
	    pattern->name, ll, ti, tj);
      return;
    }
  }

  DEBUG(DEBUG_INFLUENCE, "influence pattern '%s'+%d matched at %m\n",
	pattern->name, ll, m, n);

  /* For I patterns, add a low intensity, both colored, influence
   * source at *.
   */
  if (pattern->class & CLASS_I) {
#if 0
    q->color[ti][tj] = BLACK | WHITE;
#else
    if (pattern->class & CLASS_s)
      q->color[ti][tj] = BLACK | WHITE;
    else if (q->color_to_move != color)
      q->color[ti][tj] |= q->color_to_move;
#endif
    q->strength[ti][tj] = pattern->value;
    q->south_attenuation[ti][tj] = 1.5;
    q->west_attenuation[ti][tj] = 1.5;
    q->north_attenuation[ti][tj] = 1.5;
    q->east_attenuation[ti][tj] = 1.5;
    DEBUG(DEBUG_INFLUENCE,
	  "  low intensity influence source at %m, strength %f\n",
	  ti, tj, pattern->value);
    return;
  }
  
  /* Loop through pattern elements and perform necessary actions
   * for A, D, B, and E patterns. */
  for (k = 0; k < pattern->patlen; ++k) { /* match each point */
    if ((   (pattern->class & CLASS_E)
	    && pattern->patn[k].att == ATT_O)
	|| ((pattern->class & (CLASS_D | CLASS_A))
	    && pattern->patn[k].att == ATT_comma)
	|| ((pattern->class & CLASS_B)
	    && pattern->patn[k].att == ATT_not)) {
      /* transform pattern real coordinate */
      int x, y;
      TRANSFORM(pattern->patn[k].x,pattern->patn[k].y,&x,&y,ll);
      x += m;
      y += n;

      /* Influence enhancing pattern. */
      if (pattern->class & CLASS_E) {
	DEBUG(DEBUG_INFLUENCE, "  influence enhanced for %m, value %d\n",
	      x, y, (int) pattern->value);
	enhance_influence(q, x, y, ll, pattern->value);
      }
     
      /* Territorial connection, making a barrier for opponent influence. */
      if (pattern->class & (CLASS_A | CLASS_D)) {
	int blocking_color;
	if (pattern->class & CLASS_D)
	  blocking_color = color;
	else
	  blocking_color = OTHER_COLOR(color);
	DEBUG(DEBUG_INFLUENCE, "  barrier for %s influence at %m\n",
	      color_to_string(OTHER_COLOR(blocking_color)), x, y);
	if (blocking_color == WHITE)
	  q->black_permeability[x][y] = 0.0;
	else
	  q->white_permeability[x][y] = 0.0;
      }
      
      /* Low intensity influence source for the color in turn to move. */
      if (pattern->class & CLASS_B) {
	DEBUG(DEBUG_INFLUENCE, "  intrusion at %m\n", x, y);
	q->color[x][y] |= color;
	q->strength[x][y] = pattern->value;
      }
    }
  }
}

/* Match the patterns in influence.db and barriers.db in order to
 * modify attenuation in certain directions, add influence barriers,
 * and add extra influence sources at possible invasion and intrusion
 * points.
 */
static void
find_influence_patterns(struct influence_data *q, int color)
{
  int m, n;
  for (m = 0; m < board_size; m++)
    for (n = 0; n < board_size; n++)
      if (p[m][n]) {
	matchpat(m, n, influence_callback, p[m][n], &influencepat_db, q, NULL);
	matchpat(m, n, influence_callback, color, &barrierspat_db, q, NULL);
      }

  /* It may happen that we have a low intensity influence source at a
   * blocked intersection (due to an intrusion). Reset the
   * permeability at this point.
   */
  for (m = 0; m < board_size; m++)
    for (n = 0; n < board_size; n++)
      if (q->p[m][n] == EMPTY && q->strength[m][n] > 0.0) {
	if ((q->color[m][n] & WHITE) && q->white_permeability[m][n] != 1.0) {
	  DEBUG(DEBUG_INFLUENCE, "  black block removed from %m\n", m, n);
	  q->white_permeability[m][n] = 1.0;
	}
	if ((q->color[m][n] & BLACK) && q->black_permeability[m][n] != 1.0) {
	  DEBUG(DEBUG_INFLUENCE, "  white block removed from %m\n", m, n);
	  q->black_permeability[m][n] = 1.0;
	}
      }
}

/* Compute the influence values for both colors, after having made a
 * move for OTHER_COLOR(color) at (m, n). If these coordinates are -1
 * no move is made. In any case it's assumed that color is in turn to
 * move.
 */
static void
compute_influence(struct influence_data *q, int color, int m, int n,
		  int dragons_known)
{
  int i, j;
  init_influence(q, color, dragons_known);
  if (m != -1)
    q->strength[m][n] = DEFAULT_STRENGTH;

  /* Reduce depth values to speed up evaluation of constraints. Notice
   * that the depth values are also modified by compute_move_influce()
   * to avoid horizon problems, so some care must be taken here.
   */
  set_temporary_depth_values(depth - 2, backfill_depth - 3,
			     fourlib_depth - 2, ko_depth - 2);
  find_influence_patterns(q, color);
  restore_depth_values();
  
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if (q->color[i][j] & WHITE)
	accumulate_influence(q, i, j, WHITE);
      if (q->color[i][j] & BLACK)
	accumulate_influence(q, i, j, BLACK);
    }
  segment_influence(q);
  if ((q == &initial_influence
       && (printmoyo & PRINTMOYO_INITIAL_INFLUENCE))
      || (m == debug_influence_i && n == debug_influence_j && m >= 0))
    print_influence(q);
}

/* Return the color of the territory at (m, n). If it's territory for
 * neither color, EMPTY is returned.
 * The definition of territory in terms of the influences is totally
 * ad hoc.
 */
static int
whose_territory(struct influence_data *q, int m, int n)
{
  float bi = q->black_influence[m][n];
  float wi = q->white_influence[m][n];

  if (bi > 0.0 && wi == 0.0)
    return BLACK;

  if (wi > 0.0 && bi == 0.0)
    return WHITE;

  return EMPTY;
}

/* Return the color who has a moyo at (m, n). If neither color has a
 * moyo there, EMPTY is returned.
 * The definition of moyo in terms of the influences is totally ad
 * hoc.
 */
static int
whose_moyo(struct influence_data *q, int m, int n)
{
  float bi = q->black_influence[m][n];
  float wi = q->white_influence[m][n];

  int territory_color = whose_territory(q, m, n);
  if (territory_color != EMPTY)
    return territory_color;
  
  if (bi > 5.0 * wi && bi > 5.0 && wi < 10.0)
    return BLACK;

  if (wi > 5.0 * bi && wi > 5.0 && bi < 10.0)
    return WHITE;
  
  return EMPTY;
}

/* Return the color who has dominating influence ("area") at (m, n).
 * If neither color dominates the influence there, EMPTY is returned.
 * The definition of area in terms of the influences is totally ad
 * hoc.
 */
static int
whose_area(struct influence_data *q, int m, int n)
{
  float bi = q->black_influence[m][n];
  float wi = q->white_influence[m][n];

  int moyo_color = whose_moyo(q, m, n);
  if (moyo_color != EMPTY)
    return moyo_color;
  
  if (bi > 2.0 * wi && bi > 1.0 && wi < 40.0)
    return BLACK;

  if (wi > 2.0 * bi && wi > 1.0 && bi < 40.0)
    return WHITE;
  
  return EMPTY;
}

/* Segment the influence map into connected regions of territory,
 * moyo, or area. What to segment on is determined by the the function
 * pointer region_owner. The segmentation is performed for both
 * colors. The connected regions may include stones of the own color,
 * but only empty intersections (and dead opponent stones) count
 * toward the region size.
 */
static void
segment_region(struct influence_data *q, owner_function_ptr region_owner,
	       int type, int (*segmentation)[MAX_BOARD])
{
  int m, n;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if (q->w[m][n] == UNMARKED && region_owner(q, m, n) != EMPTY) {
	/* Found an unlabelled intersection. Use flood filling to find
         * the rest of the region.
	 */
	int size = 0;
	int queue_start = 0;
	int queue_end = 1;
	int color = region_owner(q, m, n);
	q->number_of_regions++;
	q->w[m][n] = MARKED;
	q->queuei[0] = m;
	q->queuej[0] = n;
	while (queue_start < queue_end) {
	  int i = q->queuei[queue_start];
	  int j = q->queuej[queue_start];
	  int k;
	  queue_start++;
	  if (q->p[i][j] != color)
	    size++;
	  segmentation[i][j] = q->number_of_regions;
	  for (k=0; k<4; k++) {
	    int di = deltai[k];
	    int dj = deltaj[k];
	    if (ON_BOARD(i+di, j+dj)
		&& q->w[i+di][j+dj] == UNMARKED
		&& region_owner(q, i+di, j+dj) == color) {
	      q->queuei[queue_end] = i+di;
	      q->queuej[queue_end] = j+dj;
	      queue_end++;
	      q->w[i+di][j+dj] = MARKED;
	    }
	  }
	}
	if (color == WHITE)
	  q->region_type[q->number_of_regions] = WHITE_REGION | type;
	else
	  q->region_type[q->number_of_regions] = BLACK_REGION | type;
	q->region_size[q->number_of_regions] = size;
	if (0)
	  gprintf("Region %d of type %d (color %s) at %m. Size %d\n",
		  q->number_of_regions, q->region_type[q->number_of_regions],
		  color_to_string(color), m, n, size);
      }
    }
  /* Reset the working area w. */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      q->w[m][n] = UNMARKED;
}

/* Segment the influence map into connected regions of territory,
 * moyo, and area for both colors. The region numbers are stored in
 * the territory_segmentation, moyo_segmentation, and
 * area_segmentation arrays of the influence_data struct respectively.
 * All types of regions use the same numbering, with zero as a dummy
 * for nonregions. The region_type and region_size arrays hold
 * information about each region.
 */
static void
segment_influence(struct influence_data *q)
{
  int i, j;
  q->number_of_regions = 0;
  q->region_type[0] = 0;
  q->region_size[0] = 0;
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      q->territory_segmentation[i][j] = 0;
      q->moyo_segmentation[i][j] = 0;
      q->area_segmentation[i][j] = 0;
    }
  segment_region(q, whose_territory, IS_TERRITORY, q->territory_segmentation);
  segment_region(q, whose_moyo,      IS_MOYO,      q->moyo_segmentation);
  segment_region(q, whose_area,      IS_AREA,      q->area_segmentation);
}

/* Compute the influence before a move has been made, which can
 * later be compared to the influence after a move. Assume that
 * the other color is in turn to move.
 */
void
compute_initial_influence(int color, int dragons_known)
{
  int i, j;
  compute_influence(&initial_influence, OTHER_COLOR(color), -1, -1,
		    dragons_known);
  /* Invalidate information in move_influence. */
  influence_movei = -1;
  influence_movej = -1;
  influence_color = EMPTY;
  /* Clear delta_territory cache. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      delta_territory_cache[i][j] = NOT_COMPUTED;
}

/* Let color play at (m, n) and compute the influence after this move,
 * assuming that the other color is in turn to move next.
 */
static void
compute_move_influence(int m, int n, int color)
{
  /* Don't recompute if we already have the current values stored. */
  if (influence_movei == m
      && influence_movej == n
      && influence_color == color)
    return;

  if (tryko(m, n, color, "compute_move_influence")) {
    increase_depth_values();
    compute_influence(&move_influence, OTHER_COLOR(color), m, n, 1);
    decrease_depth_values();
    popgo();
  }
  else {
    gprintf("Computing influence for illegal move %m\n", m, n);
    return;
  }

  influence_movei = m;
  influence_movej = n;
  influence_color = color;
}

/* Count the amount of territory for color. */
static int
sum_territory(struct influence_data *q, int color)
{
  int i, j;
  int territory = 0;
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if ((q->p[i][j] == EMPTY || q->strength[i][j] == 0)
	  && whose_territory(q, i, j) == color)
	territory++;
  return territory;
}

/* Count the amount of moyo for color. */
static int
sum_moyo(struct influence_data *q, int color)
{
  int i, j;
  int moyo = 0;
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if ((q->p[i][j] == EMPTY || q->strength[i][j] == 0)
	  && whose_moyo(q, i, j) == color)
	moyo++;
  return moyo;
}

/* Count the amount of moyo, that is not also territory, for color. */
static int
sum_strict_moyo(struct influence_data *q, int color)
{
  int i, j;
  int moyo = 0;
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if ((q->p[i][j] == EMPTY || q->strength[i][j] == 0)
	  && whose_moyo(q, i, j) == color
	  && whose_territory(q, i, j) != color)
	moyo++;
  return moyo;
}

/* Count the amount of area for color. */
static int
sum_area(struct influence_data *q, int color)
{
  int i, j;
  int area = 0;
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if ((q->p[i][j] == EMPTY || q->strength[i][j] == 0)
	  && whose_area(q, i, j) == color)
	area++;
  return area;
}

/* Count the amount of area, that is not also moyo, for color. */
static int
sum_strict_area(struct influence_data *q, int color)
{
  int i, j;
  int area = 0;
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if ((q->p[i][j] == EMPTY || q->strength[i][j] == 0)
	  && whose_area(q, i, j) == color
	  && whose_moyo(q, i, j) != color)
	area++;
  return area;
}

/* Return the color who has territory at (m, n), or EMPTY. */
int
influence_territory_color(int m, int n)
{
  return whose_territory(&initial_influence, m, n);
}

/* Return the color who has moyo at (m, n), or EMPTY. */
int
influence_moyo_color(int m, int n)
{
  return whose_moyo(&initial_influence, m, n);
}

/* Return the color who has area at (m, n), or EMPTY. */
int
influence_area_color(int m, int n)
{
  return whose_area(&initial_influence, m, n);
}

/* Compute the difference in territory made by a move by color at (m, n). */
int
influence_delta_territory(int m, int n, int color)
{
  int delta;
  if (delta_territory_cache[m][n] != NOT_COMPUTED)
    return delta_territory_cache[m][n];
  if (0)
    gprintf("influence_delta_territory for %m %s = ", m, n,
	    color_to_string(color));
  compute_move_influence(m, n, color);
  delta = (sum_territory(&move_influence, color) -
	   sum_territory(&initial_influence, color) -
	   sum_territory(&move_influence, OTHER_COLOR(color)) +
	   sum_territory(&initial_influence, OTHER_COLOR(color)));
  if (0)
    gprintf("%d\n", delta);
  delta_territory_cache[m][n] = delta;
  return delta;
}

/* Compute the difference in moyo made by a move by color at (m, n). */
int
influence_delta_moyo(int m, int n, int color)
{
  int delta;
  if (0)
    gprintf("influence_delta_moyo for %m %s = ", m, n,
	    color_to_string(color));
  compute_move_influence(m, n, color);
  delta = (sum_moyo(&move_influence, color) -
	   sum_moyo(&initial_influence, color) -
	   sum_moyo(&move_influence, OTHER_COLOR(color)) +
	   sum_moyo(&initial_influence, OTHER_COLOR(color)));
  if (0)
    gprintf("%d\n", delta);
  return delta;
}

/* Compute the difference in strict moyo made by a move by color at (m, n). */
int
influence_delta_strict_moyo(int m, int n, int color)
{
  int delta;
  if (0)
    gprintf("influence_delta_strict_moyo for %m %s = ", m, n,
	    color_to_string(color));
  compute_move_influence(m, n, color);
  delta = (sum_strict_moyo(&move_influence, color) -
	   sum_strict_moyo(&initial_influence, color) -
	   sum_strict_moyo(&move_influence, OTHER_COLOR(color)) +
	   sum_strict_moyo(&initial_influence, OTHER_COLOR(color)));
  if (0)
    gprintf("%d\n", delta);
  return delta;
}

/* Compute the difference in area made by a move by color at (m, n). */
int
influence_delta_area(int m, int n, int color)
{
  int delta;
  if (0)
    gprintf("influence_delta_area for %m %s = ", m, n,
	    color_to_string(color));
  compute_move_influence(m, n, color);
  delta = (sum_area(&move_influence, color) -
	   sum_area(&initial_influence, color) -
	   sum_area(&move_influence, OTHER_COLOR(color)) +
	   sum_area(&initial_influence, OTHER_COLOR(color)));
  if (0)
    gprintf("%d\n", delta);
  return delta;
}

/* Compute the difference in strict area made by a move by color at (m, n). */
int
influence_delta_strict_area(int m, int n, int color)
{
  int delta;
  if (0)
    gprintf("influence_delta_strict_area for %m %s = ", m, n,
	    color_to_string(color));
  compute_move_influence(m, n, color);
  delta = (sum_strict_area(&move_influence, color) -
	   sum_strict_area(&initial_influence, color) -
	   sum_strict_area(&move_influence, OTHER_COLOR(color)) +
	   sum_strict_area(&initial_influence, OTHER_COLOR(color)));
  if (0)
    gprintf("%d\n", delta);
  return delta;
}

/* Print the influence map when we have computed influence for the
 * move at (i, j).
 */
void
debug_influence_move(int i, int j)
{
  debug_influence_i = i;
  debug_influence_j = j;
}

/* Print influence for debugging purposes. */
static void
print_influence(struct influence_data *q)
{
  int i, j;
  if (printmoyo & PRINTMOYO_NUMERIC_INFLUENCE) {
    /* Print the white influence values. */
    for (i=0; i<board_size; i++) {
      for (j=0; j<board_size; j++) {
	if (q->p[i][j] == WHITE)
	  fprintf(stderr, "  O");
	else if (q->p[i][j] == BLACK)
	  fprintf(stderr, "  X");
	else
	  fprintf(stderr, "%3d", (int) q->white_influence[i][j]);
      }
      fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");
    
    /* Print the black influence values. */
    for (i=0; i<board_size; i++) {
      for (j=0; j<board_size; j++) {
	if (q->p[i][j] == WHITE)
	  fprintf(stderr, "  O");
	else if (q->p[i][j] == BLACK)
	  fprintf(stderr, "  X");
	else
	  fprintf(stderr, "%3d", (int) q->black_influence[i][j]);
      }
      fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");
  }

  if (printmoyo & PRINTMOYO_PRINT_INFLUENCE) {
    /* Draw colored board illustrating territory, moyo, and area. */
    start_draw_board();
    for (i=0; i<board_size; i++)
      for (j=0; j<board_size; j++) {
	int c = EMPTY;
	int color = COLOR_BLACK;
	if (q->p[i][j] == WHITE) {
	  c = 'O';
	  color = COLOR_BLACK;
	}
	else if (q->p[i][j] == BLACK) {
	  c = 'X';
	  color = COLOR_BLACK;
	}
	else if (whose_territory(q, i, j) == WHITE) {
	  c = 'o';
	  color = COLOR_CYAN;
	}
	else if (whose_territory(q, i, j) == BLACK) {
	  c = 'x';
	  color = COLOR_CYAN;
	}
	else if (whose_moyo(q, i, j) == WHITE) {
	  c = 'o';
	  color = COLOR_YELLOW;
	}
	else if (whose_moyo(q, i, j) == BLACK) {
	  c = 'x';
	  color = COLOR_YELLOW;
	}
	else if (whose_area(q, i, j) == WHITE) {
	  c = 'o';
	  color = COLOR_RED;
	}
	else if (whose_area(q, i, j) == BLACK) {
	  c = 'x';
	  color = COLOR_RED;
	}
	draw_color_char(i, j, c, color);
      }
    end_draw_board();
    
  }
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
