/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/*
 * This file is about life & death.
 * 
 * For now, we only consider eye spaces.  Code for extending and
 * shrinking eye spaces is in mouse.c. 
 *
 * We will only handle small eye spaces for now (<= 10 during testing).
 * At these small sizes, we make the simplifying assumption that
 * if any part of the surrounding wall is captured, the resulting
 * eye space will lead to 0 eyes.  This assumtion can be dealt with
 * later when the basic code works.
 *
 * Strategy:
 * Try to fill the entire eye space until there are only eyes of
 * size 1 left. Then nobody can play there, and we can count the eyes.
 */

/* FIXME: TODO
 *  - Handle ko
 *  - Handle seki
 *  - Try to find out how many outside liberties are needed for the optimal
 *    sequence.
 * - Try to find the optimal move. For example, consider this situation:
 *
 *      D E F G H J
 *      . . . X X X 9
 *      . . O X . O 8
 *      . O . X O . 7
 *      + O . X . O 6
 *      . O . X . X 5
 *      . . O X X . 4
 *      . . O O O X 3
 *      . . O . O . 2
 *      . . . . . . 1
 *
 *   In this situation, the current implementation suggests an X move at
 *   H8 as the defense move.  This works, but H6 is more efficient since 
 *   when the external liberties are filled in, we need to make at least
 *   one extra move if O plays H6.
 * - There is no need for all levels of the search to return attacking
 *   or defending moves. This is only needed at the top level. In fact,
 *   there is no need to store it into the hash table either, so the table
 *   could be halved in size (short instead of int).  However, I don't
 *   want to do this simplification yet.  Maybe handling of seki or ko
 *   will have to use some scheme like this.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "liberty.h"
#include "location_set.h"


#define DEBUG_LIMIT   4

#define MAX_EYE_SIZE  10


static int false_marginal_eye(int m, int  n, 
			      int *attacki, int *attackj, 
			      int *defendi, int *defendj,
			      int *max, int *min, 
			      struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
			      int add_moves, int color);
static int minimize_eyes(int color, int *min, int *attacki, int *attackj, 
			 struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
			 Location_set *eye,
			 Location_set *diag_defends, 
			 Location_set *diag_attacks,
			 Location_set *moves);
static int maximize_eyes(int color, int *max, int *defendi, int *defendj, 
			 struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
			 Location_set *eye,
			 Location_set *diag_defends, 
			 Location_set *diag_attacks,
			 Location_set *moves);
static void get_boundary(Location_set *boundary, 
			 struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
			 Location_set *eyepoints, int color);
static void life_showboard(void);

/* Used by life_showboard(). */
static int stackp_when_called = 0;


/* ================================================================ */
/*                   Hash table for recognize_eye.                  */
/* ================================================================ */


/*
 * Layout of the hash table:
 *
 *   Each entry is one 32 bit integer. (FIXME: Don't waste 32 bits on 
 *   64 bit machines.) 
 *
 *   byte 0 (LSB):  Result byte (see below)
 *   byte 1:        Where the attacker should play (index in the set)
 *   byte 2:        Where the attacker should play.
 *   byte 3 (MSB):  Reserved
 *
 *   The result byte:
 *   +-------------------------------+
 *   | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
 *   +-------------------------------+
 *   |vis|   result  |vis|   result  |
 *   +-------------------------------+
 *   |     attack    |    defense    |
 *   +-------------------------------+
 *
 *   The byte contains the same info twice, once for the attacker, 
 *   and once for the defender (the eye owner).
 *
 *   vis: This node is visited by the search. When the result is 0,
 *        we are still searching, otherwise the result can be used
 *        as below. This bit is used to break cycles in the search.
 *   result:
 *     0:   No result known yet.  This is used during the search.
 *     1:   0 eyes
 *     2:   1 eye
 *     3:   2 eyes
 *     4-7: Reserved
 *     Ideas:
 *     4:   seki
 *     5:   ko 0-1 eyes
 *     6:   ko 1-2 eyes
 *     7:   ko 0-2 eyes (is this possible?)
 */

/* defines for the entire entry (obsolete) */
#define ATTACK_MOVE_SHIFT  8
#define ATTACK_BYTE        (0xff << ATTACK_MOVE_SHIFT) /* bit mask */
#define ATTACK_NO_MOVE     (255 << ATTACK_MOVE_SHIFT)  /* value 255 */
#define DEFENSE_MOVE_SHIFT 16
#define DEFENSE_BYTE       (0xff << DEFENSE_MOVE_SHIFT) /* bit mask */
#define DEFENSE_NO_MOVE    (255 << DEFENSE_MOVE_SHIFT)

/* defines for the result byte */
#define ATTACK_VISITED       0x80
#define ATTACK_RESULT_SHIFT  4
#define DEFENSE_VISITED      0x08
#define DEFENSE_RESULT_SHIFT 0
#define RESULT_MASK          0x07


/* The hash table will not have to handle any collisions, since it
 * is complete. 
 */
#define EYE_HASH_SIZE  (3*3*3*3*3*3*3*3*3*3) /* 3^MAX_EYE_SIZE */


static int  eyehash_table[EYE_HASH_SIZE];


static void  eyehash_clear(void);
static int   eyehash_index(Location_set *eyepoints);


/*
 * Clear the hash table.
 */

static void
eyehash_clear()
{
  memset(eyehash_table, 0, sizeof(eyehash_table));
}


/*
 * Return the index to the hash entry for the current position, and the
 * side to move.
 *
 * FIXME: This has to be changed when we take into account ko.
 */

static int
eyehash_index(Location_set *eyepoints)
{
  Location_set_iterator   iter;
  int  hashval;
  int  i, j;

  /* Step through all locations in `eyepoints'. */
  hashval = 0;
  for (ls_iter_init(eyepoints, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
    hashval = hashval * 3 + p[i][j];
#if 0
  hashval = hashval * 2 + (to_move - 1);
#else
  hashval = hashval * 2;
#endif

  return hashval;
}


/*
 * Get the attacking move from the cell at `hashindex'.  
 *
 * The move is coded as the nth move within `eyepoints'.  The NULL move
 * (a pass) is coded as 255.
 */

static void
eyehash_get_attack_move(Location_set *eyepoints, int hashindex,
			int *attacki, int *attackj)
{
  Location_set_iterator  iter;
  int  move_index;
  int  m, n;
  int  i;

  move_index = (eyehash_table[hashindex] >> ATTACK_MOVE_SHIFT) & 255;

  if (move_index == 255) {
    m = -1;
    n = -1;
  } else
    for (ls_iter_init(eyepoints, &iter), i = 0;
	 ls_iter_next(&iter, &m, &n), i < move_index;
	 ls_iter_update(&iter), i++)
      /* NOTHING */;

  *attacki = m;
  *attackj = n;
}


/*
 * Set the attacking move in the cell at `hashindex' to (m, n).
 */

static void
eyehash_set_attack_move(Location_set *eyepoints, int hashindex,
			int attacki, int attackj)
{
  Location_set_iterator  iter;
  int  m,n;
  int  i;

  if (attacki == -1) {
    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
    return;
  }

  for (ls_iter_init(eyepoints, &iter), i = 0;
       ls_iter_next(&iter, &m, &n);
       ls_iter_update(&iter), i++)
    if (m == attacki && n == attackj) {
      /* We don't have to mask away the previous entry because we
       * never change anything.
       */
      eyehash_table[hashindex] |= i << ATTACK_MOVE_SHIFT;
      return;
    }

  assert(0);			/* shouldn't happen */
}


/*
 * Get the defending move from the cell at `hashindex'.  
 */

static void
eyehash_get_defense_move(Location_set *eyepoints, int hashindex,
			 int *defendi, int *defendj)
{
  Location_set_iterator  iter;
  int  move_index;
  int  m, n;
  int  i;

  move_index = (eyehash_table[hashindex] >> DEFENSE_MOVE_SHIFT) & 255;

  if (move_index == 255) {
    m = -1;
    n = -1;
  } else
    for (ls_iter_init(eyepoints, &iter), i = 0;
	 ls_iter_next(&iter, &m, &n), i < move_index;
	 ls_iter_update(&iter), i++)
      /* NOTHING */;

  *defendi = m;
  *defendj = n;
}


/*
 * Set the defending move in the cell at `hashindex' to (m, n).
 */

static void
eyehash_set_defense_move(Location_set *eyepoints, int hashindex,
			 int defendi, int defendj)
{
  Location_set_iterator  iter;
  int  m,n;
  int  i;

  if (defendi == -1) {
    eyehash_table[hashindex] |= DEFENSE_NO_MOVE;
    return;
  }

  for (ls_iter_init(eyepoints, &iter), i = 0;
       ls_iter_next(&iter, &m, &n);
       ls_iter_update(&iter), i++)
    if (m == defendi && n == defendj) {
      eyehash_table[hashindex] |= i << DEFENSE_MOVE_SHIFT;
      return;
    }

  assert(0);			/* Shouldn't happen */
}


/* ================================================================ */
/*                          Eye recognition                         */
/* ================================================================ */


/* The boundary of the eye we are investigating. */
static Location_set  boundary;


/*
 * Compute the maximum and minimum number of eyes reachable from the
 * eye space at (m, n). The maximum is defined for color to play
 * first, and the minimum is defined for other_color to play first.
 *
 * If max != min, (*attacki, *attackj) is set to the move which
 * reduces the eye and (*defendi, *defendj) is set to the move which
 * defends it. Most of the time the attack point and the defense point
 * will be the same.
 */

void
recognize_eye2(int m, int  n, 
	       int *attacki, int *attackj, int *defendi, int *defendj,
	       int *max, int *min, 
	       struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
 	       int add_moves, int color)
{
  int eye_color, other;
  int i, j;
  int k;
  Location_set           eyepoints;
  Location_set           vital_moves;
  Location_set           diag_defends;
  Location_set           diag_attacks;
  Location_set_iterator  iter;
  int   result1, result2;
  int   max1, min1;
  int   attacki1, attackj1;
  int   defendi1, defendj1;
  int  save_stackp;

  /* FIXME: Remove the unused parameter entirely if it's not needed. */
  UNUSED(color);
  
  save_stackp = stackp;

  /* For use by debug outputs. */
  stackp_when_called = stackp;
  
  /* Set `eye_color' to the owner of the eye. */
  eye_color = eyedata[m][n].color;
  if (eye_color == BLACK_BORDER)
    eye_color = BLACK;
  if (eye_color == WHITE_BORDER)
    eye_color = WHITE;
  other = OTHER_COLOR(eye_color);

  /* Treat eyes with false marginals as a special case. 
   * This should be taken care of early, since it calls recognize_eye
   * recursively.
   */
  if (false_marginal_eye(m, n, attacki, attackj, defendi, defendj,
			 max, min, eyedata, add_moves, eye_color))
    return;

  /* Collect all the points of the eye into the set. 
   * Also, collect all diagonal points for half eyes.
   */
  location_set_clear(&eyepoints);
  location_set_clear(&diag_defends);
  location_set_clear(&diag_attacks);
  for (i = 0; i < board_size; ++i)
    for (j = 0; j < board_size; ++j) {
      if (eyedata[i][j].origini == m && eyedata[i][j].originj == n) {
	location_set_add(&eyepoints, i, j);
	if (halfeye(i, j)) {
	  for (k=0; k<half_eye[i][j].num_defends; k++)
	    location_set_add(&diag_defends,
			     half_eye[i][j].ai[k], half_eye[i][j].aj[k]);
	  for (k=0; k<half_eye[i][j].num_attacks; k++)
	    location_set_add(&diag_attacks,
			     half_eye[i][j].di[k], half_eye[i][j].dj[k]);
	}
      }    
    }

  /* Get the boundary. */
  get_boundary(&boundary, eyedata, &eyepoints, eye_color);

  /* Clear the hash table for this eye. */
  eyehash_clear();

  if (debug & DEBUG_EYES) {
    fprintf(stderr, "Eye (%s): ", color_to_string(eye_color));
    location_set_print(stderr, &eyepoints);
    fprintf(stderr, "Boundary:    ");
    location_set_print(stderr, &boundary);
  }

  location_set_clear(&vital_moves);
  DEBUG(DEBUG_EYES, "================  MINIMIZE ================\n");
  result1 = minimize_eyes(eye_color, &min1, &attacki1, &attackj1, 
			  eyedata, &eyepoints, &diag_defends, &diag_attacks,
			  &vital_moves);
  DEBUG(DEBUG_EYES, "================  MAXIMIZE ================\n");
  result2 = maximize_eyes(eye_color, &max1, &defendi1, &defendj1, 
			  eyedata, &eyepoints, &diag_defends, &diag_attacks,
			  &vital_moves);

  *min = min1;
  *max = max1;
  if (min1 != max1) {
    if (attacki) *attacki = attacki1;
    if (attackj) *attackj = attackj1;
    if (defendi) *defendi = defendi1;
    if (defendj) *defendj = defendj1;
    DEBUG(DEBUG_EYES, "  vital points (attack):  %m\n", *attacki, *attackj);
    DEBUG(DEBUG_EYES, "  vital points (defense): %m\n", *defendi, *defendj);
    if (add_moves) {
      for (ls_iter_init(&vital_moves, &iter);
	   ls_iter_next(&iter, &i, &j);
	   ls_iter_update(&iter))
	{
	  add_vital_eye_move(i, j, m, n, eye_color);
	}
    }
  }

  assert(stackp == save_stackp);
}


static int
false_marginal_eye(int m, int  n, 
		   int *attacki, int *attackj, int *defendi, int *defendj,
		   int *max, int *min, 
		   struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
		   int add_moves, int color)
{
  int  other = OTHER_COLOR(color);
  int  libi[2];
  int  libj[2];
  int  min1;
  int  max1;

  /* If the eye point is empty, this eye has definitely not a false margin. */
  if (p[m][n] != other) 
    return 0;

  /* A string within an eye with a false margin always has one liberty. */
  if (findlib(m, n, 2, libi, libj) != 1)
    return 0;

  /* See if it is a real false margin. This is the real test. */
  if (eyedata[libi[0]][libj[0]].false_margin == 0)
    return 0;

  /* Ok, so it really is a false margin.  What we do now, is that we
   * play at the false margin and treat this as an ordinary eye.
   */
  if (trymove(libi[0], libj[0], color, "false_marginal_eye", -1, -1)) {
    recognize_eye2(m, n, attacki, attackj, defendi, defendj,
		   &max1, &min1, eyedata, add_moves, color);

    /* We can only get the minimum eyes since the opponent always gets
     * the first move. 
     */
    *max = min1;
    *min = min1;

    popgo();
  }
  else
    /* Something is strange. */
    assert(0);

  return 1;
}


/*
 * We have a set of locations in eyepoints forming an eye shape.
 * Check min number of eyes we can get if the opponent moves first.
 * `eye_color' is the owner of the eye.
 *
 * Return values:
 *   0: Ok. Min eyes in *min.
 *  -1: No result.  A cycle in the reading.
 */

static int
minimize_eyes(int eye_color, int *min, int *attacki, int *attackj, 
	      struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
	      Location_set *eyepoints, 
	      Location_set *diag_defends, 
	      Location_set *diag_attacks,
	      Location_set *moves)
{
  int  other = OTHER_COLOR(eye_color);
  int  i, j;
  int  m, n;
  Location_set attacks;
  Location_set_iterator   iter;
  Location_set_iterator   iter2;
  int  num_other;
  int  num_legal_moves;
  int  num_eyes;
  int  bordercolor;
  int  eyesize;
  int  max1;
  int  localmax;
  int  hashindex;
  int  result;
  int  attacki1, attackj1;
  int  localattacki = -1;
  int  localattackj = -1;

  int  save_stackp;

  if (stackp - stackp_when_called < DEBUG_LIMIT)
    TRACE("entering minimize_eyes: stackp = %d\n", stackp);

  /* Collect attacking moves, union of eyepoints and diagonal vital moves */
  location_set_copy(&attacks, eyepoints);
  location_set_union(&attacks, diag_attacks);

  /* Check the hash table and see if we have been here before. */
  /* FIXME: The hashindex could be calculated incrementally. */
  hashindex = eyehash_index(&attacks);
  RTRACE("hashindex = %d\n", hashindex);
  if ((eyehash_table[hashindex] & ATTACK_VISITED) != 0) {
    result = (eyehash_table[hashindex] >> ATTACK_RESULT_SHIFT) & RESULT_MASK;
    if (result == 0) {
      /* No result known yet.  We have come back to the same position
       * again within the reading.  Just return and hope that we get a
       * better result on some other branch. 
       */
      if (stackp - stackp_when_called < DEBUG_LIMIT) {
	TRACE("exiting minimize_eyes directly because of a reading loop.\n");
	TRACE("hashentry = %x\n", eyehash_table[hashindex]);
      }
      return -1;
    }
    else if (result <= 3) {
      *min = result - 1;
      eyehash_get_attack_move(&attacks, hashindex, attacki, attackj);

      if (stackp - stackp_when_called < DEBUG_LIMIT) {
	if (attacki)
	  TRACE("exiting minimize_eyes - got result %d eyes (move %m) from the table.\n",
		*min, *attacki, *attackj);
	else
	  TRACE("exiting minimize_eyes - got result %d eyes from the table.\n",
		*min);
      }
      RTRACE("hashindex = %d\n", hashindex);
      return 0;
    }
  }

  if (verbose
      && stackp - stackp_when_called < DEBUG_LIMIT
      && (debug & DEBUG_EYES))
    life_showboard();

  /* Collect how many moves the attacking color can do and how many stones
   * he has within the eye.
   */
  num_other = 0;
  num_legal_moves = 0;
  num_eyes = 0;
  for (ls_iter_init(&attacks, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (p[i][j] == eye_color)
      continue;

    if (p[i][j] == other) {
      num_other++;
      continue;
    } 

    if (legal(i, j, other))
      num_legal_moves++;
    else {
      bordercolor = find_border(i, j, NULL, &eyesize);
      /* FIXME: Do we need to use the bordercolor? */
      if (eyesize == 1 && !eyedata[i][j].marginal)
	num_eyes++;
    } 
  }

  /* Mark this node as visited, but without result so far. */
  eyehash_table[hashindex] |= ATTACK_VISITED;

  /* Check if we are done. This means that there are no stones of
   * the attacking color and that all the eye points are illegal
   * to move into. The only way this could happen is that they are
   * of size 1 and without possibility to capture anything.
   */
  if (num_other == 0 && num_legal_moves == 0) {
    if (num_eyes > 2)
      num_eyes = 2;
    *min = num_eyes;
    if (attacki) *attacki = -1;
    if (attackj) *attackj = -1;
    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
    eyehash_table[hashindex] |= (num_eyes + 1) << ATTACK_RESULT_SHIFT;

    if (stackp - stackp_when_called < DEBUG_LIMIT) {
      TRACE("exiting minimize_eyes (end of search) - result = %d\n", num_eyes);
    }
    RTRACE("hashindex = %d\n", hashindex);

    return 0;
  } 

  localmax = 9999;

  /* Try to see if doing nothing is enough.
   */
  if (stackp - stackp_when_called < DEBUG_LIMIT)
    TRACE("minimize_eyes: trying a PASS\n");
  
  result = maximize_eyes(eye_color, &max1, &attacki1, &attackj1, 
			 eyedata, eyepoints, diag_defends, diag_attacks,
			 NULL);
  if (result == 0) {
    if (max1 < localmax) {
      localmax = max1;
      localattacki = -1;
      localattackj = -1;
    }

    /* Check if the pass yielded 0 eyes.  If so, we can as well
     * return here. */
    if (localmax == 0) {
      *min = 0;
      if (attacki) *attacki = -1;
      if (attackj) *attackj = -1;
      eyehash_table[hashindex] |= ATTACK_NO_MOVE;
      eyehash_table[hashindex] |= (0 + 1) << ATTACK_RESULT_SHIFT;

      if (stackp - stackp_when_called < DEBUG_LIMIT)
	TRACE("exiting minimize_eyes - a pass sufficed\n");

      RTRACE("hashindex = %d\n", hashindex);
      return 0;
    }
  }  

  /* Now try all possible moves and see what we get. */
  for (ls_iter_init(&attacks, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    int  boundary_captured;

    if (p[i][j] != EMPTY)
      continue;

    /* Try the move and see if we can reduce the eyes. */
    save_stackp = stackp;
    boundary_captured = 0;
    if (stackp - stackp_when_called < DEBUG_LIMIT)
      TRACE("minimize_eyes: trymove at %m\n", i, j);
    if (trymove(i, j, other, "minimize_eyes", -1, -1)) {

      /* Check if any part of the boundary got captured.  If so, we assume
       * that we can get no eyes.
       *
       * FIXME: This is ok for now, but is it ok forever?
       */
      for (ls_iter_init(&boundary, &iter2);
	   ls_iter_next(&iter2, &m, &n);
	   ls_iter_update(&iter2))
      {
	if (p[m][n] == EMPTY) {
	  boundary_captured = 1;
	  break;
	}
      }

      if (!boundary_captured) {
	/* No boundary string was captured. The attacker has made his move.
	 * Now let's answer him and see how many eyes we can get.
	 */
	result = maximize_eyes(eye_color, &max1, &attacki1, &attackj1, 
			  eyedata, eyepoints, diag_defends, diag_attacks,
			  NULL);
	if (result == 0) {
	  if (max1 < localmax) {
	    localmax = max1;
	    localattacki = i;
	    localattackj = j;
 	    if (moves)
 	      location_set_clear(moves);
 	  } 
 	  if (max1 == localmax && moves)
 	    location_set_add(moves, i, j);
	}
      }

      popgo();
    }

    assert(stackp == save_stackp);

    /* Check if the boundary was captured or the move yielded 0 eyes.
     * If so, we can as well return here. 
     */
#if 0
    /* FIXME: Could use the test (!moves && ...) */
    if (boundary_captured || localmax == 0) {
      *min = 0;
      if (attacki) *attacki = i;
      if (attackj) *attackj = j;
      
      eyehash_set_attack_move(attacks, hashindex, i, j);
      eyehash_table[hashindex] |= (0 + 1) << ATTACK_RESULT_SHIFT;
      
      if (stackp - stackp_when_called < DEBUG_LIMIT) {
	if (attacki)
	  TRACE("exiting minimize_eyes (2) - result = 0, move %m\n",
		*attacki, *attackj);
	else
	  TRACE("exiting minimize_eyes (2) - result = 0\n");
      }
      RTRACE("hashindex = %d\n", hashindex);
      return 0;
    }
#else
    
    /* We really want all eye minimizing moves */
    if (boundary_captured) {
      if (localmax) {
	localmax = 0;
	localattacki = i;
	localattackj = j;
	if (moves)
	  location_set_clear(moves);
      }
      if (moves)
	location_set_add(moves, i, j);
    }
#endif
  }

  /* If we didn't find anywhere to play, see how many eye spaces there were. */
  if (localmax == 9999) {
    if (num_legal_moves > 0) {
      /* We didn't get a result even though there were legal moves.
       * This has to be because of a reading loop. Propagate the
       * non-result further up.
       */
      return -1;
    }

    /* FIXME: should be assert(0) */
    assert(1);			/* Shouldn't get here. */

    localmax = num_eyes;
    if (localmax > 2)
      localmax = 2;
    *min = localmax;
    if (attacki) *attacki = -1;
    if (attackj) *attackj = -1;

    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
    eyehash_table[hashindex] |= (localmax + 1) << ATTACK_RESULT_SHIFT;

    if (stackp - stackp_when_called < DEBUG_LIMIT) {
      if (attacki)
	TRACE("exiting minimize_eyes (3) - result = %d, move %m\n", 
	      num_eyes, *attacki, *attackj);
      else
	TRACE("exiting minimize_eyes (3) - result = %d\n", num_eyes);
    }
    RTRACE("hashindex = %d\n", hashindex);
    return 0;
  }

  if (localmax > 2)
    localmax = 2;

  *min = localmax;
  if (attacki) *attacki = localattacki;
  if (attackj) *attackj = localattackj;

  eyehash_set_attack_move(&attacks, hashindex, localattacki, localattackj);
  eyehash_table[hashindex] |= (localmax + 1) << ATTACK_RESULT_SHIFT;

  if (stackp - stackp_when_called < DEBUG_LIMIT) {
    if (attacki)
      TRACE("exiting minimize_eyes (4) - result = %d, move %m\n", 
	    localmax, *attacki, *attackj);
    else
      TRACE("exiting minimize_eyes (4) - result = %d\n", localmax);
  }
  RTRACE("hashindex = %d\n", hashindex);
  return 0;
}


/*
 * We have a set of locations in eyepoints forming an eye shape.
 * Check max number of eyes we can get if the opponent moves first.
 * `eye_color' is the owner of the eye.
 *
 * Return values:
 *   0: Ok. Max eyes in *max.
 *  -1: No result.  A cycle in the reading.
 */

static int
maximize_eyes(int eye_color, int *max, int *defendi, int *defendj, 
	      struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
	      Location_set *eyepoints,
	      Location_set *diag_defends, Location_set *diag_attacks,
	      Location_set *moves)
{
  int  i, j;
  Location_set defends;
  Location_set_iterator   iter;
  int  bordercolor;
  int  eyesize;
  int  min1;
  int  localmin;
  int  num_eyes;
  int  result;
  int  hashindex;
  int  defendi1, defendj1;
  int  localdefendi = -1;
  int  localdefendj = -1;
  int  special_case = 0;
  int  save_stackp;

  if (stackp - stackp_when_called < DEBUG_LIMIT)
    TRACE("entering maximize_eyes: stackp = %d\n", stackp);

  /* Collect defending moves, union of eyepoints and diagonal vital moves */
  location_set_copy(&defends, eyepoints);
  location_set_union(&defends, diag_defends);

  /* Check the hash table and see if we have been here before. */
  /* FIXME: The hashindex could be calculated incrementally. */
  hashindex = eyehash_index(&defends);
  RTRACE("hashindex = %d\n", hashindex);
  if ((eyehash_table[hashindex] & DEFENSE_VISITED)!= 0) {
    result = (eyehash_table[hashindex] >> DEFENSE_RESULT_SHIFT) & RESULT_MASK;
    if (result == 0) {
      /* No result known yet.  We have come back to the same position
       * again within the reading.  Just return and hope that we get a
       * better result on some other branch. 
       */
      if (stackp - stackp_when_called < DEBUG_LIMIT) {
	TRACE("exiting maximize_eyes directly because of a reading loop.\n");
      }
      RTRACE("hashindex = %d\n", hashindex);
      return -1;
    }
    else if (result <= 3) {
      *max = result - 1;
      if ((eyehash_table[hashindex] & DEFENSE_BYTE) != DEFENSE_NO_MOVE)
	eyehash_get_defense_move(&defends, hashindex, defendi, defendj);

      if (stackp - stackp_when_called < DEBUG_LIMIT) {
	if (defendi)
	  TRACE("exiting maximize_eyes - got result %d eyes (move %m) from the table.\n",
		*max, *defendi, *defendj);
	else
	  TRACE("exiting maximize_eyes - got result %d eyes from the table.\n",
		*max);
      }
      RTRACE("hashindex = %d\n", hashindex);

      return 0;
    }
  }

  if (stackp - stackp_when_called < DEBUG_LIMIT && verbose)
    life_showboard();

  localmin = -1;
  num_eyes = 0;

  /* Mark this node as visited, but without result so far. */
  eyehash_table[hashindex] |= DEFENSE_VISITED;

  /* Try all possible moves and see what we get. */
  for (ls_iter_init(&defends, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (p[i][j] != EMPTY)
      continue;

    /* If the eye is of size 1, of the eye owners color, and 
     * none of the neighbouring strings is in atari, then this
     * is a proper eye. Don't play there.
     */
    bordercolor = find_border(i, j, NULL, &eyesize);
    if (eyesize == 1) {
      if ((i == 0 
	   || (p[i-1][j] == eye_color
	       && countlib(i-1, j) > 1))
	  && (i == board_size-1
	      || (p[i+1][j] == eye_color
		  && countlib(i+1, j) > 1))
	  && (j == 0
	      || (p[i][j-1] == eye_color
		  && countlib(i, j-1) > 1))
	  && (j == board_size-1
	      || (p[i][j+1] == eye_color
		  && countlib(i, j+1) > 1)))
      {
	num_eyes++;
	continue;
      }
    }

    /* Try the move and see if we can keep the eyes. */
    save_stackp = stackp;
    if (stackp - stackp_when_called < DEBUG_LIMIT)
      TRACE("maximize_eyes: trymove at %m\n", i, j);
    if (trymove(i, j, eye_color, "maximize_eyes", -1, -1)) {

      /* Ok, we made our move.  Now let the opponent do his, and see
       * how many eyes we can get. 
       */
      result = minimize_eyes(eye_color, &min1, &defendi1, &defendj1,
			  eyedata, eyepoints, diag_defends, diag_attacks,
			  NULL);
      if (result == 0) {
	if (min1 > localmin) {
	  localmin = min1;
	  localdefendi = i;
	  localdefendj = j;
 	  if (moves)
 	    location_set_clear(moves);
	}
 	if (min1 == localmin && moves)
 	  location_set_add(moves, i, j);
      }

      popgo();
    }
    assert(stackp == save_stackp);
  }
  
  /* If there were nowhere to play, get number of eyes */
  if (localmin == -1)
    localmin = num_eyes;

  if (localmin > 2)
    localmin = 2;

  *max = localmin;
  if (defendi) *defendi = localdefendi;
  if (defendj) *defendj = localdefendj;
  eyehash_set_defense_move(&defends, hashindex, localdefendi, localdefendj);

  eyehash_table[hashindex] |= (localmin + 1) << DEFENSE_RESULT_SHIFT;

  /* If this was the special case, we have to copy the result to the
   * attack cell.
   */
  if (special_case) {
    int hashindex;

    hashindex = eyehash_index(&defends);
    TRACE("Special case: Use defense result also for attack.\n");
    RTRACE("Storing move in hashindex %d.\n", hashindex);
    eyehash_table[hashindex] |= (localmin + 1) << ATTACK_RESULT_SHIFT;
    eyehash_set_attack_move(&defends, hashindex, -1, -1);
  }

  if (stackp - stackp_when_called < DEBUG_LIMIT) {
    if (defendi)
      TRACE("exiting maximize_eyes (1) - result = %d, move %m\n", 
	    localmin, *defendi, *defendj);
    else
      TRACE("exiting maximize_eyes (1) - result = %d\n", localmin);
  }
  RTRACE("hashindex = %d\n", hashindex);

  return 0;
}


static void
get_boundary(Location_set *boundary,
	     struct eye_data eyedata[MAX_BOARD][MAX_BOARD],
	     Location_set *eyepoints, int color)
{
  Location_set_iterator  iter;
  int  i, j;

  location_set_copy(boundary, eyepoints);
  for (ls_iter_init(boundary, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (eyedata[i][j].marginal)
      location_set_del(boundary, i, j);
  }

  location_set_expand(boundary, 0);
  location_set_subtract(boundary, eyepoints);

  /* Only keep those parts of the boundary that are actually there. */
  for (ls_iter_init(boundary, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (p[i][j] != color)
      location_set_del(boundary, i, j);
  }
}


/*
 * Show the board in a form suitable for debugging life code.
 * This is also callable from GDB.
 */

static void
life_showboard()
{
  int i, j;

  start_draw_board();
  for (i=0; i<board_size; i++) {
    for (j=0; j<board_size; j++) {
      int c;
      int color = p[i][j];

      if (color == WHITE) {
	if (move_in_stack(i, j, stackp_when_called))
	  c = 'o';
	else
	  c = 'O';
      }
      else if (color == BLACK) {
	if (move_in_stack(i, j, stackp_when_called))
	  c = 'x';
	else
	  c = 'X';
      }
      else
	c = EMPTY;
      draw_char(i, j, c);
    }
  }
  end_draw_board();
}

/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
