/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "liberty.h"
#include "random.h"
#include "move_reasons.h"


/*
 * Private data structures used to collect information about moves.
 */

#define MAX_MOVE_REASONS 1000
#define MAX_WORMS 2*MAX_BOARD*MAX_BOARD/3
#define MAX_DRAGONS MAX_WORMS
#define MAX_CONNECTIONS 4*MAX_WORMS
#define MAX_SEMEAIS MAX_WORMS
#define MAX_WORM_PAIRS MAX_WORMS
#define MAX_EYES MAX_BOARD*MAX_BOARD/2
#define MAX_LUNCHES MAX_WORMS

static struct move_data move[MAX_BOARD][MAX_BOARD];
static struct move_reason move_reasons[MAX_MOVE_REASONS];
static int next_reason;

/* Worms */
static int wormi[MAX_WORMS];
static int wormj[MAX_WORMS];
static int next_worm;

/* Dragons */
static int dragoni[MAX_DRAGONS];
static int dragonj[MAX_DRAGONS];
static int next_dragon;

/* Connections */
static int conn_dragon1[MAX_CONNECTIONS];
static int conn_dragon2[MAX_CONNECTIONS];
static int next_connection;

/* Semeais */
static int semeai_myworm[MAX_SEMEAIS];
static int semeai_yourworm[MAX_SEMEAIS];
static int next_semeai;

/* Unordered worm pairs */
static int worm_pair1[MAX_WORM_PAIRS];
static int worm_pair2[MAX_WORM_PAIRS];
static int next_worm_pair;

/* Eye shapes */
static int eyei[MAX_EYES];
static int eyej[MAX_EYES];
static int eyecolor[MAX_EYES];
static int next_eye;

/* Lunches */
static int lunch_dragon[MAX_LUNCHES]; /* eater */
static int lunch_worm[MAX_LUNCHES];   /* food */
static int next_lunch;

/* Initialize move reason data structures. */
void
clear_move_reasons(void)
{
  int i;
  int j;
  int k;
  next_reason = 0;
  next_worm = 0;
  next_dragon = 0;
  next_connection = 0;
  next_semeai = 0;
  next_worm_pair = 0;
  next_eye = 0;
  next_lunch = 0;
  
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      move[i][j].value             = 0.0;
      move[i][j].territorial_value = 0.0;
      move[i][j].strategical_value = 0.0;
      move[i][j].influence_value   = 0.0;
      move[i][j].shape_value       = 0.0;
      move[i][j].followup_value    = 0.0;
      move[i][j].secondary_value   = 0.0;
      move[i][j].min_value         = 0.0;
      move[i][j].max_value         = HUGE_MOVE_VALUE;
      move[i][j].min_territory     = 0.0;
      move[i][j].max_territory     = HUGE_MOVE_VALUE;
      for (k=0; k<MAX_REASONS; k++)
	move[i][j].reason[k] = -1;
    }
}

/*
 * Find the index of a worm in the list of worms. If necessary,
 * add a new entry. (ai, aj) must point to the origin of the worm.
 */
static int
find_worm(int ai, int aj)
{
  int k;
  for (k=0; k<next_worm; k++)
    if ((wormi[k] == ai) && (wormj[k] == aj))
      return k;

  /* Add a new entry. */
  assert(next_worm < MAX_WORMS);
  wormi[next_worm] = ai;
  wormj[next_worm] = aj;
  next_worm++;
  return next_worm - 1;
}

/*
 * Find the index of a dragon in the list of dragons. If necessary,
 * add a new entry. (ai, aj) must point to the origin of the dragon.
 */
static int
find_dragon(int ai, int aj)
{
  int k;
  for (k=0; k<next_dragon; k++)
    if ((dragoni[k] == ai) && (dragonj[k] == aj))
      return k;

  /* Add a new entry. */
  assert(next_dragon < MAX_DRAGONS);
  dragoni[next_dragon] = ai;
  dragonj[next_dragon] = aj;
  next_dragon++;
  return next_dragon - 1;
}

/*
 * Find the index of a connection in the list of connections.
 * If necessary, add a new entry.
 */
static int
find_connection(int dragon1, int dragon2)
{
  int k;

  if (dragon1 > dragon2) {
    /* Swap to canonical order. */
    int tmp = dragon1;
    dragon1 = dragon2;
    dragon2 = tmp;
  }
  
  for (k=0; k<next_connection; k++)
    if ((conn_dragon1[k] == dragon1) && (conn_dragon2[k] == dragon2))
      return k;

  /* Add a new entry. */
  assert(next_connection < MAX_CONNECTIONS);
  conn_dragon1[next_connection] = dragon1;
  conn_dragon2[next_connection] = dragon2;
  next_connection++;
  return next_connection - 1;
}

/*
 * Find the index of a semeai in the list of semeais.
 * If necessary, add a new entry.
 */
static int
find_semeai(int myworm, int yourworm)
{
  int k;

  for (k=0; k<next_semeai; k++)
    if ((semeai_myworm[k] == myworm) && (semeai_yourworm[k] == yourworm))
      return k;

  /* Add a new entry. */
  assert(next_semeai < MAX_SEMEAIS);
  semeai_myworm[next_semeai] = myworm;
  semeai_yourworm[next_semeai] = yourworm;
  next_semeai++;
  return next_semeai - 1;
}

/*
 * Find the index of an unordered pair of worms in the list of worm pairs.
 * If necessary, add a new entry.
 */
static int
find_worm_pair(int worm1, int worm2)
{
  int k;

  /* Make sure the worms are ordered canonically. */
  if (worm1 > worm2) {
    int tmp = worm1;
    worm1 = worm2;
    worm2 = tmp;
  }
  
  for (k=0; k<next_worm_pair; k++)
    if ((worm_pair1[k] == worm1) && (worm_pair2[k] == worm2))
      return k;

  /* Add a new entry. */
  assert(next_worm_pair < MAX_WORM_PAIRS);
  worm_pair1[next_worm_pair] = worm1;
  worm_pair2[next_worm_pair] = worm2;
  next_worm_pair++;
  return next_worm_pair - 1;
}

/*
 * Find the index of an eye space in the list of eye spaces.
 * If necessary, add a new entry.
 */
static int
find_eye(int i, int j, int color)
{
  int k;

  for (k=0; k<next_eye; k++)
    if ((eyei[k] == i) && (eyej[k] == j) && eyecolor[k] == color)
      return k;

  /* Add a new entry. */
  assert(next_eye < MAX_EYES);
  eyei[next_eye] = i;
  eyej[next_eye] = j;
  eyecolor[next_eye] = color;
  next_eye++;
  return next_eye - 1;
}

/*
 * See if a lunch is already in the list of lunches, otherwise add a new
 * entry. A lunch is in this context a pair of eater (a dragon) and food
 * (a worm).
 */
void
add_lunch(int ai, int aj, int bi, int bj)
{
  int k;
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  int worm1 = find_worm(worm[bi][bj].origini, worm[bi][bj].originj);

  for (k=0; k<next_lunch; k++)
    if ((lunch_dragon[k] == dragon1) && (lunch_worm[k] == worm1))
      return;

  /* Add a new entry. */
  assert(next_lunch < MAX_LUNCHES);
  lunch_dragon[next_lunch] = dragon1;
  lunch_worm[next_lunch] = worm1;
  next_lunch++;
  return;
}

/*
 * Remove a lunch from the list of lunches.  A lunch is in this context a pair
 * of eater (a dragon) and food (a worm).  
*/
void
remove_lunch(int ai, int aj, int bi, int bj)
{
  int k;
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  int worm1 = find_worm(worm[bi][bj].origini, worm[bi][bj].originj);

  for (k=0; k<next_lunch; k++)
    if ((lunch_dragon[k] == dragon1) && (lunch_worm[k] == worm1))
      break;

  if (k == next_lunch)
    return; /* Not found */
  
  /* Remove entry k. */
  lunch_dragon[k] = lunch_dragon[next_lunch - 1];
  lunch_worm[k] = lunch_worm[next_lunch - 1];
  next_lunch--;
}

/*
 * Find a reason in the list of reasons. If necessary, add a new entry.
 */
static int
find_reason(int type, int what)
{
  int k;
  for (k=0; k<next_reason; k++)
    if ((move_reasons[k].type == type) && (move_reasons[k].what == what))
      return k;

  /* Add a new entry. */
  assert(next_reason < MAX_MOVE_REASONS);
  move_reasons[next_reason].type = type;
  move_reasons[next_reason].what = what;
  next_reason++;
  return next_reason - 1;
}

/*
 * Add a move reason for (ti, tj) if it's not already there or the
 * table is full.
 */ 
static void
add_move_reason(int ti, int tj, int type, int what)
{
  int k;
  for (k=0; k<MAX_REASONS; k++) {
    int r = move[ti][tj].reason[k];
    if (r < 0)
      break;
    if (move_reasons[r].type == type
	&& move_reasons[r].what == what)
      return;  /* Reason already listed. */
  }
  /* Reason not found, add it if there is place left. */
  if (k<MAX_REASONS)
    move[ti][tj].reason[k] = find_reason(type, what);
}

/*
 * Remove a move reason for (ti, tj). Ignore silently if the reason
 * wasn't there.
 */ 
static void
remove_move_reason(int ti, int tj, int type, int what)
{
  int k;
  int n = -1; /* Position of the move reason to be deleted. */
  for (k=0; k<MAX_REASONS; k++) {
    int r = move[ti][tj].reason[k];
    if (r < 0)
      break;
    if (move_reasons[r].type == type
	&& move_reasons[r].what == what)
      n = k;
  }

  if (n == -1)
    return; /* Move reason wasn't there. */

  /* Now move the last move reason to position n, thereby removing the
   * one we were looking for.
   */
  k--;
  move[ti][tj].reason[n] = move[ti][tj].reason[k];
  move[ti][tj].reason[k] = -1;
}

/*
 * Check whether a move reason already is recorded for a move.
 */
static int
move_reason_known(int ti, int tj, int type, int what)
{
  int k;
  int r;
  for (k=0; k<MAX_REASONS; k++) {
    r = move[ti][tj].reason[k];
    if (r < 0)
      break;
    if (move_reasons[r].type == type
	&& move_reasons[r].what == what)
      return 1;
  }
  return 0;
}

/*
 * Add to the reasons for the move at (ti, tj) that it defends the worm
 * at (ai, aj).
 */
void
add_defense_move(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini,
				  worm[ai][aj].originj);
  add_move_reason(ti, tj, DEFEND_MOVE, worm_number);
}

/* Query whether a defense move is already known. */
int
defense_move_known(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  return move_reason_known(ti, tj, DEFEND_MOVE, worm_number);
}

/*
 * Add to the reasons for the move at (ti, tj) that it attacks the worm
 * at (ai, aj).
 */
void
add_attack_move(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  add_move_reason(ti, tj, ATTACK_MOVE, worm_number);
}

/* Query whether an attack move is already known. */
int
attack_move_known(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  return move_reason_known(ti, tj, ATTACK_MOVE, worm_number);
}

/*
 * Remove from the reasons for the move at (ti, tj) that it defends
 * the worm at (ai, aj). We do this by adding a NON_DEFEND move
 * reason and wait until later to actually remove it. Otherwise it may
 * be added again.
 *
 * We must also check that there do exist a defense move reason for
 * this worm. Otherwise we may end up in an infinite loop when trying
 * to actually remove it.
 */
void
remove_defense_move(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  if (move_reason_known(ti, tj, DEFEND_MOVE, worm_number))
    add_move_reason(ti, tj, NON_DEFEND_MOVE, worm_number);
}


/*
 * Remove from the reasons for the move at (ti, tj) that it attacks
 * the worm at (ai, aj). We do this by adding a NON_ATTACK move
 * reason and wait until later to actually remove it. Otherwise it may
 * be added again.
 *
 * We must also check that there do exist an attack move reason for
 * this worm. Otherwise we may end up in an infinite loop when trying
 * to actually remove it.
 */
void
remove_attack_move(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  if (move_reason_known(ti, tj, ATTACK_MOVE, worm_number))
    add_move_reason(ti, tj, NON_ATTACK_MOVE, worm_number);
}

/*
 * Add to the reasons for the move at (ti, tj) that it connects the
 * dragons at (ai, aj) and (bi, bj). Require that the dragons are
 * distinct.
 */
void
add_connection_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  int dragon2 = find_dragon(dragon[bi][bj].origini, dragon[bi][bj].originj);
  int connection;
  if (dragon1 == dragon2)
    return;
  connection = find_connection(dragon1, dragon2);
  add_move_reason(ti, tj, CONNECT_MOVE, connection);
}

/*
 * Add to the reasons for the move at (ti, tj) that it cuts the
 * dragons at (ai, aj) and (bi, bj). Require that the dragons are
 * distinct.
 */
void
add_cut_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  int dragon2 = find_dragon(dragon[bi][bj].origini, dragon[bi][bj].originj);
  int connection;
  if (dragon1 == dragon2)
    return;
  connection = find_connection(dragon1, dragon2);

  /*
   * Ignore the cut or connection if either (ai, aj) or (bi, bj)
   * points to a tactically captured worm.
   */
  if ((worm[ai][aj].attacki != -1 && worm[ai][aj].defendi == -1)
      || (worm[bi][bj].attacki != -1 && worm[bi][bj].defendi == -1))
    return;
  
  add_move_reason(ti, tj, CUT_MOVE, connection);
}

/*
 * Add to the reasons for the move at (ti, tj) that it is an anti-suji.
 * This means that it's a locally inferior move or for some other reason
 * must *not* be played.
 */
void
add_antisuji_move(int ti, int tj)
{
  add_move_reason(ti, tj, ANTISUJI_MOVE, 0);
}

/*
 * Add to the reasons for the move at (ti, tj) that it wins a semeai
 * between my worm at (ai, aj) and your worm at (bi, bj).
 */
void
add_semeai_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int myworm = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  int yourworm = find_worm(worm[bi][bj].origini, worm[bi][bj].originj);
  int this_semeai = find_semeai(myworm, yourworm);
  add_move_reason(ti, tj, SEMEAI_MOVE, this_semeai);
}

/*
 * Add to the reasons for the move at (ti, tj) that it's the vital
 * point for the eye space at (ai, aj) of color.
 */
void
add_vital_eye_move(int ti, int tj, int ai, int aj, int color)
{
  int eye;
  if (color == WHITE)
    eye = find_eye(white_eye[ai][aj].origini, white_eye[ai][aj].originj,
		   color);
  else
    eye = find_eye(black_eye[ai][aj].origini, black_eye[ai][aj].originj,
		   color);
  add_move_reason(ti, tj, VITAL_EYE_MOVE, eye);
}

/*
 * Add to the reasons for the move at (ti, tj) that it attacks
 * either (ai, aj) or (bi, bj) (e.g. a double atari). This move
 * reason is only used for double attacks on opponent stones.
 *
 * Before accepting the move reason, check that the worms are
 * distinct and that neither is undefendable.
 */
void
add_attack_either_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int worm1 = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  int worm2 = find_worm(worm[bi][bj].origini, worm[bi][bj].originj);
  int worm_pair;
  if (worm1 == worm2)
    return;
  if (worm[ai][aj].attack_code != 0 && worm[ai][aj].defend_code == 0)
    return;
  if (worm[bi][bj].attack_code != 0 && worm[bi][bj].defend_code == 0)
    return;
  worm_pair = find_worm_pair(worm1, worm2);
  add_move_reason(ti, tj, ATTACK_EITHER_MOVE, worm_pair);
}

/*
 * Add to the reasons for the move at (ti, tj) that it defends
 * both (ai, aj) and (bi, bj) (e.g. from a double atari). This move
 * reason is only used for defense of own stones.
 */
void
add_defend_both_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int worm1 = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  int worm2 = find_worm(worm[bi][bj].origini, worm[bi][bj].originj);
  int worm_pair = find_worm_pair(worm1, worm2);
  add_move_reason(ti, tj, DEFEND_BOTH_MOVE, worm_pair);
}

/*
 * Add to the reasons for the move at (ti, tj) that it secures
 * territory by blocking.
 */
void
add_block_territory_move(int ti, int tj)
{
  add_move_reason(ti, tj, BLOCK_TERRITORY_MOVE, 0);
}

/*
 * Add to the reasons for the move at (ti, tj) that it expands
 * territory.
 */
void
add_expand_territory_move(int ti, int tj)
{
  add_move_reason(ti, tj, EXPAND_TERRITORY_MOVE, 0);
}

/*
 * Add to the reasons for the move at (ti, tj) that it expands
 * moyo.
 */
void
add_expand_moyo_move(int ti, int tj)
{
  add_move_reason(ti, tj, EXPAND_MOYO_MOVE, 0);
}

/*
 * Increase or decrease the shape value for the move at (ti, tj).
 */
void
add_shape_value(int ti, int tj, float value)
{
  move[ti][tj].shape_value += value;
}

/*
 * Add to the reasons for the move at (ti, tj) that it attacks
 * the dragon (ai, aj) on a strategical level.
 */
void
add_strategical_attack_move(int ti, int tj, int ai, int aj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  add_move_reason(ti, tj, STRATEGIC_ATTACK_MOVE, dragon1);
}

/*
 * Add to the reasons for the move at (ti, tj) that it defends
 * the dragon (ai, aj) on a strategical level.
 */
void
add_strategical_defense_move(int ti, int tj, int ai, int aj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  add_move_reason(ti, tj, STRATEGIC_DEFEND_MOVE, dragon1);
}

/*
 * Add to the reasons for the move at (ti, tj) that the mouse
 * code reports an attack on the dragon (ai, aj).
 */
void
add_mouse_attack_move(int ti, int tj, int ai, int aj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  add_move_reason(ti, tj, MOUSE_ATTACK_MOVE, dragon1);
}

/*
 * Add to the reasons for the move at (ti, tj) that the mouse
 * code reports a defense of the dragon (ai, aj).
 */
void
add_mouse_defense_move(int ti, int tj, int ai, int aj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  add_move_reason(ti, tj, MOUSE_DEFEND_MOVE, dragon1);
}

/*
 * Add value of followup moves. 
 */
void
add_followup_value(int ti, int tj, float value)
{
  move[ti][tj].followup_value += value;
}

/*
 * Set a minimum allowed value for the move.
 */
void
set_minimum_move_value(int ti, int tj, float value)
{
  if (value > move[ti][tj].min_value)
    move[ti][tj].min_value = value;
}

/*
 * Set a maximum allowed value for the move.
 */
void
set_maximum_move_value(int ti, int tj, float value)
{
  if (value < move[ti][tj].max_value)
    move[ti][tj].max_value = value;
}

/*
 * Set a minimum allowed territorial value for the move.
 */
void
set_minimum_territorial_value(int ti, int tj, float value)
{
  if (value > move[ti][tj].min_territory)
    move[ti][tj].min_territory = value;
}

/*
 * Set a maximum allowed territorial value for the move.
 */
void
set_maximum_territorial_value(int ti, int tj, float value)
{
  if (value < move[ti][tj].max_territory)
    move[ti][tj].max_territory = value;
}

/* Test all moves that defends, attacks, connects or cuts to see if
 * they also attack or defend some other worm.
 *
 * FIXME: Before starting to read, we should see if the considered move
 *        already is listed in the move_reasons.
 *
 * FIXME: We would like to see whether an arbitrary move works to cut
 *        or connect something else too.
 */

static void
find_more_attack_and_defense_moves(int color)
{
  int unstable_worms[MAX_WORMS];
  int N = 0;  /* number of unstable worms */
  int m, n;
  int k;
  int other = OTHER_COLOR(color);

  if (hurry >= 3)
    return;

  TRACE("\nLooking for additional attack and defense moves\n");

  /* Identify the unstable worms and store them in a list. */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      if (p[m][n]
	  && worm[m][n].origini == m
	  && worm[m][n].originj == n
	  && worm[m][n].attacki != -1
	  && worm[m][n].defendi != -1) {
	unstable_worms[N] = find_worm(m, n);
	N++;
      }

  /* To avoid horizon effects, we temporarily increase the depth values. */
  increase_depth_values();
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      for (k=0; k<MAX_REASONS; k++) {
	int r = move[m][n].reason[k];
	int what;
	if (r < 0)
	  break;
	what = move_reasons[r].what;
	if ((   move_reasons[r].type == ATTACK_MOVE
		&& p[wormi[what]][wormj[what]] == other)
	    || (move_reasons[r].type == DEFEND_MOVE
		&& p[wormi[what]][wormj[what]] == color)
	    || move_reasons[r].type == CONNECT_MOVE
	    || move_reasons[r].type == CUT_MOVE
	    || move_reasons[r].type == ATTACK_EITHER_MOVE
	    || move_reasons[r].type == DEFEND_BOTH_MOVE)
	  break;
      }

      if (k<MAX_REASONS && move[m][n].reason[k] != -1) {
	/* Try the move at (m,n) and see what happens. */
	TRACE("trying %m\n",m ,n);
	if (trymove(m, n, color, "find_more_attack_and_defense_moves",
		    -1, -1)) {
	  for (k=0; k<N; k++) {
	    int ai = wormi[unstable_worms[k]];
	    int aj = wormj[unstable_worms[k]];
	    /* string of our color, see if there still is an attack,
	     * unless we already know the move works as defense move.
	     */
	    if (p[ai][aj] == color
		&& !move_reason_known(m, n, DEFEND_MOVE, unstable_worms[k]))
	      if (!attack(ai, aj, NULL, NULL)) {
		  TRACE("found extra point of defense of %m at %m\n",
			ai, aj, m, n);
		  add_defense_move(m, n, ai, aj);
	      }

	    /* string of opponent color, see if there still is a defense,
	     * unless we already know the move works as attack move.
	     */
	    if (p[ai][aj] == other
		&& !move_reason_known(m, n, ATTACK_MOVE, unstable_worms[k]))
	      if (!find_defense(ai, aj, NULL, NULL)) {
		/* Maybe find_defense() doesn't find the defense. Try to
		 * defend with the stored defense move.
		 */
		int attack_works = 1;
		if (trymove(worm[ai][aj].defendi, worm[ai][aj].defendj, other, 
			    "find_more_attack_and_defense_moves", -1, -1)) {
		  if (!attack(ai, aj, NULL, NULL))
		    attack_works = 0;
		  popgo();
		}
		
		if (attack_works) {
		  TRACE("found extra point of attack of %m at %m\n",
			ai, aj, m, n);
		  add_attack_move(m, n, ai, aj);
		}
	      }
	  }
	  popgo();
	}
      }
    }

  decrease_depth_values();
}

/*
 * Remove attacks on own stones and defense of opponent stones, i.e.
 * moves which are only relevant for the opponent.
 *
 * Comment: I once thought it would be useful to take these into
 * account (cf. the proverb "opponent's key points are your key
 * points") but now it seems they only lead to trouble. It's easiest
 * just to remove them altogether.
 *
 * FIXME: There is a certain amount of superfluous code that tests
 * against this case which now can be removed.
 */
static void
remove_opponent_attack_and_defense_moves(int color)
{
  int m, n, k;
  int ai, aj;
  int found_one;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      do {
	/* If we do remove something, we must restart the loop to be
         * sure that we look through all move_reasons.
	 */
	found_one = 0;
	for (k=0; k<MAX_REASONS; k++) {
	  int r = move[m][n].reason[k];
	  if (r < 0)
	    break;
	  ai = wormi[move_reasons[r].what];
	  aj = wormj[move_reasons[r].what];
	  if (move_reasons[r].type == ATTACK_MOVE
	      && p[ai][aj] == color) {
	    remove_move_reason(m, n, ATTACK_MOVE, move_reasons[r].what);
	    found_one = 1;
	    break;
	  }
	  else if (move_reasons[r].type == DEFEND_MOVE
		   && p[ai][aj] == OTHER_COLOR(color)) {
	    remove_move_reason(m, n, DEFEND_MOVE, move_reasons[r].what);
	    found_one = 1;
	    break;
	  }
	}
      } while (found_one);
    }
}

/*
 * Remove attacks and defenses that have earlier been marked as
 * NON_ATTACK or NON_DEFEND respectively, because they actually don't
 * work.
 */
static void
do_remove_false_attack_and_defense_moves(void)
{
  int m, n, k;
  int found_one;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      do {
	/* If we do remove something, we must restart the loop to be
         * sure that we look through all move_reasons.
	 */
	found_one = 0;
	for (k=0; k<MAX_REASONS; k++) {
	  int r = move[m][n].reason[k];
	  if (r < 0)
	    break;
	  if (move_reasons[r].type == NON_ATTACK_MOVE) {
	    int ai = wormi[move_reasons[r].what];
	    int aj = wormj[move_reasons[r].what];
	    remove_move_reason(m, n, ATTACK_MOVE, move_reasons[r].what);
	    /* Remove ourselves too. */
	    remove_move_reason(m, n, NON_ATTACK_MOVE, move_reasons[r].what);
	    /* If (m, n) was the attack point of the worm (ai, aj), we
	     * call change_attack to remove it.
	     *
	     * FIXME: We should look through the move reasons to see
	     * whether we could come up with an alternate attack point.
	     */
	    change_attack(ai, aj, -1, -1, 0);
	    /* If there was no attack, there is nothing to defend against. */
	    if (worm[ai][aj].defend_code != 0) {
	      remove_move_reason(m, n, DEFEND_MOVE, move_reasons[r].what);
	      change_defense(ai, aj, -1, -1, 0);
	    }
	    /* FIXME: We should also remove all defense reasons for
             * this worm.
	     */
	    found_one = 1;
	    break;
	  }
	  else if (move_reasons[r].type == NON_DEFEND_MOVE) {
	    int ai = wormi[move_reasons[r].what];
	    int aj = wormj[move_reasons[r].what];
	    remove_move_reason(m, n, DEFEND_MOVE, move_reasons[r].what);
	    /* Remove ourselves too. */
	    remove_move_reason(m, n, NON_DEFEND_MOVE, move_reasons[r].what);
	    /* If (m, n) was the defense point of the worm (ai, aj), we
	     * call change_defense to remove it.
	     *
	     * FIXME: We should look through the move reasons to see
	     * whether we could come up with an alternate defense point.
	     */
	    change_defense(ai, aj, -1, -1, 0);
	    found_one = 1;
	    break;
	  }
	}
      } while (found_one);
    }
}

/*
 * It's often bad to run away with a worm that is in a strategically
 * weak position. This function gives heuristics for determining
 * whether a move at (ti, tj) to defend the worm (ai, aj) is
 * strategically sound.
 *
 * FIXME: These heuristics need improvement. The biggest weakness is
 * that they sometimes fail to detect when we're running away towards
 * open ground. It would help much to have a reliable escape route
 * mechanism. 
 */
static int
strategically_sound_defense(int ai, int aj, int ti, int tj)
{
  int color = p[ai][aj];
  int i;
  /* If the number of liberties is more than doubled by the move we
   * guess that it's good in at least some respect.
   */
  if (accurate_approxlib(ti, tj, color, MAXLIBS, NULL, NULL)
      > 2*worm[ai][aj].liberties)
    return 1;

  if (mouse) {
    /* If the influence function does not think that we are running
     * into anyone's territory it's okay. The reason why we don't
     * accept running into our own territory is positions like
     * 
     * |XXXXX.
     * |OXOOXX
     * |OOOaOX
     * |...O.X
     * +------
     * 
     * where the tactical defense at 'a' is suicidal.
     */
    
    if (influence_territory_color(ti, tj) != OTHER_COLOR(color))
      return 1;

    /* If we connect to a living (or sufficiently healthy) dragon the
     * defense is okay. If the move attacks something it's also okay.
     * FIXME: To make this properly, we should verify that the attack
     *        really is effective.
     */
    
    for (i=0; i<MAX_REASONS; i++) {
      int rr = move[ti][tj].reason[i];
      int type;
      int what;
      if (rr == -1)
  	break;
      type = move_reasons[rr].type;
      what = move_reasons[rr].what;
      if ((type == ATTACK_MOVE
  	   && p[wormi[what]][wormj[what]] != color))
	return 1;
      if (type == CONNECT_MOVE) {
	int mouse_status = dragon[dragoni[what]][dragonj[what]].mouse_status;
	if (mouse_status == ALIVE || mouse_status == UNCHECKED)
	  return 1;
      }
    }

    /* If the move hasn't been accepted yet, we try to play it and ask
     * the mouse code whether the stones can live.
     */

    if (mouse_effective_defense(ai, aj, ti, tj))
      return 1;
  }
  else { /* !mouse */
    
    /* If the influence function does not think that we are running into
     * opponent territory it's okay.
     */
    if (influence_territory_color(ti, tj) != OTHER_COLOR(color))
      return 1;
  
    /* Otherwise we require that the defending move also attacks
     * something or connects to something.
     */
    for (i=0; i<MAX_REASONS; i++) {
      int rr = move[ti][tj].reason[i];
      int type;
      int what;
      if (rr == -1)
  	break;
      type = move_reasons[rr].type;
      what = move_reasons[rr].what;
      if ((type == ATTACK_MOVE
  	   && p[wormi[what]][wormj[what]] != color)
  	  || type == CONNECT_MOVE) {
  	return 1;
      }
    }
  }
  
  return 0;
}
	


/*
 * Any move that captures or defends a worm also connects or cuts
 * the surrounding dragons. Find these secondary move reasons.
 *
 * FIXME: There is a certain amount of optimizations that could be
 *        done here.
 * FIXME: Even when we defend a worm, it's possible that the opponent
 *        still can secure a connection, e.g. underneath a string with
 *        few liberties. Thus a defense move isn't necessarily a cut move.
 *
 * FIXME: Connections are transitive. If a move connects A with B and B
 *        with C, we should infer that it connects A with C as well.
 */
static int adji[MAXCHAIN];
static int adjj[MAXCHAIN];
static int adjsize[MAXCHAIN];
static int adjlib[MAXCHAIN];

static void
induce_secondary_move_reasons(int color)
{
  int m;
  int n;
  int k;
  int i;
  int neighbors;
  int ai;
  int aj;
  int di = -1;
  int dj = -1;
  int best_strength = 0;
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      for (k=0; k<MAX_REASONS; k++) {

	int r = move[m][n].reason[k];
	if (r < 0)
	  break;

	if (move_reasons[r].type != ATTACK_MOVE
	    && move_reasons[r].type != DEFEND_MOVE)
	  continue;

	ai = wormi[move_reasons[r].what];
	aj = wormj[move_reasons[r].what];

	if ((   (move_reasons[r].type == ATTACK_MOVE)
		&& (p[ai][aj] == color))
	    || ((move_reasons[r].type == DEFEND_MOVE)
		&& (p[ai][aj] != color)))
	  continue; /* Only a move for the opponent. */
	
	if (worm[ai][aj].defendi == -1)
	  continue; /* No defense. */

	/*
	 * If this is a defense move and the defense is futile for
         * strategical reasons, we shouldn't induce a cutting move
         * reason.
	 */
	if (move_reasons[r].type == DEFEND_MOVE
	    && !strategically_sound_defense(ai, aj, m, n))
	  continue;

	/* Find all dragons surrounding the worm. */
	chainlinks(ai, aj, &neighbors, adji, adjj, adjsize, adjlib);

	/*
	 * Find the strongest of the surrounding dragons and say that
	 * all other dragons are connected or cut with respect to that
	 * one.
	 *
	 * For the time being we use the vitality value of the dragon
	 * to measure strength. We probably want to change this later.
	 */
	for (i=0; i<neighbors; i++)
	  if (i == 0 || dragon[adji[i]][adjj[i]].vitality > best_strength) {
	    di = dragon[adji[i]][adjj[i]].origini;
	    dj = dragon[adji[i]][adjj[i]].originj;
	    best_strength = dragon[di][dj].vitality;
	  }

	for (i=0; i<neighbors; i++) {
	  int ei = dragon[adji[i]][adjj[i]].origini;
	  int ej = dragon[adji[i]][adjj[i]].originj;

	  if (di != ei || dj != ej) {
	    if (move_reasons[r].type == ATTACK_MOVE)
	      add_connection_move(m, n, di, dj, ei, ej);
	    else
	      add_cut_move(m, n, di, dj, ei, ej);
	  }
	}
      }
}

static void
list_move_reasons(int color)
{
  int m;
  int n;
  int k;
  int ai = -1;
  int aj = -1;
  int bi = -1;
  int bj = -1;
  int dragon1 = -1;
  int dragon2 = -1;
  int worm1 = -1;
  int worm2 = -1;
  int myworm = -1;
  int yourworm = -1;
  int ecolor = 0;

  gprintf("\nMove reasons:\n");
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      for (k=0; k<MAX_REASONS; k++) {

	int r = move[m][n].reason[k];
	if (r < 0)
	  break;

	switch(move_reasons[r].type) {
	  case ATTACK_MOVE:
	  case DEFEND_MOVE:
	    ai = wormi[move_reasons[r].what];
	    aj = wormj[move_reasons[r].what];

	    if (move_reasons[r].type == ATTACK_MOVE
		&& p[ai][aj] != color)
	      gprintf("Move at %m attacks %m%s\n", m, n, ai, aj,
		      (worm[ai][aj].defendi == -1) ? " (defenseless)" : "");
	    else if ((move_reasons[r].type == DEFEND_MOVE
		      && p[ai][aj] == color))
	      gprintf("Move at %m defends %m\n", m, n, ai, aj);
	    break;

	  case CONNECT_MOVE:
	  case CUT_MOVE:
	    dragon1 = conn_dragon1[move_reasons[r].what];
	    dragon2 = conn_dragon2[move_reasons[r].what];
	    ai = dragoni[dragon1];
	    aj = dragonj[dragon1];
	    bi = dragoni[dragon2];
	    bj = dragonj[dragon2];
	    if (move_reasons[r].type == CONNECT_MOVE)
	      gprintf("Move at %m connects %m and %m\n", m, n, ai, aj, bi, bj);
	    else
	      gprintf("Move at %m cuts %m and %m\n", m, n, ai, aj, bi, bj);
	    break;

	  case ANTISUJI_MOVE:
	    gprintf("Move at %m is an antisuji\n", m, n);
	    break;

	  case SEMEAI_MOVE:
	    myworm = semeai_myworm[move_reasons[r].what];
	    yourworm = semeai_yourworm[move_reasons[r].what];
	    ai = wormi[myworm];
	    aj = wormj[myworm];
	    bi = wormi[yourworm];
	    bj = wormj[yourworm];
	    gprintf("Move at %m urgent for semeai between %m and %m\n", m, n,
		    ai, aj, bi, bj);
	    break;

	  case VITAL_EYE_MOVE:
	    ai = eyei[move_reasons[r].what];
	    aj = eyej[move_reasons[r].what];
	    ecolor = eyecolor[move_reasons[r].what];
	    if (ecolor == WHITE)
	      gprintf("Move at %m vital eye point for dragon %m (eye %m)\n",
		      m, n, white_eye[ai][aj].dragoni,
		      white_eye[ai][aj].dragonj, ai, aj);
	    else
	      gprintf("Move at %m vital eye point for dragon %m (eye %m)\n",
		      m, n, black_eye[ai][aj].dragoni,
		      black_eye[ai][aj].dragonj, ai, aj);
	    break;

	  case NON_ATTACK_MOVE:
	  case NON_DEFEND_MOVE:
	    ai = wormi[move_reasons[r].what];
	    aj = wormj[move_reasons[r].what];

	    if (move_reasons[r].type == NON_ATTACK_MOVE
		&& p[ai][aj] != color)
	      gprintf("Move at %m does in fact not attack %m\n", m, n, ai, aj);
	    else if ((move_reasons[r].type == NON_DEFEND_MOVE
		      && p[ai][aj] == color))
	      gprintf("Move at %m does in fact not defend %m\n", m, n, ai, aj);
	    break;

	  case ATTACK_EITHER_MOVE:
	  case DEFEND_BOTH_MOVE:
	    worm1 = worm_pair1[move_reasons[r].what];
	    worm2 = worm_pair2[move_reasons[r].what];
	    ai = wormi[worm1];
	    aj = wormj[worm1];
	    bi = wormi[worm2];
	    bj = wormj[worm2];

	    if (move_reasons[r].type == ATTACK_EITHER_MOVE)
	      gprintf("Move at %m attacks either %m or %m\n",
		      m, n, ai, aj, bi, bj);
	    else
	      gprintf("Move at %m defends both %m and %m\n",
		      m, n, ai, aj, bi, bj);
	    break;

          case MOUSE_ATTACK_MOVE:
	    ai = dragoni[move_reasons[r].what];
	    aj = dragonj[move_reasons[r].what];
	    gprintf("Move at %m mouse-attacks %m\n", m, n, ai, aj);
	    break;

          case MOUSE_DEFEND_MOVE:
	    ai = dragoni[move_reasons[r].what];
	    aj = dragonj[move_reasons[r].what];
	    gprintf("Move at %m mouse-defends %m\n", m, n, ai, aj);
	    break;

	  case BLOCK_TERRITORY_MOVE:
	    gprintf("Move at %m blocks territory\n", m, n);
	    break;

	  case EXPAND_TERRITORY_MOVE:
	    gprintf("Move at %m expands territory\n", m, n);
	    break;

	  case EXPAND_MOYO_MOVE:
	    gprintf("Move at %m expands moyo\n", m, n);
	    break;

	  case STRATEGIC_ATTACK_MOVE:
	  case STRATEGIC_DEFEND_MOVE:
	    ai = dragoni[move_reasons[r].what];
	    aj = dragonj[move_reasons[r].what];

	    if (move_reasons[r].type == STRATEGIC_ATTACK_MOVE)
	      gprintf("Move at %m strategically attacks %m\n", m, n, ai, aj);
	    else
	      gprintf("Move at %m strategically defends %m\n", m, n, ai, aj);
	    break;

	}
      }
}

/*
 * Measure the "effective" size of a dragon. This measure is a
 * reasonable approximation of how much area a dragon cover, including
 * some amount of surrounding empty spaces.
 *
 * FIXME: Improve this further.
 */
static float effective_dragon_size(int ai, int aj)
{
  return dragon[ai][aj].effective_size;
}

/*
 * An attempt to estimate the safety of a dragon. This should be
 * possible to improve considerably.
 *
 * The resulting value is interpreted so that 1.0 means a fully safe
 * dragon while 0.0 is an almost dead dragon.
 */
static float
dragon_safety(int ai, int aj, int ignore_dead_dragons)
{
  int status = dragon[ai][aj].status;
  int vitality = dragon[ai][aj].vitality;
  float safety = 0.0;

  /* Kludge: If a dragon is dead, we return 1.0 in order not
   * to try to run away.
   */
  if (status == DEAD && ignore_dead_dragons)
    return 1.0;

  if (status == ALIVE)
    return 1.0;

  if (status == CRITICAL)
    return 0.0;

  if (status == UNKNOWN) {
    safety = 0.125 * vitality;
    if (dragon[ai][aj].safety == CRITICAL)
      safety -= 0.8;
  }

  if (safety > 1.0)
    safety = 1.0;
  if (safety < 0.0)
    safety = 0.0;

  return safety;
}

/*
 * Strategical value of connecting (or cutting) the dragon at (ai, aj)
 * to the dragon at (bi, bj). Notice that this function is assymetric.
 *
 * FIXME: We shouldn't care about connecting killing nakade stones to
 *        the outside.
 *
 * FIXME: This is all badly ad hoc. We certainly shouldn't rely this
 *        much on the vitality measure.
 */
static float
connection_value2(int ai, int aj, int bi, int bj, int ti, int tj)
{
  int status1 = dragon[ai][aj].status;
  int status2 = dragon[bi][bj].status;
  int vitality1 = dragon[ai][aj].vitality;
  int vitality2 = dragon[bi][bj].vitality;
  int halfeyes1 = 2 * dragon[ai][aj].genus + dragon[ai][aj].heyes;
  int halfeyes2 = 2 * dragon[bi][bj].genus + dragon[bi][bj].heyes;
  /* The difference between effective_size and size is a kind of
   * (generalized) liberty count.
   */
  float lib1 = dragon[ai][aj].effective_size - dragon[ai][aj].size;
  float lib2 = dragon[bi][bj].effective_size - dragon[bi][bj].size;
  float lib3 = 0;
  
  float dragon_size = effective_dragon_size(ai, aj);

  /* Relative safety of dragon before and after cut/connection */
  float before_safety = 0.0;
  float after_safety = 0.0;

  /* Play at (ti, tj) and count the number of liberties for (ai, aj)
   * after this move.
   */
  if (trymove(ti, tj, p[ai][aj], "connection_value2", ai, aj)) {
    if (p[ai][aj] == EMPTY)
      lib3 = 0;
    else
      lib3 = countlib(ai, aj);
    popgo();
  }
  
  /* Yet another kludge. */
  if (dragon[ai][aj].safety == CRITICAL)
    vitality1 -= 5;
  else
    vitality1 += 3;
  
  if (status1 == ALIVE) {
    if (vitality1 >= 10)
      before_safety = 1.0;
    else if (vitality1 <= 0)
      before_safety = 0.5;
    else
      before_safety = 0.5 + 0.05 * vitality1;

    after_safety = before_safety + 0.1 * halfeyes2;

    if (status2 != DEAD)
      after_safety += 0.1;

    if (status2 == ALIVE)
      after_safety = 1.0;

    if (after_safety > 1.0)
      after_safety = 1.0;
  }
  
  else if (status1 == CRITICAL) {
    before_safety = 0.0;
      
    if (status2 == ALIVE)
      after_safety = 1.0;
    else if (halfeyes2 > 0)
      after_safety = 1.0;
    else if (status2 != DEAD)
      after_safety = 0.5 + 0.05 * vitality2;
  }
    
  else if (status1 == UNKNOWN) {
    if (vitality1 >= 10)
      before_safety = 1.0;
    else if (vitality1 < -6)
      before_safety = 0.0;
    else
      before_safety = 0.0625 * (vitality1 + 6);

    after_safety = before_safety;
    
    if (status2 == ALIVE)
      after_safety += 0.5 + 0.05 * vitality2 + 0.1 * halfeyes2;
    else if (halfeyes1 + halfeyes2 >= 4)
      after_safety = 1.0;
    else if (halfeyes1 + halfeyes2 >= 3
	     && ((dragon[ai][aj].heyei == ti
		  && dragon[ai][aj].heyej == tj)
		 || (dragon[bi][bj].heyei == ti
		     && dragon[bi][bj].heyej == tj)))
      after_safety = 1.0;
    else if (status2 == DEAD)
      after_safety += 0.2 * halfeyes2;
    else {
      after_safety += 0.2 * halfeyes2;
      if (dragon[bi][bj].effective_size > 4.0)
	after_safety += 0.1;
      if (dragon[bi][bj].effective_size > 10.0)
	after_safety += 0.1;
      if (vitality2>0)
	after_safety += 0.05 * vitality2;

      if (lib2 > lib3) {
	if (lib1 < lib2)
	  after_safety += 0.2 * lib2/lib1;
      }
      else {
	if (lib1 < lib3)
	  after_safety += 0.2 * lib3/lib1;
      }
    }
  }
    
  else if (status1 == DEAD) {
    before_safety = 0.0;

    if (status2 == ALIVE)
      after_safety = 1.0;
    else if (halfeyes1 + halfeyes2 >= 4)
      after_safety = 1.0;
    else if (halfeyes1 + halfeyes2 >= 3
	     && ((dragon[ai][aj].heyei == ti
		  && dragon[ai][aj].heyej == tj)
		 || (dragon[bi][bj].heyei == ti
		     && dragon[bi][bj].heyej == tj)))
      after_safety = 1.0;
    else if (status2 != DEAD) {
      after_safety += 0.01 * halfeyes2;
      if (dragon[bi][bj].effective_size > 4.0)
	after_safety += 0.01;
      if (dragon[bi][bj].effective_size > 10.0)
	after_safety += 0.01;
      if (vitality2>0)
      after_safety += 0.05 * vitality2;
    }
  }

  /* If this dragon consists of only one worm and that worm can be
   * tactically captured or defended by this move, we have already
   * counted the points as territorial value, unless it's assumed to
   * be dead.
   */
  if (status1 != DEAD
      && dragon[ai][aj].size == worm[ai][aj].size
      && (move_reason_known(ti, tj, ATTACK_MOVE, find_worm(ai, aj))
	  || move_reason_known(ti, tj, DEFEND_MOVE, find_worm(ai, aj))))
    after_safety = 0.0;

  if (after_safety < before_safety)
    after_safety = before_safety;

  if (after_safety > 1.0)
    after_safety = 1.0;

  return 2.0 * (after_safety - before_safety) * dragon_size;
}


/*
 * Estimate the direct territorial value of a move at (m,n).
 */
static void
estimate_territorial_value(int m, int n, int color)
{
  int k;
  int ai = -1;
  int aj = -1;
  int bi = -1;
  int bj = -1;
  int myworm = -1;
  int yourworm = -1;
  int ecolor = 0;
  
  float this_value = 0.0;
  float tot_value = 0.0;
  float secondary_value = 0.0;

  int does_block = 0;
  int does_expand = 0;
  
  for (k=0; k<MAX_REASONS; k++) {
    int r = move[m][n].reason[k];
    if (r < 0)
      break;

    this_value = 0.0;
    switch (move_reasons[r].type) {
    case ATTACK_MOVE:
      /* FIXME: Need to take ko attacks properly into account. */
      ai = wormi[move_reasons[r].what];
      aj = wormj[move_reasons[r].what];
      
      /* Attack on our stones. */
      if (p[ai][aj] == color)
	break;
      
      /* Defenseless stone. */
      if (worm[ai][aj].defend_code == 0) {
	secondary_value += worm[ai][aj].size;
	break;
      }

      this_value = 2 * worm[ai][aj].effective_size;

      /* If the stones are dead, there is only a secondary value in
       * capturing them tactically as well.
       */
      if (dragon[ai][aj].status == DEAD) {
	secondary_value += 0.2 * this_value;
	break;
      }
      
      TRACE("  %m: %f - attack on worm %m\n",
	    m, n, this_value, ai, aj);
      
      tot_value += this_value;
      break;
      
    case DEFEND_MOVE:
      /* FIXME: Need to take ko defense properly into account. */
      ai = wormi[move_reasons[r].what];
      aj = wormj[move_reasons[r].what];
      
      /* Defense of enemy stones. */
      if (p[ai][aj] != color)
	break;
      
      /* Defenseless stone. */
      if (worm[ai][aj].defend_code == 0) {
	secondary_value += worm[ai][aj].size;
	break;
      }
      
      /* 
       * Estimate value 
       */
      if (strategically_sound_defense(ai, aj, m, n)) {
	this_value = 2 * worm[ai][aj].effective_size;

	/* If the stones are dead, we use the convention that
	 * defending them has a strategical value rather than
	 * territorial. Admittedly this make more sense for attacks on
	 * dead stones.
	 */
	if (dragon[ai][aj].status == DEAD) {
	  secondary_value += 0.2 * this_value;
	  break;
	}
      
	
	TRACE("  %m: %f - defense of worm %m\n",
	      m, n, this_value, ai, aj);
      }
      
      tot_value += this_value;
      break;

    case CONNECT_MOVE:
    case CUT_MOVE:
    case STRATEGIC_ATTACK_MOVE:
    case STRATEGIC_DEFEND_MOVE:
    case BLOCK_TERRITORY_MOVE:
      does_block = 1;
      break;
      
    case EXPAND_MOYO_MOVE:
    case EXPAND_TERRITORY_MOVE:
      does_expand = 1;
      break;
      
    case SEMEAI_MOVE:
      /*
       * The territorial value of winning a semeai is about twice
       * the sum of the effective sizes of the two worms plus other
       * adjoining dead worms. If the alternative to winning is
       * seki, the values are about half this size. The same goes
       * for making seki instead of losing the semeai.
       *
       * FIXME: Need to make a difference between the different
       *        kinds of possible semeai results. There is also
       *        some secondary value in playing an already won
       *        semeai to eliminate ko threats, and vice versa.
       *
       * FIXME: Valuation much too simplified.
       */
      myworm = semeai_myworm[move_reasons[r].what];
      yourworm = semeai_yourworm[move_reasons[r].what];
      ai = wormi[myworm];
      aj = wormj[myworm];
      bi = wormi[yourworm];
      bj = wormj[yourworm];
      this_value = 2 * (worm[ai][aj].effective_size
			+ worm[bi][bj].effective_size);
      TRACE("  %m: %f - semeai between %m and %m\n",
	    m, n, this_value, ai, aj, bi, bj);
      
      tot_value += this_value;
      break;
      
    case VITAL_EYE_MOVE:
      ai = eyei[move_reasons[r].what];
      aj = eyej[move_reasons[r].what];
      ecolor = eyecolor[move_reasons[r].what];
      
      if (ecolor == WHITE) {
	bi = white_eye[ai][aj].dragoni;
	bj = white_eye[ai][aj].dragonj;
      }
      else {
	bi = black_eye[ai][aj].dragoni;
	bj = black_eye[ai][aj].dragonj;
      }
      
      if (bi == -1) /* Maybe we should assert this not to happen. */
	break; 
      
      /* FIXME: These are ad hoc and need improvement. */
      /* It is not sufficient to test for status==CRITICAL, need
       * to count the eyes as well.
       */
      if (dragon[bi][bj].status == CRITICAL
	  && 2 * dragon[bi][bj].genus + dragon[bi][bj].heyes == 3) {
	this_value = 2 * dragon[bi][bj].effective_size;
	TRACE("  %m: %f - vital for %m\n",
	      m, n, this_value, bi, bj);
	tot_value += this_value;
      }
      else {
	if (2*dragon[bi][bj].genus + dragon[bi][bj].heyes < 6)
	  secondary_value += dragon[bi][bj].effective_size / 4;
	else
	  secondary_value += dragon[bi][bj].effective_size / 20;
      }
      break;

    case MOUSE_ATTACK_MOVE:
    case MOUSE_DEFEND_MOVE:
      ai = dragoni[move_reasons[r].what];
      aj = dragonj[move_reasons[r].what];
      this_value = 2 * dragon[ai][aj].effective_size;
      TRACE("  %m: %f - mouse attack/defend for %m\n",
	      m, n, this_value, ai, aj);
      tot_value += this_value;

      break;

    }
  }

  /* Currently no difference in the valuation between blocking and
   * expanding moves.
   */
  this_value = 0.0;
  if (does_block || does_expand) {
    this_value = influence_delta_territory(m, n, color);
    if (this_value != 0.0)
      TRACE("  %m: %f - block or expand territory\n", m, n, this_value);
  }

  /* Test if min_territory or max_territory values constrain the
   * delta_territory value.
   */
  if (this_value < move[m][n].min_territory && move[m][n].min_territory > 0) {
    this_value = move[m][n].min_territory;
    TRACE("  %m:   %f - revised to meet minimum territory value\n",
	  m, n, this_value);
  }
  if (this_value > move[m][n].max_territory) {
    this_value = move[m][n].max_territory;
    TRACE("  %m:   %f - revised to meet maximum territory value\n",
	  m, n, this_value);
  }
    
  tot_value += this_value;

  move[m][n].territorial_value = tot_value;
  move[m][n].secondary_value += secondary_value;
}


/*
 * Estimate the influence value of a move at (m,n).
 *
 * FIXME: This is much too simplified.
 */
static void
estimate_influence_value(int m, int n, int color)
{
  int k;
  int does_expand = 0;

  /* If the influence_value is non-zero, we have already computed it
   * once and don't have to redo it. This may happen when move values
   * are reevaluated after looking for endgame moves or after
   * revise_semeai().
   */
  if (move[m][n].influence_value != 0.0)
    return;
  
  for (k=0; k<MAX_REASONS; k++) {
    int r = move[m][n].reason[k];
    if (r < 0)
      break;

    switch (move_reasons[r].type) {
    case CONNECT_MOVE:
    case CUT_MOVE:
    case BLOCK_TERRITORY_MOVE:
    case EXPAND_TERRITORY_MOVE:
    case STRATEGIC_ATTACK_MOVE:
    case STRATEGIC_DEFEND_MOVE:
    case EXPAND_MOYO_MOVE:
      does_expand = 1;
      break;
    }
  }

  if (!does_expand)
    return;
  
  move[m][n].influence_value = 0.5 * influence_delta_strict_moyo(m, n, color);
  move[m][n].influence_value += 0.25 * influence_delta_strict_area(m, n, color);
  
  if (move[m][n].influence_value != 0.0)
    TRACE("  %m: %f - influence\n",
	  m, n, move[m][n].influence_value);
}

/*
 * Estimate the strategical value of a move at (m,n).
 */
static void
estimate_strategical_value(int m, int n, int color)
{
  int k;
  int l;
  int ai = -1;
  int aj = -1;
  int bi = -1;
  int bj = -1;
  int dragon1 = -1;
  int dragon2 = -1;
  int worm1 = -1;
  int worm2 = -1;
  int myworm = -1;
  int yourworm = -1;
  int ecolor = 0;
  
  float this_value = 0.0;
  float tot_value = 0.0;

  /* Strategical value of connecting or cutting dragons. */
  static float dragon_value[MAX_DRAGONS];

  for (k=0; k<next_dragon; k++)
    dragon_value[k] = 0.0;
  
  for (k=0; k<MAX_REASONS; k++) {
    int r = move[m][n].reason[k];
    if (r < 0)
      break;
    
    this_value = 0.0;
    switch (move_reasons[r].type) {
      case ATTACK_MOVE:
      case DEFEND_MOVE:
	worm1 = move_reasons[r].what;
	ai = wormi[worm1];
	aj = wormj[worm1];
      
	/* Attack on our stones. */
	if (move_reasons[r].type == ATTACK_MOVE
	    && p[ai][aj] == color)
	  break;
	
	/* Defense of enemy stones. */
	if (move_reasons[r].type == DEFEND_MOVE
	    && p[ai][aj] != color)
	  break;
	
	/* Defenseless stone */
	if (worm[ai][aj].defendi == -1)
	  break;
	
	/* This is totally ad hoc, just guessing the value of
         * potential cutting points.
	 */
	if (worm[ai][aj].cutstone2 > 1) {
	  this_value = 10.0 * (worm[ai][aj].cutstone2 - 1);
	  TRACE("  %m: %f - %m cutstone\n", m, n, 
		this_value, ai, aj);
	}
	
	tot_value += this_value;
	
	/* If the string is a lunch for a weak dragon, the attack or
         * defense has a strategical value. This can be valued along
	 * the same lines as strategic_attack/strategic_defend.
	 *
	 * FIXME: Need to estimate the safety after the move here too.
	 */

	/* Require the defense to be strategically viable. */
	if (move_reasons[r].type == DEFEND_MOVE
	    && !strategically_sound_defense(ai, aj, m, n))
	  break;
	
	/* Can't use k in this loop too. */
	for (l=0; l<next_lunch; l++)
	  if (lunch_worm[l] == worm1) {
	    dragon1 = lunch_dragon[l];
	    bi = dragoni[dragon1];
	    bj = dragonj[dragon1];

	    this_value = (dragon[bi][bj].effective_size
			  * (1.0 - dragon_safety(bi, bj, 0)));

	    /* If this dragon consists of only one worm and that worm
	     * can be tactically captured or defended by this move, we
	     * have already counted the points as territorial value,
	     * unless it's assumed to be dead.
	     */
	    if (dragon[bi][bj].status != DEAD
		&& dragon[bi][bj].size == worm[bi][bj].size
		&& (move_reason_known(m, n, ATTACK_MOVE, find_worm(bi, bj))
		    || move_reason_known(m, n, DEFEND_MOVE,
					 find_worm(bi, bj))))
	      this_value = 0.0;

	    if (this_value > dragon_value[dragon1])
	      dragon_value[dragon1] = this_value;
	  }

	break;
	
      case ATTACK_EITHER_MOVE:
      case DEFEND_BOTH_MOVE:
	/* This is complete nonsense, but still better than nothing.
	 * FIXME: Do this in a reasonable way.
	 */
	worm1 = worm_pair1[move_reasons[r].what];
	worm2 = worm_pair2[move_reasons[r].what];
	ai = wormi[worm1];
	aj = wormj[worm1];
	bi = wormi[worm2];
	bj = wormj[worm2];

	/* If we already have unconditional attack or defense of
         * either worm, this move reason has no additional value.
	 */
	if (move_reasons[r].type == ATTACK_EITHER_MOVE
	    && (move_reason_known(m, n, ATTACK_MOVE, worm1)
		|| move_reason_known(m, n, ATTACK_MOVE, worm2)))
	  break;
	if (move_reasons[r].type == DEFEND_BOTH_MOVE
	    && (move_reason_known(m, n, DEFEND_MOVE, worm1)
		|| move_reason_known(m, n, DEFEND_MOVE, worm2)))
	  break;
	
	
	this_value = 2 * min(worm[ai][aj].effective_size,
			     worm[bi][bj].effective_size);
	if (move_reasons[r].type == ATTACK_EITHER_MOVE)
	  TRACE("  %m: %f - attacks either %m or %m\n",
		m, n, this_value, ai, aj, bi, bj);
	else
	  TRACE("  %m: %f - defends both %m and %m\n",
		m, n, this_value, ai, aj, bi, bj);

	tot_value += this_value;
	break;
	
      case CONNECT_MOVE:
      case CUT_MOVE:
	dragon1 = conn_dragon1[move_reasons[r].what];
	dragon2 = conn_dragon2[move_reasons[r].what];
	ai = dragoni[dragon1];
	aj = dragonj[dragon1];
	bi = dragoni[dragon2];
	bj = dragonj[dragon2];

	this_value = connection_value2(ai, aj, bi, bj, m, n);
	if (this_value > dragon_value[dragon1])
	  dragon_value[dragon1] = this_value;
	
	this_value = connection_value2(bi, bj, ai, aj, m, n);
	if (this_value > dragon_value[dragon2])
	  dragon_value[dragon2] = this_value;
	
	break;
	
      case SEMEAI_MOVE:
	/*
	 * The strategical value of winning a semeai is that adjoining
	 * own dragons (usually) becomes fully secure, while adjoining
	 * enemy dragons do not.
	 *
	 * FIXME: Need to make a difference between the different
	 *        kinds of possible semeai results. 
	 *
	 * FIXME: Valuation not implemented at all yet. */
	myworm = semeai_myworm[move_reasons[r].what];
	yourworm = semeai_yourworm[move_reasons[r].what];
	ai = wormi[myworm];
	aj = wormj[myworm];
	bi = wormi[yourworm];
	bj = wormj[yourworm];
	break;
	
      case VITAL_EYE_MOVE:
	/*
	 * The value of the threatened group itself has already been
	 * accounted for in territiorial_value. Now we need to determine
	 * the effect this has on surrounding groups.
	 *
	 * FIXME: Valuation not implemented.
	 */
	ai = eyei[move_reasons[r].what];
	aj = eyej[move_reasons[r].what];
	ecolor = eyecolor[move_reasons[r].what];

	if (ecolor == WHITE) {
	  bi = white_eye[ai][aj].dragoni;
	  bj = white_eye[ai][aj].dragonj;
	}
	else {
	  bi = black_eye[ai][aj].dragoni;
	  bj = black_eye[ai][aj].dragonj;
	}

	if (bi == -1) /* Maybe we should assert this not to happen. */
	  break; 


#if 0
	if (dragon[bi][bj].status == CRITICAL) {
	  this_value = ???
	  TRACE("  %m: %f - vital for %m\n",
		m, n, this_value, bi, bj);
	  tot_value += this_value;
	}
#endif
	break;

      case STRATEGIC_ATTACK_MOVE:
      case STRATEGIC_DEFEND_MOVE:
	/* The right way to do this is to estimate the safety of the
	 * dragon before and after the move. Unfortunately we are
	 * missing good ways to do this currently.
	 *
	 * Temporary solution is to only look at an ad hoc measure of
	 * the dragon safety and ignoring the effectiveness of the
	 * move.
	 *
	 * FIXME: Improve the implementation.
	 */
	dragon1 = move_reasons[r].what;
	ai = dragoni[dragon1];
	aj = dragonj[dragon1];
	
	this_value = (dragon[ai][aj].effective_size
		      * (1.0 - dragon_safety(ai, aj, 1)));

	/* To prefer good connections adn cuts, we lower this value somewhat.
	 */
	this_value *= 0.75;

	/* If this dragon consists of only one worm and that worm can
	 * be tactically captured or defended by this move, we have
	 * already counted the points as territorial value, unless
	 * it's assumed to be dead.
	 */
	if (dragon[ai][aj].status != DEAD
	    && dragon[ai][aj].size == worm[ai][aj].size
	    && (move_reason_known(m, n, ATTACK_MOVE, find_worm(ai, aj))
		|| move_reason_known(m, n, DEFEND_MOVE,
				     find_worm(ai, aj))))
	  this_value = 0.0;
	
	if (this_value > dragon_value[dragon1])
	  dragon_value[dragon1] = this_value;

	break;
    }
  }
  
  for (k=0; k<next_dragon; k++) {
    if (dragon_value[k] > 0)
      TRACE("  %m: %f - strategic effect on %m\n",
	    m, n, dragon_value[k], dragoni[k], dragonj[k]);
    tot_value += dragon_value[k];
  }

  move[m][n].strategical_value = tot_value;
}

/* Look through the move reasons to see whether (m, n) is an antisuji move. */
static int
is_antisuji_move(int m, int n)
{
  int k;
  for (k=0; k<MAX_REASONS; k++) {
    int r = move[m][n].reason[k];
    if (r < 0)
      break;
    if (move_reasons[r].type == ANTISUJI_MOVE)
      return 1; /* This move must not be played. End of story. */
  }
  return 0;
}


/* Compare two move reasons, used for sorting before presentation. */
static int
compare_move_reasons(const void *p1, const void *p2)
{
  const int mr1 = *(const int *) p1;
  const int mr2 = *(const int *) p2;

  if (move_reasons[mr1].type != move_reasons[mr2].type)
    return move_reasons[mr2].type - move_reasons[mr1].type;
  else
    return move_reasons[mr2].what - move_reasons[mr1].what;
}


/*
 * Combine the reasons for a move at (m, n) into an old style value.
 * These heuristics are now somewhat less ad hoc but probably still
 * need a lot of improvement.
 */
static float
value_move_reasons(int m, int n, int color)
{
  float tot_value;


  /* Is it an antisuji? */
  if (is_antisuji_move(m, n))
    return 0.0; /* This move must not be played. End of story. */

  /* If this move has no reason at all, we can skip some steps. */
  if (move[m][n].reason[0] >= 0) {
    int num_reasons;
    /* Sort the move reasons. This makes it easier to visually compare
     * the reasons for different moves in the trace outputs.
     */
    num_reasons = 0;
    while (move[m][n].reason[num_reasons] >= 0)
      num_reasons++;
    qsort(&(move[m][n].reason[0]), num_reasons, sizeof(move[m][n].reason[0]),
	  compare_move_reasons);

    /* Estimate the value of various aspects of the move. */
    estimate_territorial_value(m, n, color);
    estimate_strategical_value(m, n, color);
    estimate_influence_value(m, n, color);
  }

  /* FIXME: The relative weights shouldn't be hardcoded like this. */
  tot_value = (move[m][n].territorial_value
	       + move[m][n].strategical_value
	       + move[m][n].influence_value);

  if (tot_value > 0.0) {
    tot_value += 0.5 * move[m][n].followup_value;
    if (move[m][n].followup_value != 0.0)
      TRACE("  %m: %f - followup\n",
	    m, n, move[m][n].followup_value);

    tot_value += 0.05 * move[m][n].secondary_value;
    if (move[m][n].secondary_value != 0.0)
      TRACE("  %m: %f - secondary\n",
	    m, n, 0.05 * move[m][n].secondary_value);

    if (move[m][n].shape_value != 0.0) {
    /* FIXME: Should use pow, but that requires linking with -lm */
#if 0
      float shape_factor = pow(1.05, move[m][n].shape_value);
#else
      float shape_factor = 1.0;
      int i;
      for (i=0; i<abs(move[m][n].shape_value) ;i++)
	shape_factor *= 1.05;
      if (move[m][n].shape_value < 0)
	shape_factor = 1.0/shape_factor;
#endif
      tot_value *= shape_factor;
      TRACE("  %m: %f - shape (shape value %f, shape factor %f)\n",
	    m, n, tot_value*(1.0-1.0/shape_factor),
	    move[m][n].shape_value, shape_factor);
    }
  }

  /* Test if min_value or max_value values constrain the total value. */
  if (tot_value < move[m][n].min_value && move[m][n].min_value > 0) {
    tot_value = move[m][n].min_value;
    TRACE("  %m:   %f - minimum accepted value\n",
	  m, n, tot_value);
  }
  if (tot_value > move[m][n].max_value) {
    tot_value = move[m][n].max_value;
    TRACE("  %m:   %f - maximum accepted value\n",
	  m, n, tot_value);
  }
    
  /* FIXME: Improve the trace outputs. */

  if (tot_value > 0
      || move[m][n].territorial_value > 0
      || move[m][n].strategical_value > 0) {
    TRACE("New move generation values %m to %f\n", 
	  m, n, tot_value);
    move_considered(m, n, (int) tot_value);
  }

  return tot_value;
}

/*
 * Loop over all possible moves and value the move reasons for each.
 */
static void
value_moves(int color)
{
  int m;
  int n;

  TRACE("\nMove valuation:\n");
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      move[m][n].value = value_move_reasons(m, n, color);
      if (move[m][n].value == 0.0)
	continue;
      /* Maybe this test should be performed elsewhere. This is just to get
       * some extra safety.
       */
      if (!legal(m, n, color)) {
	move[m][n].value = 0.0;
	TRACE("Move at %m wasn't legal.\n", m, n);
      }
      else {
	/* Add a random number between 0 and 0.01 to use in comparisons. */
	move[m][n].value += 0.01 * gg_drand();
      }
    }
}

/*
 * Review the move reasons to find which (if any) move we want to play.
 */
int
review_move_reasons(int *i, int *j, float *val, int color)
{
  int m;
  int n;
  float tval;
  float bestval;
  int good_move_found = 0;
  
  find_more_attack_and_defense_moves(color);
  remove_opponent_attack_and_defense_moves(color);
  do_remove_false_attack_and_defense_moves();
  induce_secondary_move_reasons(color);

  if (verbose)
    list_move_reasons(color);
  
  value_moves(color);

  while (!good_move_found) {
    bestval = 0.0;
    
    /* Search through all board positions for the highest valued move. */
    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++) {
	if (move[m][n].value == 0.0)
	  continue;
	
	tval = move[m][n].value;
	
	if (tval > bestval) {
	  bestval = tval;
	  *i = m;
	  *j = n;
	}
      }
    
    /* Call confirm_safety() to check that we're not about to make a
     * blunder. Otherwise reject this move and scan through all move
     * values once more.
     */
    if (bestval > 0.0
	&& !confirm_safety(*i, *j, color, (int) bestval/2, NULL, NULL)) {
      TRACE("Move at %m would be a blunder.\n", *i, *j);
      move[*i][*j].value = 0.0;
      good_move_found = 0;
    }
    else
      good_move_found = 1;
  }
  
  if (bestval > 0.0) {
    *val = bestval;
    return 1;
  }
  return 0;
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */

