/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "gnugo.h"
#include "interface.h"
#include "ttsgf.h"
#include "sgfana.h"

#define DEBUG_COMMANDS "\
capture <pos>    try to capture indicated group\n\
dead             Toggle display of dead stones\n\
defend <pos>     try to defend indicated group\n\
listdragons      print dragon info \n\
showarea         display area\n\
showdragons      display dragons\n\
showmoyo         display moyo\n\
showterri        display territory\n\
"


/* some options for the ascii interface */
static int opt_showboard = 1;
static int showscore = 0;
static int showdead = 0;
static int emacs = 0;
SGFNode *curnode = 0;
char *filename = NULL;
static int last_move_i;      /* The position of the last move */
static int last_move_j;      /* -""-                          */

static void endgame(Gameinfo *gameinfo);
static void nngs_endgame(Gameinfo *gameinfo);
static void showcapture(Position *pos, char *line);
static void showdefense(Position *pos, char *line);
static void ascii_goto(Gameinfo *gameinfo, char *line);
static int ascii2pos(Position *pos, char *line, int *i, int *j);

/* If sgf game info is written can't reset parameters like handicap, etc. */
static int sgf_initialized = 0;

/*
 * Create letterbar for the top and bottom of the ASCII board.
 */

static void
make_letterbar(int boardsize, char *letterbar)
{
  int i, letteroffset;
  char spaces[64];
  char letter[64];

  if (boardsize <= 25)
    strcpy(spaces, " ");
  strcpy(letterbar, "   ");
  
  for (i=0; i<boardsize; i++) {
    letteroffset = 'A';
    if (i+letteroffset >= 'I')
      letteroffset++;
    strcat(letterbar, spaces);
    sprintf(letter, "%c", i+letteroffset);
    strcat(letterbar, letter);
  }
}


/* This array contains +'s and -'s for the empty board positions.
 * hspot_size contains the board size that the grid has been
 * initialised to.
 */

static int hspot_size;
static char hspots[MAX_BOARD][MAX_BOARD];


/*
 * Mark the handicap spots on the board.
 */

static void
set_handicap_spots(int boardsize)
{
  if (hspot_size == boardsize)
    return;
  
  hspot_size = boardsize;
  
  memset(hspots, '.', sizeof(hspots));

  /* small sizes are easier to hardwire... */
  if (boardsize==2 || boardsize==4)
    return;
  if (boardsize==3) {
    /* just the middle one */
    hspots[boardsize/2][boardsize/2] = '+';
    return;
  }
  if (boardsize==5) {
    /* place the outer 4 */
    hspots[1][1] = '+';
    hspots[boardsize-2][1] = '+';
    hspots[1][boardsize-2] = '+';
    hspots[boardsize-2][boardsize-2] = '+';
    /* and the middle one */
    hspots[boardsize/2][boardsize/2] = '+';
    return;
  }

  if (!(boardsize%2)) {
    /* If the board size is even, no center handicap spots. */
    if (boardsize > 2 && boardsize < 12) {
      /* Place the outer 4 only. */
      hspots[2][2] = '+';
      hspots[boardsize-3][2] = '+';
      hspots[2][boardsize-3] = '+';
      hspots[boardsize-3][boardsize-3] = '+';
    }
    else {
      /* Place the outer 4 only. */
      hspots[3][3] = '+';
      hspots[boardsize-4][3] = '+';
      hspots[3][boardsize-4] = '+';
      hspots[boardsize-4][boardsize-4] = '+';
    }
  }
  else {
    /* Uneven board size */
    if (boardsize > 2 && boardsize < 12) {
      /* Place the outer 4... */
      hspots[2][2] = '+';
      hspots[boardsize-3][2] = '+';
      hspots[2][boardsize-3] = '+';
      hspots[boardsize-3][boardsize-3] = '+';

      /* ...and the middle one. */
      hspots[boardsize/2][boardsize/2] = '+';
    }
    else {
      /* Place the outer 4... */
      hspots[3][3] = '+';
      hspots[boardsize-4][3] = '+';
      hspots[3][boardsize-4] = '+';
      hspots[boardsize-4][boardsize-4] = '+';

      /* ...and the inner 4... */
      hspots[3][boardsize/2] = '+';
      hspots[boardsize/2][3] = '+';
      hspots[boardsize/2][boardsize-4] = '+';
      hspots[boardsize-4][boardsize/2] = '+';

      /* ...and the middle one. */
      hspots[boardsize/2][boardsize/2] = '+';
    }
  }

  return;
}


/*
 * Display the board position when playing in ASCII.
 */

static void
ascii_showboard(Position *pos)
{
   int i, j;
   char letterbar[64];
   int last_pos_was_move;
   int pos_is_move;
   int dead;
   
   make_letterbar(pos->boardsize, letterbar);
   set_handicap_spots(pos->boardsize);

   printf("\n");
   printf("    White has captured %d pieces\n", pos->black_captured);
   printf("    Black has captured %d pieces\n", pos->white_captured);
   if (showscore) {
     printf("    aprox. White territory: %d\n",
	    terri_eval[WHITE] - pos->black_captured);
     printf("    aprox. Black territory: %d\n",
	    terri_eval[BLACK] - pos->white_captured);
   }
   printf("\n");

   fflush (stdout);
   printf("%s%s\n", (emacs ? "EMACS1\n" : ""), letterbar);
   fflush (stdout);

   for (i = 0; i < pos->boardsize; i++) {
     printf(" %2d", pos->boardsize-i);
     last_pos_was_move = 0;
     for (j = 0; j < pos->boardsize; j++) {
       if ((last_move_i==i) && (last_move_j==j))
	 pos_is_move = 128;
       else
	 pos_is_move = 0;
       dead = (dragon_status(i, j)==DEAD) && showdead;
       switch (pos->board[i][j] + pos_is_move + last_pos_was_move) {
       case EMPTY+128:
       case EMPTY:
	 printf(" %c", hspots[i][j]);
	 last_pos_was_move = 0;
	 break;
       case BLACK:
	 printf(" %c", dead ? 'x' : 'X');
	 last_pos_was_move = 0;
	 break;
       case WHITE:
	 printf(" %c", dead ? 'o' : 'O');
	 last_pos_was_move = 0;
	 break;
       case BLACK+128:
	 printf("(%c)", dead ? 'X' : 'X');
	 last_pos_was_move = 256;
	 break; /* FIXME: Why have the newly placed stones the status dead? */
       case WHITE+128:
	 printf("(%c)", dead ? 'O' : 'O');
	 last_pos_was_move = 256;
	 break;
       case EMPTY+256:
	 printf("%c", hspots[i][j]);
	 last_pos_was_move = 0;
	 break;
       case BLACK+256:
	 printf("%c", dead ? 'x' : 'X');
	 last_pos_was_move = 0;
	 break;
       case WHITE+256:
	 printf("%c", dead ? 'o' : 'O');
	 last_pos_was_move = 0;
	 break;
       default: 
	 fprintf(stderr,"Illegal board value %d\n", pos->board[i][j]);
	 exit(EXIT_FAILURE);
	 break;
       }
     }

     if (last_pos_was_move==0) {
       if (pos->boardsize>10)
	 printf(" %2d", pos->boardsize-i);
       else
	 printf(" %1d", pos->boardsize-i);
     }
     else {
       if (pos->boardsize>10)
	 printf("%2d", pos->boardsize-i);
       else
         printf("%1d", pos->boardsize-i);
     }
     printf("\n");
   }

   fflush (stdout);
   printf("%s\n\n", letterbar);
   fflush (stdout);
   
}  /* end ascii_showboard */

/*
 * command help
 */

static void
show_commands(void)
{
  printf("\nCommands:\n");
  printf(" back             Take back your last move\n");
  printf(" boardsize        Set boardsize (on move 1 only!)\n");
  printf(" comment          Write a comment to outputfile\n");
  printf(" depth <num>      Set depth for reading\n");
  printf(" display          Display game board\n");
  printf(" exit             Exit GNU Go\n");
  printf(" force <move>     Force a move for current color\n");
  printf(" forward          Go to next node in game tree\n");
  printf(" goto <movenum>   Go to movenum in game tree\n");
  printf(" hurry <amount>   Play faster\n");
  printf(" handicap         Set handicap (on move 1 only!)\n");
  printf(" help             Display this help menu\n");
  printf(" helpdebug        Display debug help menu\n");
  printf(" info             Display program settings\n");
  printf(" komi             Set komi (on move 1 only!)\n");
  printf(" last             Goto last node in game tree\n");
  printf(" pass             Pass on your move\n");
  printf(" play <num>       Play <num> moves\n");
  printf(" playblack        Play as Black (switch if White)\n");
  printf(" playwhite        Play as White (switch if Black)\n");
  printf(" quit             Exit GNU Go\n");
  printf(" resign           Resign the current game\n");
  printf(" save <file>      Save the current game\n");
  printf(" score            Toggle display of score On/Off\n");
  printf(" showboard        Toggle display of board On/Off\n");
  printf(" switch           Switch the color you are playing\n");
  printf(" undo             Take the last move back (same as back)\n");
  printf(" <move>           A move of the format <letter><number>");
  printf("\n");
}

enum commands {INVALID=-1, END, EXIT, QUIT, RESIGN, 
	       PASS, MOVE, FORCE, SWITCH,
	       PLAY, PLAYBLACK, PLAYWHITE,
	       SETHANDICAP, SETBOARDSIZE, SETKOMI,
	       SETDEPTH,
               INFO, DISPLAY, SHOWBOARD, HELP, UNDO, COMMENT, SCORE,
               CMD_DEAD,CMD_BACK,CMD_FORWARD,CMD_LAST,
               CMD_CAPTURE,CMD_DEFEND,
               CMD_HELPDEBUG,CMD_SHOWAREA,CMD_SHOWMOYO,CMD_SHOWTERRI,
               CMD_GOTO,CMD_SAVE,CMD_SHOWDRAGONS,CMD_LISTDRAGONS,
	       NNGS, SETHURRY
};


/*
 * Check if we have a valid command.
 */

static int
get_command(char *command)
{
  char c;
  int d;

  /* Check to see if a move was input. */
  if (!((sscanf(command, "%c%d", &c, &d) != 2)
	|| ((c=toupper(c)) < 'A')
	|| ((c=toupper(c)) > 'Z')
	|| (c == 'I')))
    return MOVE;

  /* help shortcut */
  if (command[0]=='?')
    return HELP;

  /* Kill leading spaces. */
  while (command[0]==' ')
    command++;

  if (!strncmp(command, "playblack", 9)) return PLAYBLACK;
  if (!strncmp(command, "playwhite", 9)) return PLAYWHITE;
  if (!strncmp(command, "showboard", 9)) return SHOWBOARD;
  if (!strncmp(command, "showdragons", 9)) return CMD_SHOWDRAGONS;
  if (!strncmp(command, "listdragons", 9)) return CMD_LISTDRAGONS;
  if (!strncmp(command, "boardsize", 9)) return SETBOARDSIZE;
  if (!strncmp(command, "handicap", 8)) return SETHANDICAP;
  if (!strncmp(command, "display", 7)) return DISPLAY;
  if (!strncmp(command, "helpdebug", 7)) return CMD_HELPDEBUG;
  if (!strncmp(command, "resign", 6)) return RESIGN;
  if (!strncmp(command, "showmoyo", 6)) return CMD_SHOWMOYO;
  if (!strncmp(command, "showterri", 6)) return CMD_SHOWTERRI;
  if (!strncmp(command, "showarea", 6)) return CMD_SHOWAREA;
  if (!strncmp(command, "depth", 5)) return SETDEPTH;
  if (!strncmp(command, "switch", 5)) return SWITCH;
  if (!strncmp(command, "komi", 4)) return SETKOMI;
  if (!strncmp(command, "play", 4)) return PLAY;
  if (!strncmp(command, "info", 4)) return INFO;
  if (!strncmp(command, "force", 4)) return FORCE;
  if (!strncmp(command, "hurry", 5)) return SETHURRY;
  if (!strncmp(command, "pass", 4)) return PASS;
  if (!strncmp(command, "nngs", 4)) return NNGS;
  if (!strncmp(command, "save", 3)) return CMD_SAVE;
  if (!strncmp(command, "end", 3)) return END;
  if (!strncmp(command, "move", 3)) return MOVE;
  if (!strncmp(command, "undo", 3)) return UNDO;
  if (!strncmp(command, "comment", 3)) return COMMENT;
  if (!strncmp(command, "score", 3)) return SCORE;
  if (!strncmp(command, "dead", 3)) return CMD_DEAD;
  if (!strncmp(command, "capture", 3)) return CMD_CAPTURE;
  if (!strncmp(command, "defend", 3)) return CMD_DEFEND;
  if (!strncmp(command, "exit", 2)) return EXIT;
  if (!strncmp(command, "quit", 1)) return QUIT;
  if (!strncmp(command, "help", 1)) return HELP;
  if (!strncmp(command, "back", 1)) return CMD_BACK;
  if (!strncmp(command, "forward", 1)) return CMD_FORWARD;
  if (!strncmp(command, "last", 1)) return CMD_LAST;
  if (!strncmp(command, "goto", 1)) return CMD_GOTO;


  /* No valid command was found. */
  return INVALID;
}


/*
 * Write sgf root node. 
 */

static void
init_sgf(Gameinfo *ginfo)
{
  if (sgf_initialized)
    return;
  sgf_initialized = 1;
  sgffile_write_gameinfo(ginfo, "ascii");
  if (ginfo->handicap > 0)
    position_recordhand(&ginfo->position, !filename);
  else
    ginfo->to_move = BLACK;
}


/*
 * Generate the computer move. 
 */

static void
computer_move(Gameinfo *gameinfo, int *pass)
{
  int i, j;
  int move_val;

  init_sgf(gameinfo);
  
  /* Generate computer move. */
  move_val = get_move(gameinfo, &i, &j, gameinfo->to_move);  
  last_move_i = i;
  last_move_j = j;
  curnode = sgfAddPlay(curnode, gameinfo->to_move, i, j);
  
  if (move_val) {
    /* An ordinary move */
    if (gameinfo->to_move==WHITE)
      mprintf("White(%d): %m\n", gameinfo->move_number, i, j);
    if (gameinfo->to_move==BLACK)
      mprintf("Black(%d): %m\n", gameinfo->move_number, i, j);
    *pass = 0;
  }
  else {
    if (gameinfo->to_move==WHITE)
      mprintf("White(%d): Pass\n", gameinfo->move_number);
    if (gameinfo->to_move==BLACK)
      mprintf("Black(%d): Pass\n", gameinfo->move_number);
    (*pass)++;
  }

  gameinfo->to_move = gameinfo->to_move == BLACK ? WHITE : BLACK;
}


/*
 * Make a move.
 */

static void
do_move(Gameinfo *gameinfo, char *command, int *pass, int force)
{
  char c;
  int d, i, j;

  sscanf(command, "%c%d", &c, &d);
  /* Got a move... check if it's legal. */
  /* Compensate for no 'I' on the board. */
  c = toupper(c);
  if ((c - 'J') > (gameinfo->position.boardsize - 9)
      || (d <= 0 || d > gameinfo->position.boardsize)) {
    printf("\nInvalid move: %c%d\n", c, d);
    return;
  }
  
  if (c > 'I')
    c--;
  i = gameinfo->position.boardsize - d;
  j = c - 'A';
  
  if (!position_legal(&gameinfo->position, i, j, gameinfo->to_move)) {
    if (c >= ('I'))
      c++;
    printf("\nIllegal move: %c%d\n", c, d);
    return;
  }
  else {
    *pass = 0;
    gameinfo->move_number++;
    TRACE("\nyour move: %m\n\n", i, j);
    position_play_move(&gameinfo->position, i, j, gameinfo->to_move);
    last_move_i = i;
    last_move_j = j;
    /* FIXME: This call to init_sgf should not be here. */
    init_sgf(gameinfo);
    sgffile_move_made(i, j, gameinfo->to_move, 0);
    curnode = sgfAddPlay(curnode, gameinfo->to_move, i, j);
  }
  
  if (opt_showboard && !emacs) {
    ascii_showboard(&gameinfo->position);
    printf("GNU Go is thinking...\n");
  }
  if (force) {
    gameinfo->computer_player = OTHER_COLOR(gameinfo->computer_player);
    gameinfo->to_move = OTHER_COLOR(gameinfo->to_move);
    return;
  }
  gameinfo->to_move = OTHER_COLOR(gameinfo->to_move);
  computer_move(gameinfo, pass);
}


/*
 * Make a pass.
 */

static void
do_pass(Gameinfo *gameinfo, int *pass, int force)
{
  (*pass)++;
  position_play_move(&gameinfo->position, -1, -1, gameinfo->to_move);
  last_move_i = -1;
  last_move_j = -1;
  gameinfo->move_number++;
  init_sgf(gameinfo);
  sgffile_move_made(-1, -1, gameinfo->to_move, 0);
  curnode = sgfAddPlay(curnode, gameinfo->to_move, -1, -1);

  if (force) {
    gameinfo->computer_player = OTHER_COLOR(gameinfo->computer_player);
    gameinfo->to_move = OTHER_COLOR(gameinfo->to_move);
    return;
  }
  gameinfo->to_move = OTHER_COLOR(gameinfo->to_move);
  computer_move(gameinfo, pass);
}


/*
 * Play a game against an ascii client.
 */

void
play_ascii(Gameinfo *gameinfo, char *fname, char *until)
{
  int m, num, sz;
  float fnum;
  int pass = 0;  /* two passes and its over */
  int tmp;
  char line[80];
  char *line_ptr = line;
  char *command;
  char *tmpstring;
  int nngs = 0; /* NNGS style customization */
  
  filename = fname;

#ifdef HAVE_SETLINEBUF
  setlinebuf(stdout); /* Need at least line buffer NNGS and gnugo-gnugo */
#else
  setbuf(stdout, NULL); /* else set it to completely UNBUFFERED */
#endif
  
  /* Start the move numbering at 0... */
  /* when playing with gmp it starts at 0 /NE */
  gameinfo->move_number = 0;

  if (filename) {
    load_sgf_header(sgf_root, gameinfo);
    sgffile_write_gameinfo(gameinfo, "ascii");
    load_sgf_file(sgf_root, gameinfo, until);
    sgf_initialized = 1;
    curnode = sgfNodeCheck(0);
  }
  else {
    if (sgfGetIntProperty(sgf_root, "SZ", &sz)) {
      position_clear(&gameinfo->position, sz);
    }
    if (gameinfo->handicap > 0) {
      gameinfo->handicap = position_placehand(&gameinfo->position, 
					     gameinfo->handicap);
      gameinfo->to_move = WHITE;
    }
    else
      gameinfo->to_move = BLACK;

    curnode = sgf_root;
  }

  printf("\nBeginning ASCII mode game.\n\n");
  ginfo_print(gameinfo);
  
  /* Does the computer play first?  If so, make a move. */
  if (gameinfo->computer_player == gameinfo->to_move)
    computer_move(gameinfo, &pass);
  
  /* main ASCII Play loop */
  while (pass < 2 && !time_to_die) {
    /* Display game board. */
    if (opt_showboard)
      ascii_showboard(&gameinfo->position);

    /* Print the prompt */
    if (gameinfo->to_move == WHITE)
      printf("White(%d): ", gameinfo->move_number+1);
    if (gameinfo->to_move == BLACK)
      printf("Black(%d): ", gameinfo->move_number+1);
      
    /* Read a line of input. */
    line_ptr = line;
    if (!fgets(line, 80, stdin))
      break; /* EOF or some error */
      
    while (command = strtok(line_ptr, ";"), line_ptr = 0, command) {
      
      /* Get the command or move. */
      switch (get_command(command)) {
      case RESIGN:
	printf("\nGNU Go wins by resignation.");
	sgfWriteResult(gameinfo->to_move==BLACK ? (float) -1000 : (float) 1000);
      case END:
      case EXIT:
      case QUIT:
	printf("\nThanks for playing GNU Go.\n\n");
	return ;
	break;
      case HELP:
	show_commands();
	break;
      case CMD_HELPDEBUG:
	printf(DEBUG_COMMANDS);
	break;
      case SHOWBOARD:
	opt_showboard = !opt_showboard;
	break;
      case INFO:
	printf("\n");
	ginfo_print(gameinfo);
	break;
      case SETBOARDSIZE:
	if (gameinfo->move_number > 0) {
	  printf("Boardsize can be modified on move 1 only!\n");
	  break;
	}
	if (sgf_initialized) {
	  printf("Boardsize cannot be modified after game record is started!\n");
	  break;
	}
	command += 10;
	if (sscanf(command, "%d", &num) != 1) {
	  printf("\nInvalid command syntax!\n");
	  break;
	}
	if (num<2 || num>25) {
	  printf("\nInvalid board size: %d\n", num);
	  break;
	}
	/* Init board. */
	position_clear(&gameinfo->position, num);
	/* In case max handicap changes on smaller board. */
	gameinfo->handicap = position_placehand(&gameinfo->position, 
					       gameinfo->handicap);
	break;
      case SETHANDICAP:
	if (gameinfo->move_number > 0) {
	  printf("Handicap can be modified on move 1 only!\n");
	  break;
	}
	if (sgf_initialized) {
	  printf("Handicap cannot be modified after game record is started!\n");
	  break;
	}
	command += 9;
	if (sscanf(command, "%d", &num) != 1) {
	  printf("\nInvalid command syntax!\n");
	  break;
	}
	if (num<0 || num>MAX_HANDICAP) {
	  printf("\nInvalid handicap: %d\n", num);
	  break;
	}
	/* Init board. */
	position_clear(&gameinfo->position, gameinfo->position.boardsize);
	/* place stones on board but don't record sgf in case we change more info */
	gameinfo->handicap = position_placehand(&gameinfo->position, num);
	printf("\nSet handicap to %d\n", gameinfo->handicap);
	gameinfo->to_move = WHITE;
	break;
      case SETKOMI:
	if (gameinfo->move_number > 0) {
	  printf("Komi can be modified on move 1 only!\n");
	  break;
	}
	if (sgf_initialized) {
	  printf("Komi cannot be modified after game record is started!\n");
	  break;
	}
	command += 5;
	if (sscanf(command, "%f", &fnum) != 1) {
	  printf("\nInvalid command syntax!\n");
	  break;
	}
	gameinfo->komi = fnum;
	printf("\nSet Komi to %.1f\n", fnum);
	break;
	
      case SETDEPTH:
	{
	  command += 6;
	  if (sscanf(command, "%d", &num) != 1) {
	    printf("\nInvalid command syntax!\n");
	    break;
	  }
	  nominal_depth = num;
	  printf("\nSet depth to %d\n", nominal_depth);
	  break;
	}
      
      case SETHURRY:
	{
	  command += 6;
	  if (sscanf(command, "%d", &num) != 1) {
	    printf("\nInvalid command syntax!\n");
	    break;
	  }
	  hurry = num;
	  printf("\nSet hurry to %d\n", hurry);
	  break;
	}
      
      case DISPLAY:
	if (!opt_showboard)
	  ascii_showboard(&gameinfo->position);
	break;
      case FORCE:
	command += 6; /* skip the force part... */
	switch (get_command(command)) {
	case MOVE:
	  do_move(gameinfo, command, &pass, 1);
	  break;
	case PASS:
	  do_pass(gameinfo, &pass, 1);
	  break;
	default:
	  printf("Illegal forced move: %s %d\n", command,
		 get_command(command));
	  break;
	}
	break;
      case MOVE:
	do_move(gameinfo, command, &pass, 0);
	break;
      case PASS:
	do_pass(gameinfo, &pass, 0);
	break;
      case PLAY:
	command += 5;
	if (sscanf(command, "%d", &num) != 1) {
	  printf("\nInvalid command syntax!\n");
	  break;
	}
	if (num >= 0)
	  for (m=0; m<num; m++) {
	    gameinfo->computer_player = OTHER_COLOR(gameinfo->computer_player);
	    computer_move(gameinfo, &pass);
	    if (pass >= 2)
	      break;
	  }
	else {
	  printf("\nInvalid number of moves specified: %d\n", num);
	  break;
	}
	break;
      case PLAYBLACK:
	if (gameinfo->computer_player == WHITE)
	  gameinfo->computer_player = BLACK;
	if (gameinfo->computer_player == gameinfo->to_move)
	  computer_move(gameinfo, &pass);
	break;
      case PLAYWHITE:
	if (gameinfo->computer_player == BLACK)
	  gameinfo->computer_player = WHITE;
	if (gameinfo->computer_player == gameinfo->to_move)
	  computer_move(gameinfo, &pass);
	break;
      case SWITCH:
	gameinfo->computer_player = OTHER_COLOR(gameinfo->computer_player);
	computer_move(gameinfo, &pass);
	break;
	/*FIXME: Don't forget boardupdate after back! */
      case UNDO:
      case CMD_BACK:
	/*reloading the game*/
	if (gameinfo->move_number > 0) {
	  curnode = sgf_root;
	  gameinfo->to_move = sgfPlayTree(&curnode, &gameinfo->position, 
					  &gameinfo->move_number);
	  gameinfo->move_number--;
	}
	else
	  printf("\nBeginning of game.\n");
        break;
      case CMD_FORWARD:
	if (curnode->child) {
	  gameinfo->to_move = sgfAddMoveFromTree(curnode->child,
						 gameinfo, 
						 gameinfo->to_move);
	  curnode = curnode->child;
	}
	else
	  printf("\nEnd of game tree.\n");
	break;
      case CMD_LAST:
	while (curnode->child) {
	  gameinfo->to_move = sgfAddMoveFromTree(curnode->child, 
						 gameinfo, gameinfo->to_move);
	  curnode = curnode->child;
	}
	break;
      case COMMENT:
	printf("\nEnter comment. Press ENTER when ready.\n");
	fgets(line, 80, stdin);
	sgfAddComment(curnode, line);
	break;
      case SCORE:
	showscore = !showscore;
	break;
      case CMD_DEAD:
	showdead = !showdead;
	break;
      case CMD_CAPTURE:
	strtok(command, " ");
	showcapture(&gameinfo->position, strtok(0, " "));
	break;
      case CMD_DEFEND:
	strtok(command, " ");
	showdefense(&gameinfo->position, strtok(0, " "));
	break;
      case CMD_SHOWMOYO:
	tmp = printmoyo;
	printmoyo = 4;
	/* FIXME: Use examine_position(). */
	make_worms();
	make_dragons(EMPTY);
	make_moyo(gameinfo->to_move);
	print_moyo(0);
	printmoyo = tmp;
	break;
      case CMD_SHOWTERRI:
	tmp = printmoyo;
	printmoyo = 1;
	/* FIXME: Use examine_position(). */
	make_worms();
	make_dragons(EMPTY);
	make_moyo(gameinfo->to_move);
	print_moyo(0);
	printmoyo = tmp;
	break;
      case CMD_SHOWAREA:
	tmp = printmoyo;
	printmoyo = 16;
	/* FIXME: Use examine_position(). */
	make_worms();
	make_dragons(EMPTY);
	make_moyo(gameinfo->to_move);
	print_moyo(0);
	printmoyo = tmp;
	break;
      case CMD_SHOWDRAGONS:
	/* FIXME: Use examine_position(). */
	make_worms();
	make_dragons(EMPTY);
	showboard(1);
	break;
      case CMD_GOTO:
	strtok(command, " ");
	ascii_goto(gameinfo, strtok(0, " "));
	break;
      case CMD_SAVE:
	strtok(command, " ");
	tmpstring = strtok(0, " ");
	if (tmpstring) {
	  tmpstring[strlen(tmpstring)-1] = 0;
	  sgf_write_header(sgf_root, analyzerflag & ANALYZE_OVERWRITE,
			   gameinfo->seed, gameinfo->komi);
	  writesgf(sgf_root, analyzerfile);
	}
	printf("You may resume the game with -l [filename] --mode ascii\n");
	break;
      case CMD_LISTDRAGONS:
	/* FIXME: Use examine_position(). */
	make_worms();
	make_dragons(EMPTY);
	show_dragons();
	break;
      case NNGS:
	/* Setting the nngs flag will customize the ascii interface
	 * for use by the NNGS connecting program. This mode of
	 * operation is unlikely to be useful for other purposes.
	 */
	nngs = 1;
	break;
      case INVALID:
      default:
	printf("\nInvalid command: %s",command);
	break;
      }
    }
  }

  /* two passes : game over */

  if (pass >= 2) {
    printf("My final evaluation of the game :\n");
    position_who_wins(&gameinfo->position, gameinfo->computer_player,
		      gameinfo->komi, stdout);
    if (nngs)
      nngs_endgame(gameinfo);
    else
      endgame(gameinfo);
  }

  printf("\nThanks for playing GNU Go.\n\n");
}

void
play_ascii_emacs(Gameinfo *gameinfo, char *fname, char *until)
{
  emacs=1;
  play_ascii(gameinfo, fname, until);
}


/*
 * endgame() scores the game.
 */

static void
endgame(Gameinfo *gameinfo)
{
  char line[12];
  int done = 0;
  int d, i, j;
  char c;
  float score;
  int black_points = 0;
  int white_points = 0;
  int xyzzy = 1;
  float komi;

  komi = gameinfo->komi;

  printf("\nGame over. Let's count!.\n");  
  showdead = 1;
  ascii_showboard(&gameinfo->position);
  while (!done) {
    printf("Dead stones are marked with small letters (x,o).\n");
    printf("\nIf you don't agree, enter the location of a group\n\
to toggle its state or \"done\".\n");

    if (!fgets(line, 12, stdin))
      return; /* EOF or some error */
    
    for (i=0; i<12; i++)
      line[i] = (isupper ((int) line[i]) ? tolower (line[i]) : line[i]);
    if (!strncmp(line, "done", 4))
      done = 1;
    else if (!strncmp(line, "quit", 4))
      return;
    else if (!strncmp(line, "xyzzy", 5)) {
      if (xyzzy) {
	printf("\nYou are in a debris room filled with stuff washed in from the\n");
	printf("surface.  A low wide passage with cobbles becomes plugged\n");
	printf("with mud and debris here, but an awkward canyon leads\n");
	printf("upward and west.  A note on the wall says:\n");
	printf("   Magic Word \"XYZZY\"\n\n");
	xyzzy = 0;
      }
      else {
	printf("You are inside a building, a well house for a large spring.\n\n");
	xyzzy = 1;
      }
    }
    else if (!strncmp(line, "help", 4)) {
      printf("\nIf there are dead stones on the board I will remove them.\n");
      printf("You must tell me where they are. Type the coordinates of a\n");
      printf("dead group, or type \"done\"\n");
    }      
    else if (!strncmp(line, "undo", 4)) {
      printf("UNDO not allowed anymore. The status of the stones now toggles after entering the location of it.\n");
      ascii_showboard(&gameinfo->position);
    }
    else {
      sscanf(line, "%c%d", &c, &d);
      c = toupper(c);
      if (c > 'I')
	c--;
      i = gameinfo->position.boardsize - d;
      j = c - 'A';
      if ((i < 0) || (i >= gameinfo->position.boardsize)
	  || (j < 0) || (j >= gameinfo->position.boardsize) || 
	  (gameinfo->position.board[i][j] == EMPTY))
	printf("\ninvalid!\n");
      else {
	int status = dragon_status(i, j);
	status = (status == DEAD) ? ALIVE : DEAD;
	change_dragon_status(i, j, status);
	/* remove_string(i, j); */
	ascii_showboard(&gameinfo->position);
      }
    }
  }

  /* count_territory(&white_points, &black_points); */

  position_evaluate_territory(&gameinfo->position,
			      &white_points, &black_points);

  printf("\n\nBlack has %d points and has captured %d white stones\n",
	 black_points, gameinfo->position.white_captured);
  printf("White has %d points and has captured %d black stones\n",
	 white_points, gameinfo->position.black_captured);
  printf("komi is %.1f.\n", komi);

  score = (black_points - white_points 
	   - gameinfo->position.black_captured
	   + gameinfo->position.white_captured
	   - komi);
  sgfShowTerritory(0);
  sgfWriteResult(score);

  if (score>0) 
    printf("\nBLACK WINS by %.1f points\n", score);
  else if (score<0)
    printf("\nWHITE WINS by %.1f points\n", -score);    
  else
    printf("\nTIED GAME!\n");
}


/*
 * nngs_endgame() records the strings the opponent says are dead and
 * compares this to our own opinion.
 */
/* FIXME: Make some kind of position_examine_position and use that instead. */
static void
nngs_endgame(Gameinfo *gameinfo)
{
  char line[12];
  int d, i, j;
  char c;
  int worm_status[MAX_BOARD][MAX_BOARD];
  int m, n;
  static const char *snames[] = 
    {"dead", "alive", "critical", "unknown"};
  
  for (m=0; m<gameinfo->position.boardsize; m++)
    for (n=0; n<gameinfo->position.boardsize; n++)
      worm_status[m][n] = ALIVE;
  
  while (1) {
    printf("Bring out your dead!\n");

    if (!fgets(line, 12, stdin))
      return; /* EOF or some error */
    
    for (i=0; i<12; i++)
      line[i] = (isupper ((int) line[i]) ? tolower (line[i]) : line[i]);

    if (!strncmp(line, "done", 4))
      break;
    else if (!strncmp(line, "quit", 4))
      return;
    else if (!strncmp(line, "pass", 4))  /* ignore */
      continue;
    else if (!strncmp(line, "undo", 4)) {
      for (m=0; m<gameinfo->position.boardsize; m++)
	for (n=0; n<gameinfo->position.boardsize; n++)
	  worm_status[m][n] = ALIVE;
    }
    else {
      sscanf(line, "%c%d", &c, &d);
      c = toupper(c);
      if (c > 'I')
	c--;
      i = gameinfo->position.boardsize - d;
      j = c - 'A';
      if ((i < 0)
	  || (i >= gameinfo->position.boardsize)
	  || (j < 0)
	  || (j >= gameinfo->position.boardsize)
	  || (gameinfo->position.board[i][j] == EMPTY))
	printf("\ninvalid!\n");
      else {
	for (m=0; m<gameinfo->position.boardsize; m++)
	  for (n=0; n<gameinfo->position.boardsize; n++)
	    if (same_worm(m, n, i, j))
	      worm_status[m][n] = DEAD;
      }
    }
  }

  for (m=0; m<gameinfo->position.boardsize; m++)
    for (n=0; n<gameinfo->position.boardsize; n++)
      if (is_worm_origin(m, n, m, n)
	  && gameinfo->position.board[m][n] != EMPTY) {
	if (worm_status[m][n] != dragon_status(m, n)) {
	  char buf[120];
	  sprintf(buf,
		  "Disagreement over %c%d: GNU Go says %s, opponent says %s.\n",
 		  'A'+n+(n>=8), gameinfo->position.boardsize-m,
		  snames[dragon_status(m, n)],
		  snames[worm_status[m][n]]);
	  mprintf(buf);
	  sgfAddComment(curnode, buf);
	}
      }
}


static void
showcapture(Position *pos, char *line)
{
  int i, j, x, y;
  if (line)
    if (!ascii2pos(pos, line, &i, &j)) {
      printf("\ninvalid point!\n");
      return;
    }
  if (position_attack(pos, i, j, &x, &y))
    mprintf("\nSuccessfull attack of %m at %m\n", i, j, x, y);
  else
    mprintf("\n%m cannot be attacked\n", i, j);
}


static void
showdefense(Position *pos, char *line)
{
  int i, j, x, y;
  if (line)
    if (!ascii2pos(pos, line, &i, &j)) {
      printf("\ninvalid point!\n");
      return;
    }

    if (position_attack(pos, i, j, &x, &y)) {
      if (position_find_defense(pos, i, j, &x, &y))
        mprintf("\nSuccessfull defense of %m at %m\n", i, j, x, y);
      else
        mprintf("\n%m cannot be defended\n", i, j);
    }
    else
      mprintf("\nThere is no need to defend %m\n", i, j);
}


static void
ascii_goto(Gameinfo *gameinfo, char *line)
{
  int movenumber = 0;
  if (!line)
    return;
  if (!strncmp(line, "last", 4))
    movenumber = 9999;
  else {
    if (!strncmp(line, "first", 4))
      movenumber = 1;
    else
      sscanf(line, "%i", &movenumber);
  }
  printf("goto %i\n", movenumber);
  curnode = sgf_root;
  gameinfo->to_move = sgfPlayTree(&curnode, &gameinfo->position, &movenumber);
  gameinfo->move_number = movenumber-1;
  return;
}


static int
ascii2pos(Position *pos, char *line, int *i, int *j)
{
  int d;
  char c;
  if (sscanf(line, "%c%d", &c, &d) == 2) {
    c = toupper(c);
    if (c > 'I') c--;
    *i = pos->boardsize - d;
    *j = c - 'A';
    if ((*i < 0) || (*i >= pos->boardsize)
	|| (*j<0) || (*j >= pos->boardsize)
	|| (pos->board[*i][*j] == EMPTY))
      return 0;
    else
      return 1;
  }
  return 0;
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
