/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <assert.h>

#include "gnugo.h"
#include "interface.h"
#include "liberty.h"
#include "gtp.h"

#define DECLARE(func) static int func(Gameinfo *gameinfo, char *s, int id)

DECLARE(gtp_playblack);
DECLARE(gtp_playwhite);
DECLARE(gtp_what_color);
DECLARE(gtp_countlib);
DECLARE(gtp_findlib);
DECLARE(gtp_quit);
DECLARE(gtp_loadsgf);
DECLARE(gtp_showboard);
DECLARE(gtp_trymove);
DECLARE(gtp_dump_stack);
DECLARE(gtp_popgo);
DECLARE(gtp_increase_depths);
DECLARE(gtp_decrease_depths);
DECLARE(gtp_attack);
DECLARE(gtp_defend);
DECLARE(gtp_mouse_attack);
DECLARE(gtp_mouse_defend);
DECLARE(gtp_eval_eye);
DECLARE(gtp_dragon_status);

/* List of known commands. */
static struct gtp_command commands[] = {
  {"black",           gtp_playblack},
  {"white",           gtp_playwhite},
  {"color",           gtp_what_color},
  {"countlib",        gtp_countlib},
  {"findlib",         gtp_findlib},
  {"quit",            gtp_quit},
  {"loadsgf",         gtp_loadsgf},
  {"showboard",       gtp_showboard},
  {"trymove",         gtp_trymove},
  {"dump_stack",      gtp_dump_stack},
  {"popgo",           gtp_popgo},
  {"increase_depths", gtp_increase_depths},
  {"decrease_depths", gtp_decrease_depths},
  {"attack",          gtp_attack},
  {"defend",          gtp_defend},
  {"mouse_attack",    gtp_mouse_attack},
  {"mouse_defend",    gtp_mouse_defend},
  {"eval_eye",        gtp_eval_eye},
  {"dragon_status",   gtp_dragon_status},
  {NULL,              NULL}
};


/* Start playing using the Go Text Protocol. */
void
play_gtp(Gameinfo *gameinfo)
{
  /* Try to make sure that we have a useful level of buffering of stdout. */
#ifdef HAVE_SETLINEBUF
  setlinebuf(stdout);
#else
  setbuf(stdout, NULL);
#endif

  /* Prepare pattern matcher and reading code. */
  initialize_engine();

  gtp_main_loop(gameinfo, commands);
}


/****************************
 * Administrative commands. *
 ****************************/

/* Function:  Quit
 * Arguments: none
 * Fails:     never
 * Returns:   nothing
 */
static int
gtp_quit(Gameinfo *gameinfo, char *s, int id)
{
  UNUSED(gameinfo);
  UNUSED(s);
  gtp_success(id, "");
  return GTP_QUIT;
}


/******************
 * Playing moves. *
 ******************/

/* Function:  Play a black stone at the given vertex.
 * Arguments: vertex
 * Fails:     invalid vertex
 * Returns:   nothing
 */
static int
gtp_playblack(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  position_play_move(&gameinfo->position, i, j, BLACK);
  return gtp_success(id, "");
}


/* Function:  Play a white stone at the given vertex.
 * Arguments: vertex
 * Fails:     invalid vertex
 * Returns:   nothing
 */
static int
gtp_playwhite(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  position_play_move(&gameinfo->position, i, j, WHITE);
  return gtp_success(id, "");
}


/* Function:  Load an sgf file, possibly up to a move number or the first
 *            occurence of a move.           
 * Arguments: filename + move number, vertex, or nothing
 * Fails:     missing filename or failure to open or parse file
 * Returns:   nothing
 */
static int
gtp_loadsgf(Gameinfo *gameinfo, char *s, int id)
{
  char filename[GTP_BUFSIZE];
  SGFNode *sgf;
  int n;
  if (sscanf(s, "%s %n", filename, &n) != 1)
    return gtp_failure(id, "missing filename");
  
  if ((sgf = readsgffile(filename)) == NULL)
    return gtp_failure(id, "cannot open or parse '%s'", filename);

  ginfo_clear(gameinfo, 19);
  load_sgf_header(sgf, gameinfo);
  load_sgf_file(sgf, gameinfo, s+n);
  return gtp_success(id, "");
}


/*****************
 * Board status. *
 *****************/

/* Function:  Return the color at a vertex.
 * Arguments: vertex
 * Fails:     invalid vertex
 * Returns:   "black", "white", or "empty"
 */
static int
gtp_what_color(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");
  
  return gtp_success(id, color_to_string(p[i][j]));
}


/* Function:  Count number of liberties for the string at a vertex.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   Number of liberties.
 */
static int
gtp_countlib(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  return gtp_success(id, "%d", countlib(i, j));
}


/* Function:  Return the positions of the liberties for the string at a vertex.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   Sorted space separated list of vertices.
 */
static int
gtp_findlib(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int libi[MAXLIBS];
  int libj[MAXLIBS];
  int libs;
  
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  libs = findlib(i, j, MAXLIBS, libi, libj);
  gtp_printid(id, GTP_SUCCESS);
  gtp_print_vertices(libs, libi, libj, board_size);
  return gtp_finish_response();
}

/**********************
 * Retractable moves. *
 **********************/

/* Function:  Play a stone of the given color at the given vertex.
 * Arguments: move (color + vertex)
 * Fails:     invalid color, invalid vertex, illegal move
 * Returns:   nothing
 */
static int
gtp_trymove(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int color;
  if (!gtp_decode_move(s, gameinfo->position.boardsize, &color, &i, &j))
    return gtp_failure(id, "invalid color or coordinate");

  if (!trymove(i, j, color, "gtp_trymove", -1, -1))
    return gtp_failure(id, "illegal move");

  return gtp_success(id, "");
}


/* Function:  Undo a trymove.
 * Arguments: none
 * Fails:     stack empty
 * Returns:   nothing
 */
static int
gtp_popgo(Gameinfo *gameinfo, char *s, int id)
{
  UNUSED(gameinfo);
  UNUSED(s);

  if (stackp == 0)
    return gtp_failure(id, "Stack empty.\n");

  popgo();
  return gtp_success(id, "");
}


/*********************
 * Tactical reading. *
 *********************/

/* Function:  Try to attack a string.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   attack code followed by attack point if attack code nonzero.
 */
static int
gtp_attack(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int ai, aj;
  int attack_code;
  
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  attack_code = attack(i, j, &ai, &aj);
  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("%d", attack_code);
  if (attack_code > 0) {
    gtp_printf(" ");
    gtp_print_vertex(ai, aj, board_size);
  }
  return gtp_finish_response();
}  


/* Function:  Try to defend a string.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   defense code followed by defense point if defense code nonzero.
 */
static int
gtp_defend(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int di, dj;
  int defend_code;
  
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  defend_code = find_defense(i, j, &di, &dj);
  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("%d", defend_code);
  if (defend_code > 0) {
    gtp_printf(" ");
    gtp_print_vertex(di, dj, board_size);
  }
  return gtp_finish_response();
}  


/* Function:  Increase depth values by one.
 * Arguments: none
 * Fails:     never
 * Returns:   nothing
 */
static int
gtp_increase_depths(Gameinfo *gameinfo, char *s, int id)
{
  UNUSED(gameinfo);
  UNUSED(s);
  increase_depth_values();
  return gtp_success(id, "");
}  


/* Function:  Decrease depth values by one.
 * Arguments: none
 * Fails:     never
 * Returns:   nothing
 */
static int
gtp_decrease_depths(Gameinfo *gameinfo, char *s, int id)
{
  UNUSED(gameinfo);
  UNUSED(s);
  decrease_depth_values();
  return gtp_success(id, "");
}  


/******************
 * mouse reading. *
 ******************/

static char goal[MAX_BOARD][MAX_BOARD];

/* Function:  Try to attack a dragon.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   attack code followed by attack point if attack code nonzero.
 */
static int
gtp_mouse_attack(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int ai, aj;
  int attack_code;
  int save_verbose = verbose;
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  verbose=0;
  examine_position(p[i][j], EXAMINE_DRAGONS);
  verbose=save_verbose;
  
  attack_code = mouse_attack(i, j, &ai, &aj, goal);
  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("%d", attack_code);
  if (attack_code > 0) {
    gtp_printf(" ");
    gtp_print_vertex(ai, aj, board_size);
  }
  return gtp_finish_response();
}  


/* Function:  Try to defend a dragon.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   defense code followed by defense point if defense code nonzero.
 */
static int
gtp_mouse_defend(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int di, dj;
  int defend_code;
  int save_verbose = verbose;
  
  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  verbose=0;
  examine_position(p[i][j], EXAMINE_DRAGONS);
  verbose=save_verbose;

  defend_code = mouse_defend(i, j, &di, &dj, goal);
  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("%d", defend_code);
  if (defend_code > 0) {
    gtp_printf(" ");
    gtp_print_vertex(di, dj, board_size);
  }
  return gtp_finish_response();
}  


/********
 * eyes *
 ********/

/* Function:  Evaluate an eye space
 * Arguments: vertex
 * Fails:     invalid vertex, vertex not eye space of unique color
 * Returns:   Minimum and maximum number of eyes. If these differ an
 *            attack and a defense point are additionally returned.
 */

static int
gtp_eval_eye(Gameinfo *gameinfo, char *s, int id)
{
  int m, n;
  int max, min;
  int attacki, attackj;
  int defendi, defendj;
  int i, j;
  int save_verbose = verbose;

  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &m, &n))
    return gtp_failure(id, "invalid coordinate");

  verbose = 0;
  examine_position(BLACK, EXAMINE_DRAGONS);
  verbose = save_verbose;
  
  if (black_eye[m][n].color == BLACK_BORDER) {
    i = black_eye[m][n].origini;
    j = black_eye[m][n].originj;
    compute_eyes(i, j, &max, &min, &attacki, &attackj, &defendi, &defendj,
		 black_eye, 0, EMPTY);
  }
  else if (white_eye[m][n].color == WHITE_BORDER) {
    i = white_eye[m][n].origini;
    j = white_eye[m][n].originj;
    compute_eyes(i, j, &max, &min, &attacki, &attackj, &defendi, &defendj,
		 white_eye, 0, EMPTY);
  }
  else
    return gtp_failure(id, "not an eye or not of unique color");

  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("%d %d", min, max);
  if (max != min) {
    gtp_printf(" ");
    gtp_print_vertex(attacki, attackj, gameinfo->position.boardsize);
    gtp_printf(" ");
    gtp_print_vertex(defendi, defendj, gameinfo->position.boardsize);
  }
  return gtp_finish_response();
}


/*****************
 * dragon status *
 *****************/

/* Function:  Determine status of a dragon.
 * Arguments: vertex
 * Fails:     invalid vertex, empty vertex
 * Returns:   status ("alive", "critical", "dead", or "unknown"),
 *            attack point, defense point. Points of attack and
 *            defense are only given if the status is critical and the
 *            mouse code is enabled.
 *
 * FIXME: Should give attack and defense points even without the mouse
 *        option. Should be able to distinguish between life in seki
 *        and life with territory. Should also be able to identify ko.
 */

static int
gtp_dragon_status(Gameinfo *gameinfo, char *s, int id)
{
  int i, j;
  int save_verbose = verbose;

  if (!gtp_decode_coord(s, gameinfo->position.boardsize, &i, &j))
    return gtp_failure(id, "invalid coordinate");

  if (p[i][j] == EMPTY)
    return gtp_failure(id, "vertex must not be empty");

  verbose = 0;
  examine_position(BLACK, EXAMINE_DRAGONS);
  verbose = save_verbose;
  
  /* FIXME: We should also call the semeai module. */

  if (!mouse || dragon[i][j].mouse_status == UNCHECKED) {
    if (dragon[i][j].status == ALIVE)
      return gtp_success(id, "alive");
  
    if (dragon[i][j].status == DEAD)
      return gtp_success(id, "dead");
  
    if (dragon[i][j].status == UNKNOWN)
      return gtp_success(id, "unknown");

    assert(dragon[i][j].status == CRITICAL); /* Only remaining possibility. */
    return gtp_success(id, "critical");
  }

  /* Mouse code active. */
  if (dragon[i][j].mouse_status == ALIVE)
    return gtp_success(id, "alive");
  
  if (dragon[i][j].mouse_status == DEAD)
    return gtp_success(id, "dead");
  
  if (dragon[i][j].mouse_status == UNKNOWN)
    return gtp_success(id, "unknown");
  
  assert(dragon[i][j].mouse_status == CRITICAL);
  /* Status critical, need to return attack and defense point as well. */
  gtp_printid(id, GTP_SUCCESS);
  gtp_printf("critical ");
  gtp_print_vertex(dragon[i][j].mouse_attacki, dragon[i][j].mouse_attackj,
		   gameinfo->position.boardsize);
  gtp_printf(" ");
  gtp_print_vertex(dragon[i][j].mouse_defendi, dragon[i][j].mouse_defendj,
		   gameinfo->position.boardsize);
  return gtp_finish_response();
}



/*********
 * debug *
 *********/

/* Function:  Write the position to stderr.
 * Arguments: none
 * Fails:     never
 * Returns:   nothing
 */
static int
gtp_showboard(Gameinfo *gameinfo, char *s, int id)
{
  UNUSED(gameinfo);
  UNUSED(s);
  showboard(0);
  return gtp_success(id, "");
}


/* Function:  Dump stack to stderr.
 * Arguments: none
 * Fails:     never
 * Returns:   nothing
 */
static int
gtp_dump_stack(Gameinfo *gameinfo, char *s, int id)
{
  UNUSED(gameinfo);
  UNUSED(s);
  dump_stack();
  return gtp_success(id, "");
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
