/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*  Parts of this code were given to us by Tommy Thorn */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <time.h>

#include "sgf_properties.h"
#include "ttsgf.h"
#include "sgfana.h"
#include "liberty.h"
#include "interface.h"
#include "gg_utils.h"


SGFNode *lastnode=0; /*last node where move was placed*/


/* ================================================================ */
/*                     Some utility functions.                      */
/* ================================================================ */

/* General FIXME:  Move these somewhere? */

/*
 * Utility: a checking, initializing malloc
 */

void *
xalloc(unsigned int size)
{
  void *pt = malloc(size);

  if (!pt) {
    fprintf(stderr, "%s: Out of memory!\n","xalloc");
    exit(2);
  }

  memset(pt, 0, (unsigned long) size);
  return pt;
}


/* ================================================================ */
/*                           SGF Nodes                              */
/* ================================================================ */


/*
 * Allocate memory for a new SGF node.
 */

SGFNode *
sgfNewNode()
{
  return xalloc(sizeof(SGFNode));
}


/*
 * Set a the node as the default node.
 *
 * FIXME: Why not call it sgfSetDefaultNode then?
 */

void
sgfSetLastNode(SGFNode *last_node)
{
    lastnode=last_node;
}


/*
 * Add a generic text property to an SGF node.
 */

void
sgfAddProperty(SGFNode *node, const char *name, const char *value)
{
  SGFProperty *prop = node->props;

  if (prop)
    while (prop->next)
      prop=prop->next;

  sgfMkProperty(name, value, node, prop);
}


/*
 * Add an integer property to an SGF node.
 */

void
sgfAddPropertyInt(SGFNode *node, const char *name, long val)
{
  char buffer[10];

  gg_snprintf(buffer, 10, "%ld", val);
  sgfAddProperty(node, name, buffer);
}


/*
 * Add a float property to an SGF node.
 */

void
sgfAddPropertyFloat(SGFNode *node, const char *name, float val)
{
  char buffer[10];

  gg_snprintf(buffer, 10, "%3.1f", val);
  sgfAddProperty(node, name, buffer);
}


/*
 * Read a property as int from an SGF node.
 */

int
sgfGetIntProperty(SGFNode *node, const char *name, int *value)
{
  SGFProperty *prop;
  short nam = name[0] | name[1] << 8;

  for (prop = node->props; prop; prop=prop->next)
    if (prop->name == nam) {
      *value = atoi(prop->value);
      return 1;
    }

  return 0;
}


/*
 * Read a property as float from an SGF node.
 */

int
sgfGetFloatProperty(SGFNode *node, const char *name, float *value)
{
  SGFProperty *prop;
  short nam = name[0] | name[1] << 8;

  for (prop = node->props; prop; prop=prop->next)
    if (prop->name == nam) {
      *value = (float) atof(prop->value);
      /* MS-C warns of loss of data (double to float) */
      return 1;
    }

  return 0;
}


/*
 * Read a property as text from an SGF node.
 */

int
sgfGetCharProperty(SGFNode *node, const char *name, char **value)
{
  SGFProperty *prop;
  short nam = name[0] | name[1] << 8;

  for (prop = node->props; prop; prop=prop->next)
    if (prop->name == nam) {
      *value=prop->value;
      return 1;
    }

  return 0;
}


/*
 * Overwrite a property from an SGF node with text or create a new
 * one if it does not exist.
 */

void
sgfOverwriteProperty(SGFNode *node, const char *name, const char *text)
{
  SGFProperty *prop;
  short nam = name[0] | name[1] << 8;

  for (prop = node->props; prop; prop=prop->next)
    if (prop->name == nam) {
      /* FIXME: Realloc pr with strlen text. */
      strncpy(prop->value, text, strlen(prop->value));
      return;
    }

  sgfAddProperty(node, name, text);
}


/* 
 * Overwrite a float property in the gametree with val or create
 * a new one if it does not exist.
 */

void
sgfOverwritePropertyFloat(SGFNode *node, const char *name, float val)
{
  SGFProperty *prop;
  short nam = name[0] | name[1] << 8;

  for (prop = node->props; prop; prop=prop->next)
    if (prop->name == nam) {
      /* FIXME: Realloc pr. */
      gg_snprintf(prop->value,strlen(prop->value)+1, "%3.1f", val);
      return;
    }

  sgfAddPropertyFloat(node, name, (float) val);
}


/*
 * Overwrite an int property in an SGF node with val or create a new 
 * one if it does not exist.
 */

void
sgfOverwritePropertyInt(SGFNode *node, const char *name, int val)
{
  SGFProperty *prop;
  short nam = name[0] | name[1] << 8;

  for (prop = node->props; prop; prop=prop->next)
    if (prop->name == nam) {
      /* FIXME: Realloc pr. */
      gg_snprintf(prop->value,strlen(prop->value)+1,"%d", val);
      return;
   }

  sgfAddPropertyInt(node, name, (int) val);
}


/*
 * Goto previous node.
 */

SGFNode *
sgfPrev(SGFNode *node)
{
  SGFNode *q;
  SGFNode *prev;

  if (!node->parent)
    return NULL;

  q = node->parent->child;
  prev = NULL;
  while (q && q != node) {
    prev = q;
    q = q->next;
  }

  return prev;
}


/*
 * Goto root node.
 */

SGFNode *
sgfRoot(SGFNode *node)
{
  while (node->parent)
    node = node->parent;

  return node;
}


/* ================================================================ */
/*                         SGF Properties                           */
/* ================================================================ */


/*
 * Make an SGF property.
 *
 * FIXME: To speed up, we should have MkMoveProperty.
 */

SGFProperty *
sgfMkProperty(const char *name, const  char *value,
	      SGFNode *node, SGFProperty *last)
{
  SGFProperty *prop;

  prop = xalloc(sizeof(SGFProperty)+strlen(value));
  if (strlen(name) == 1) 
    prop->name = name[0] | (short) (' ' << 8);
  else
    prop->name = name[0] | name[1] << 8;
  strcpy(prop->value, value);

  if (last == NULL)
    node->props = prop;
  else
    last->next = prop;

  return prop;
}


/* ================================================================ */
/*                        High level functions                      */
/* ================================================================ */


/*
 * Returns the node to modify. When p==0 then lastnode is used,
 * except if lastnode is NULL, then the current end of game is used.
 */

SGFNode *
sgfNodeCheck(SGFNode *node)
{
   assert(sgf_root);

   if (!node) {
     if (lastnode)
       node=lastnode;
     else {
       node=sgf_root;
       if (node->child) {
	 while (node->child)
	   node = node->child;
       }
     }
   }

   return node;
}


/*
 * Add a stone to the current or the given node.
 * Return the node where the stone was added.
 */

SGFNode *
sgfAddStone(SGFNode *node, int color, int movex, int movey)
{
  char move[3];

  node=sgfNodeCheck(node);
  sprintf(move,"%c%c", movey+'a', movex+'a');
  sgfAddProperty(node, (color==BLACK) ? "AB":"AW", move);

  return node;
}


/*
 * Add a move to the gametree.
 */

SGFNode *
sgfAddPlay(SGFNode *node, int who, int movex, int movey)
{
  char move[3];
  SGFNode *new;
  SGFNode *last = sgfNodeCheck(node);

  /* a pass move? */
  if ((movex == movey) && (movey == -1))
    move[0] = 0;
  else
    sprintf(move, "%c%c", movey+'a', movex+'a');

  if (last->child) {
    lastnode = sgfStartVariantFirst(last->child);
    sgfAddProperty(lastnode, (who==BLACK) ? "B":"W", move);

    return lastnode;
  }

  new = sgfNewNode();
  last->child = new;
  new->parent = last;
  lastnode = new;
  sgfAddProperty(new, (who==BLACK) ? "B":"W", move);

  return new;
}


SGFNode *
sgfCreateHeaderNode(int boardsize, float komi)
{
    SGFNode *root=sgfNewNode();

    sgfAddPropertyInt(root, "SZ", boardsize);
    sgfAddPropertyFloat(root, "KM", komi);
    sgf_root=root;

    return root;
}


/*
 * Add a comment to the gametree.
 */

SGFNode *
sgfAddComment(SGFNode *node, const char *comment)
{
  SGFNode * ret;

  assert(sgf_root);
  ret=sgfNodeCheck(node);
  sgfAddProperty(ret,"C ",comment);

  return ret;
}


/*
 * Place text on the board at position (i,j).
 */

SGFNode *
sgfBoardText(SGFNode *node,int i,int j, const char *text)
{
  void *str = xalloc(strlen(text)+3);

  assert(sgf_root);

  node=sgfNodeCheck(node);
  sprintf(str,"%c%c:%s",j+'a',i+'a',text);
  sgfAddProperty(node,"LB",str);
  free(str);

  return node;
}


/*
 * Place a character on the board at position (i,j).
 */

SGFNode *
sgfBoardChar(SGFNode *node,int i,int j, char c)
{
  char text[2]="";

  text[0]=c;
  text[1]=0;

  return sgfBoardText(node,i,j,text);
}


/*
 * Place a number on the board at position (i,j).
 */

SGFNode * 
sgfBoardNumber(SGFNode *node,int i,int j, int number)
{
  char text[10];

  node=sgfNodeCheck(node);
  gg_snprintf(text,10,"%c%c:%i",j+'a',i+'a',number);
  sgfAddProperty(node, "LB", text);

  return node;
}


/*
 * Place a circle mark on the board at position (i,j).
 */

SGFNode * 
sgfTriangle(SGFNode *node,int i,int j)
{
  char text[3];

  node=sgfNodeCheck(node);
  gg_snprintf(text,3,"%c%c",j+'a',i+'a');
  sgfAddProperty(node,"TR",text);

  return node;
}


/*
 * Place a circle mark on the board at position (i,j).
 */

SGFNode * 
sgfCircle(SGFNode *node,int i,int j)
{
  char text[3];

  node=sgfNodeCheck(node);
  gg_snprintf(text,3,"%c%c",j+'a',i+'a');
  sgfAddProperty(node, "CR", text);

  return node;
}


/*
 * Place a square mark on the board at position (i,j).
 */

SGFNode *
sgfSquare(SGFNode *node,int i,int j)
{
  return sgfMark(node,i,j);   /*cgoban 1.9.5 does not understand SQ*/
}


/*
 * Place a (square) mark on the board at position (i,j).
 */

SGFNode *
sgfMark(SGFNode *node,int i,int j)
{
  char text[3];

  node=sgfNodeCheck(node);
  gg_snprintf(text,3,"%c%c",j+'a',i+'a');
  sgfAddProperty(node,"MA",text);

  return node;
}


/*
 * Start a new variant. Returns a pointer to the new node.
 */

SGFNode *
sgfStartVariant(SGFNode *node)
{
  node=sgfNodeCheck(node);

  while (node->next)
    node = node->next;
  node->next=sgfNewNode();
  node->next->parent=node;

  return node->next;
}


/*
 * Start a new variant as first child. Returns a pointer to the new node.
 */

SGFNode *
sgfStartVariantFirst(SGFNode *node)
{
  SGFNode * old_first_child=sgfNodeCheck(node);
  SGFNode * new_first_child=sgfNewNode();

  new_first_child->next=old_first_child;
  if (old_first_child->parent)
    new_first_child->parent=old_first_child->parent;
  old_first_child->parent=new_first_child;
  if (new_first_child->parent)
    new_first_child->parent->child=new_first_child;

  return new_first_child;
}


/*
 * Perform the moves and place the stones from the SGF Tree on the 
 * board.  Return whose turn it is to move (WHITE or BLACK).
 */

int
sgfAddMoveFromTree(SGFNode *node, Gameinfo *gameinfo, int next_to_move)
{
  int i,j;
  int next=next_to_move;
  SGFProperty *prop;

  for (prop=node->props; prop; prop=prop->next) {
    switch(prop->name) {
    case SGFAB:
      /* A black stone. */
      i=prop->value[1]-'a';
      j=prop->value[0]-'a';
      position_add_stone(&gameinfo->position, i, j, BLACK);
      break;

    case SGFAW:
      /* A white stone. */
      i=prop->value[1]-'a';
      j=prop->value[0]-'a';
      position_add_stone(&gameinfo->position, i, j, WHITE);
      break;

    case SGFPL:
      /* Player property - who is next to move? */
      if (prop->value[0] == 'w' || prop->value[0] == 'W')
	next = WHITE;
      else
	next = BLACK;
      break;

    case SGFW:
    case SGFB:
      /* An ordinary move. */
      next = (prop->name == SGFW) ? WHITE : BLACK;
      /* FIXME: We should use get_moveXY() from sgf_utils.c instead. */
      if ((!prop->value)||(prop->value[0]<'a')) {
	i = -1;
	j = -1;
      }
      else {
	i = prop->value[1]-'a';
	j = prop->value[0]-'a';
      }

      if (i >= gameinfo->position.boardsize) {
	i = -1;
	j = -1;
      }

      gameinfo->move_number++;
      position_play_move(&gameinfo->position, i, j, next);
      next=OTHER_COLOR(next);
      break;
    }
  }

  return next;
}


/*
 * Play the moves in ROOT UNTIL movenumber is reached.
 */
/* FIXME: Move to another file since they use Position instead of just 
          handling SGF trees? */
int
sgfPlayTree(SGFNode **root, Position *position, int *until)
{
  SGFNode *node;
  int next=BLACK;
  SGFProperty *prop;
  int i,j;
  int movenumber=0;

  /* Start from the empty board. */
  position_clear(position, position->boardsize);

  for (node=*root; node; node=node->child) {
    for (prop=node->props; prop; prop=prop->next) {
      switch(prop->name) {
      case SGFAB:
	i = prop->value[1] - 'a';
	j = prop->value[0] - 'a';
	position_add_stone(position, i, j, BLACK);
	break;

      case SGFAW:
	i = prop->value[1] - 'a';
	j = prop->value[0] - 'a';
	position_add_stone(position, i, j, WHITE);
	break;

      case SGFPL:
	if (prop->value[0] == 'w' || prop->value[0] == 'W')
	  next = WHITE;
	else
	  next = BLACK;
	break;

      case SGFW:
      case SGFB:
	next = prop->name == SGFW ? WHITE : BLACK;
	/* FIXME: We should use get_moveXY() from sgf_utils.c instead. */
	if ((!prop->value) || (prop->value[0]<'a')) {
	  i = -1;
	  j = -1;
	}
	else {
	  i = prop->value[1]-'a';
	  j = prop->value[0]-'a';
	}
	if (i >= position->boardsize) {
	  i = -1;
	  j = -1;
	}
	
	movenumber++;
	if (movenumber == *until) {
	  *root=node->parent;
	  return next;
	}
	position_play_move(position, i, j, next);
	next=OTHER_COLOR(next);
	break;
      }
    }
  }
  *until=movenumber;

  return next;
}




/* ================================================================ */
/*                          Read SGF tree                           */
/* ================================================================ */


#define MAX_FILE_BUFFER 200000 /* buffer for reading SGF file. */

/*
 * SGF grammar:
 *
 * Collection = GameTree { GameTree }
 * GameTree   = "(" Sequence { GameTree } ")"
 * Sequence   = Node { Node }
 * Node       = ";" { Property }
 * Property   = PropIdent PropValue { PropValue }
 * PropIdent  = UcLetter { UcLetter }
 * PropValue  = "[" CValueType "]"
 * CValueType = (ValueType | Compose)
 * ValueType  = (None | Number | Real | Double | Color | SimpleText |
 *               Text | Point  | Move | Stone)
 *
 * The above grammar has a number of simple properties which enables us
 * to write a simpler parser:
 *   1) There is never a need for backtracking
 *   2) The only recursion is on gametree.
 *   3) Tokens are only one character
 * 
 * We will use a global state to keep track of the remaining input
 * and a global char variable, `lookahead' to hold the next token.  
 * The function `nexttoken' skips whitespace and fills lookahead with 
 * the new token.
 */


#define STRICT 's'
#define LAX    'l'


static void parse_error(const char *msg, int arg);
static void nexttoken(void);
static void match(int expected);


static FILE *sgffile;


#define sgf_getch() (getc(sgffile))


static char *sgferr;
#ifdef TEST_SGFPARSER
static int sgferrarg;
#endif
static int sgferrpos;

static int lookahead;


/* ---------------------------------------------------------------- */
/*                       Parsing primitives                         */
/* ---------------------------------------------------------------- */


static void
parse_error(const char *msg, int arg)
{
  fprintf(stderr, msg, arg);
  exit(EXIT_FAILURE);
}


static void
nexttoken()
{
  do
    lookahead = sgf_getch();
  while (isspace(lookahead));
}


static void
match(int expected)
{
  if (lookahead != expected)
    parse_error("expected: %c", expected);
  else
    nexttoken();
}

/* ---------------------------------------------------------------- */
/*                        The parser proper                         */
/* ---------------------------------------------------------------- */


static void
propident(char *buffer, int size)
{
  if (lookahead == -1 || !isupper(lookahead)) 
    parse_error("Expected an upper case letter", 0);
  while (lookahead != -1 && isalpha(lookahead)) {
    if (isupper(lookahead) && size > 1) {
      *buffer++ = lookahead;
      size--;
    }
    nexttoken();
  }
  *buffer = '\0';
}


static void
propvalue(char *buffer, int size)
{
  char *p = buffer;

  match('[');
  while (lookahead != ']' && lookahead != '\0') {
    if (lookahead == '\\') {
      lookahead = sgf_getch();
      /* Follow the FF4 definition of backslash */
      if (lookahead == '\r') {
	lookahead = sgf_getch();
	if (lookahead == '\n') 
	  lookahead = sgf_getch();
      } else if (lookahead == '\n') {
	lookahead = sgf_getch();
	if (lookahead == '\r') 
	  lookahead = sgf_getch();
      }
    }
    if (size > 1) {
      *p++ = lookahead;
      size--;
    }
    lookahead = sgf_getch();
  }
  match(']');
  /* Remove trailing whitespace */
  --p;
  while (p > buffer && isspace((int) *p))
    --p;
  *++p = '\0';
}


static SGFProperty *
property(SGFNode *n, SGFProperty *last)
{
  char name[3];
  char buffer[4000];

  propident(name, sizeof(name));
  do {
    propvalue(buffer, sizeof(buffer));
    last = sgfMkProperty(name, buffer, n, last);
  } while (lookahead == '[');
  return last;
}


static void
node(SGFNode *n)
{
  SGFProperty *last = NULL;
  match(';');
  while (lookahead != -1 && isupper(lookahead))
    last = property(n, last);
}


static SGFNode *
sequence(SGFNode *n)
{
  node(n);
  while (lookahead == ';') {
    SGFNode *new = sgfNewNode();
    new->parent = n;
    n->child = new; n = new;
    node(n);
  }
  return n;
}


static void
gametree(SGFNode **p, SGFNode *parent, int mode) 
{
  if (mode == STRICT)
    match('(');
  else
    for (;;) {
      if (lookahead == -1) {
	parse_error("Empty file?", 0);
	break;
      }
      if (lookahead == '(') {
	while (lookahead == '(')
	  nexttoken();
	if (lookahead == ';')
	  break;
      }
      nexttoken();
    }

  /* The head is parsed */
  {
    SGFNode *head = sgfNewNode();
    SGFNode *last;

    head->parent = parent;
    *p = head;

    last = sequence(head);
    p = &last->child;
    while (lookahead == '(') {
      gametree(p, last->parent, STRICT);
      p = &((*p)->next);
    }
    if (mode == STRICT)
      match(')');
  }
}


/*
 * Wrapper around readsgf which reads from a file rather than a string.
 * Returns NULL if file will not open, or some other parsing error.
 */

SGFNode *
readsgffile(const char *filename)
{
  SGFNode *root;
  int tmpi = 0;

  if (strcmp(filename, "-") == 0)
    sgffile = stdin;
  else
    sgffile = fopen(filename, "r");

  if (!sgffile)
    return NULL;


  nexttoken();
  gametree(&root, NULL, LAX);

  fclose(sgffile);

  if (sgferr) {
    fprintf(stderr, "Parse error: %s at position %d\n", sgferr, sgferrpos);
    return NULL;
  }

  /* perform some simple checks on the file */
  if (!sgfGetIntProperty(root, "GM", &tmpi)) {
    fprintf(stderr, "Couldn't find the game type (GM) attribute!\n");
  }
  if (tmpi != 1) {
    fprintf(stderr, "SGF file might be for game other than go: %d\n", tmpi);
    fprintf(stderr, "Trying to load anyway.\n");
  }


  /* FIXME: These warnings should be printed, but only if some suitable
   *        trace/debug option has been specified.
   */
#if 0
  if (!sgfGetIntProperty(root, "FF", &tmpi)) {
    fprintf(stderr, "Can not determine SGF spec version (FF)!\n");
  }
  if (tmpi<3 || tmpi>4) {
    fprintf(stderr, "Unsupported SGF spec version: %d\n", tmpi);
  }
#endif

  return root;
}





#ifdef TEST_SGFPARSER
int
main()
{
  static char buffer[25000];
  static char output[25000];
  SGFNode *game;

  sgffile = stdin;

  nexttoken();
  gametree(&game, LAX);
  if (sgferr) {
    fprintf(stderr, "Parse error:");
    fprintf(stderr, sgferr, sgferrarg);
    fprintf(stderr, " at position %d\n", sgferrpos);
  } else {
    unparse_game(stdin,game);
    write(1,output,outputp-output);
  }
}
#endif


/* ================================================================ */
/*                          Write SGF tree                          */
/* ================================================================ */


#define OPTION_STRICT_FF4 0


/*
 * p->child is the next move.
 * p->next  is the next variation
 */


#if 0
void
sgf_putch(char c)
{
  fputc(c,ttsgfout);
}


void
sgf_puts(const char *s)
{
  fputs(s,ttsgfout);
}


void
sgf_putint(int i)
{
  char buf[10], *p=buf;
  
  if (i < 0)
    sgf_putch('-'), i = -i;
  do {
    *p++ = i % 10;
    i /= 10;
  } while (i > 9);
  while (p > buf)
    sgf_putch((char) (*--p + '0'));
  /* MS-C complains even of this conversion, if not cast... */
}


static void
unparse_prop(SGFProperty *p)
{
  char name[3], *cp;

  name[2] = 0;
  while (p != NULL) {
    name[0] = p->name & 255;
    name[1] = p->name >> 8;
    if (name[1] == ' ') name[1] = 0;
    sgf_puts(name);
    sgf_putch('[');
    for (cp = p->value; *cp; cp++) {
      if (*cp == '[' || *cp == ']' || *cp == '\\')
	sgf_putch('\\');
      sgf_putch(*cp);
    }
    sgf_putch(']');
    p = p->next;
  }
}
#endif


/*
 * Print all unprinted property values at node N to file.
 */

void
sgfPrintAllProperties(FILE *file, SGFNode *node)
{
  SGFProperty *prop;
  SGFProperty *last;
  
  prop = node->props;
  if (!prop)
    return;

  do {
    if (!(prop->name&0x20)) {
      last=prop;
          
      fprintf(file, "%c%c[%s]", prop->name&0xff, (prop->name>>8) & 0xff, 
	      prop->value);
      prop=prop->next;
      while (prop) {
	if (last->name==prop->name) {
	  fprintf(file,"[%s]",prop->value);
	  prop->name|=0x20;  /*indicate already printed*/
	}
	prop=prop->next;
      };
      prop=last;
      prop->name|=0x20;  /*indicate already printed*/
    }
    prop=prop->next;
  } while (prop);
}


/*
 * Print the property values of NAME at node N and mark it as printed. 
 */

int
sgfPrintCharProperty(FILE *file, SGFNode *node, const char *name)
{
  int first=1;
  SGFProperty *prop;
  short nam = name[0] | name[1] << 8;

  for (prop = node->props; prop; prop=prop->next) {
    if (prop->name == nam) {
      prop->name|=0x20;  /*indicate already printed*/
      if (first) {
	/*    if (name[1]==' ')
	      name[1]='';
	      */    
	fprintf(file,"%s",name);
	first=0;
      }
      fprintf(file,"[%s]",prop->value);
    }
  }

  return !first;
}


/*
 * Print comments form Node node.
 *
 * NOTE: cgoban does not print "C[comment1][comment2]" and I don't know
 *       what the sgfspec says.
 */

int
sgfPrintCommentProperty(FILE *file, SGFNode *node, const char *name)
{
  int first=1;
  SGFProperty *prop;
  short nam = name[0] | name[1] << 8;

  for (prop = node->props; prop; prop=prop->next) {
    if (prop->name == nam) {
      prop->name|=0x20;  /*indicate already printed*/
      if (first) {
	/*   if (name[1]==' ')
	     name[1]='';
	     */
	fprintf(file,"%s[%s",name,prop->value);
	first=0;
      } else
	  fprintf(file,"\n%s",prop->value);
    }
  }

  if (!first) {
    fprintf(file,"]");
    return 1;
  } else
    return 0;
}


void
unparse_node(FILE *file, SGFNode *node)
{
  fputs("\n;",file);
  sgfPrintCharProperty(file, node,"B ");
  sgfPrintCharProperty(file, node,"W ");
  sgfPrintCommentProperty(file, node,"N ");
  sgfPrintCommentProperty(file, node,"C ");
  if (sgfPrintCommentProperty(file, node, "PB")
      | sgfPrintCommentProperty(file,node,"BR"))
    fputs("\n",file);
  if (sgfPrintCommentProperty(file, node,"PW")
      | sgfPrintCommentProperty(file,node,"WR"))
    fputs("\n",file);
  sgfPrintAllProperties(file,node);
}


void
unparse_root(FILE *file, SGFNode *node)
{
  fputs("\n;",file);
  sgfPrintCharProperty(file, node,"FF");
  fputs("\n",file);
  if (sgfPrintCharProperty(file, node,"GM"))
      fputs("\n",file);
  else
      fputs("GM[1]\n",file);

  if (sgfPrintCharProperty(file, node,"SZ"))
      fputs("\n",file);
  if (sgfPrintCharProperty(file, node,"GN"))
      fputs("\n",file);
  if (sgfPrintCharProperty(file, node,"DT"))
      fputs("\n",file);
  if (sgfPrintCommentProperty(file, node,"PB")
     | sgfPrintCommentProperty(file, node,"BR"))
      fputs("\n",file);
  if (sgfPrintCommentProperty(file, node,"PW")
     | sgfPrintCommentProperty(file, node,"WR"))
      fputs("\n",file);
  sgfPrintCharProperty(file, node,"B ");
  sgfPrintCharProperty(file, node,"W ");
  sgfPrintCommentProperty(file, node,"N ");
  sgfPrintCommentProperty(file, node,"C ");
  sgfPrintAllProperties(file, node);
}


void
unparse_game(FILE *file, SGFNode *node)
{
  fputc('(',file);
  unparse_root(file,node);

  node = node->child;
  while (node != NULL && node->next == NULL) {
    unparse_node(file,node);
    node = node->child;
  } 

  while (node != NULL) {
    unparse_game(file,node);
    node = node->next;
  }
  fputs("\n)",file);
}

void
sgf_write_header(SGFNode *root, int overwrite, int seed, float komi)
{
  time_t curtime = time(NULL);
  struct tm *loctime = localtime(&curtime);
  char str[128];
  int dummy;

  gg_snprintf(str, 128, "GNU Go %s Random Seed %d", VERSION, seed);
  if (overwrite || !sgfGetIntProperty(root, "GN", &dummy))
    sgfOverwriteProperty(root, "GN", str);
  gg_snprintf(str, 128, "%4.4i-%2.2i-%2.2i",
	      loctime->tm_year+1900, loctime->tm_mon+1, loctime->tm_mday);
  if (overwrite || !sgfGetIntProperty(root, "DT", &dummy))
    sgfOverwriteProperty(root, "DT", str);
  if (overwrite || !sgfGetIntProperty(root, "AP", &dummy))
    sgfOverwriteProperty(root, "AP", PACKAGE":"VERSION);
  if (overwrite || !sgfGetIntProperty(root, "RU", &dummy))
    sgfOverwriteProperty(root, "RU", "Japanese");
  sgfOverwriteProperty(root, "FF", "4");
  sgfOverwritePropertyFloat(root, "KM", komi);
}

/*
 * Opens filename and writes the game stored in the sgf structure.
 */

int
writesgf(SGFNode *root, const char *filename)
{
  FILE *outfile;

  if (strcmp(filename, "-")==0) 
    outfile = stdout;
  else
    outfile = fopen(filename, "w");

  if (!outfile) {
    fprintf(stderr,"Can not open %s\n",filename);
    return 0;
  }

  unparse_game(outfile, root);
  fclose(outfile);
  return 1;
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */

